#!/usr/bin/env python

#^#
# OVSAgentServer.py
#
# the RPC Server for ovs agent.
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import xmlrpclib
import time
import sys
import socket
import sha
import os
import logging
from os.path import join, dirname, exists, getsize, basename
from OVSLog import init_log
from OVSConstants import CONFIG_DIR, SHUTDOWN_FLAG, AGENT_SHUTDOWN_MODE_FLAG
from OVSConfig import get_config, get_passwd, init_signal_handler
from OVSTimer import Timer
from OVSDaemon import daemonize
from OVSCommons import fmt_err_msg
from OVSXSysInfo import get_agent_version

#---------------------<> helper function <>----------------------------
def auth(p_username, p_password):
    """
    username/password based basic authentication
    """
    if not p_username:
        return
    p_passwd_digest = sha.new(p_password).digest().encode("base64").replace("\n", "")
    passwd = get_passwd()
    if passwd.has_key(p_username) and passwd[p_username] == p_passwd_digest:
        return p_username

#----------------------------------------------------------------------
from DocXMLRPCServer import DocXMLRPCRequestHandler
class AgentXMLRPCRequestHandler(DocXMLRPCRequestHandler):
    #for SSL support
    def setup(self):
        config = get_config()
        if not config.get("security", "ssl").lower().startswith("enable"):
            DocXMLRPCRequestHandler.setup(self)
            return
        #else: ssl is enabled
        self.connection = self.request
        self.rfile = socket._fileobject(self.request, "rb", self.rbufsize)
        self.wfile = socket._fileobject(self.request, "wb", self.wbufsize)

    def do_POST(self):
        """
        overwrite do_POST implement basic authentication
        """
        try:
            #authorization: Basic ZHlubzpkeW5v
            username, password = self.headers["authorization"].split(" ")[1].decode("base64").split(":")
            username = auth(username, password)
            if not username: raise Exception()
        except:
            #this will got a log entry with log_message
            self.send_error(403, "Unauthorized HTTP Access Attempt from %s!" % (self.client_address,))
            return

        config = get_config()
        if not config.get("security", "ssl").lower().startswith("enable"):
            DocXMLRPCRequestHandler.do_POST(self)
            return
        #else: #ssl enabled
        #copy from DocXMLRPCRequestHandler.do_POST
        #--------------------------------------------------------------
        try:
            # get arguments
            data = self.rfile.read(int(self.headers["content-length"]))
            # In previous versions of DocXMLRPCServer, _dispatch
            # could be overridden in this class, instead of in
            # SimpleXMLRPCDispatcher. To maintain backwards compatibility,
            # check to see if a subclass implements _dispatch and dispatch
            # using that method if present.
            response = self.server._marshaled_dispatch(data, getattr(self, '_dispatch', None))
        except: # This should only happen if the module is buggy
            # internal error, report as HTTP server error
            self.send_response(500)
            self.end_headers()
        else:
            # got a valid XML RPC response
            self.send_response(200)
            self.send_header("Content-type", "text/xml")
            self.send_header("Content-length", str(len(response)))
            self.end_headers()
            self.wfile.write(response)

            # shut down the connection
            self.wfile.flush()
            self.connection.shutdown()
        #--------------------------------------------------------------

    #redefine this function to use our own logging configuration
    def log_message(self, format, *args):
        logger = logging.getLogger()
        logger.info(format % args)

#----------------------------------------------------------------------
from DocXMLRPCServer import ServerHTMLDoc
from DocXMLRPCServer import XMLRPCDocGenerator
class OVSXMLRPCDocGenerator(XMLRPCDocGenerator):
    def generate_html_documentation(self):
        methods = {}
        print repr(dir(self.instance))
        for attr_name in dir(self.instance):
            attr = getattr(self.instance, attr_name)
            if callable(attr) and hasattr(attr, "exposed") and attr.exposed == True:
                methods[attr_name] = attr

        documenter = ServerHTMLDoc()
        documentation = documenter.docserver( self.server_name,
                self.server_documentation, methods)
        return documenter.page(self.server_title, documentation)

#----------------------------------------------------------------------
from DocXMLRPCServer import DocXMLRPCServer
from SocketServer import ForkingMixIn
class OVSAgentServer(OVSXMLRPCDocGenerator, ForkingMixIn, DocXMLRPCServer):
    """
    Subclass of DocXMLRPCServer,
        *) enable shutdown the server throu rpc call
        *) add access restriction to client ip
    """
    #http://hea-www.harvard.edu/~fine/Tech/addrinuse.html
    allow_reuse_address = True
    #default 40 in ForkingMixin, too limited when do multiple cp operation
    max_children = 500

    def __init__(self, *argv, **argk):
        DocXMLRPCServer.__init__(self, *argv, **argk)
        #^# enable SSL
        config = get_config()
        if config.get("security", "ssl").lower().startswith("enable"):
            # cmds to create self-signed certificates
            # openssl genrsa -out key.pem 1024
            # openssl req -new -key key.pem -out request.pem
            # openssl x509 -req -days 1000 -in request.pem -signkey key.pem -out certificate.pem
            # more, refer to http://www.openssl.org/docs/apps/openssl.html
            from OpenSSL import SSL
            KEYFILE = join(CONFIG_DIR, "cert/key.pem")
            CERTFILE = join(CONFIG_DIR, "cert/certificate.pem")
            ctx = SSL.Context(SSL.SSLv23_METHOD)
            ctx.use_privatekey_file(KEYFILE)
            ctx.use_certificate_file(CERTFILE)
            self.socket = SSL.Connection(ctx, socket.socket(self.address_family, self.socket_type))
            self.server_bind()
            self.server_activate()

        self.shutdown_flag = "%s.%s" % (SHUTDOWN_FLAG, os.environ.get("USER"))
        if exists(self.shutdown_flag):
            os.remove(self.shutdown_flag)

        #register ovs services
        from OVSServices import OVSServices
        self.register_instance(OVSServices(), allow_dotted_names=True)

        #shutdown admin function
        self.shutdown_flag = "%s.%s" % (SHUTDOWN_FLAG, os.environ.get("USER"))
        if exists(self.shutdown_flag):
            os.remove(self.shutdown_flag)
        self.register_function(self.shutdown)

    #retrict this function to admin user only
    def shutdown(self, username, password):
        logger = logging.getLogger()
        if auth(username, password) == "admin":
            logger.info("Shutdown flag created.")
            #create the server shutdown flag
            file(self.shutdown_flag, "w").close()
            return "success"
        else:
            return "failed:unprivileged user, permission denied!"

    def serve_forever(self):
        logger = logging.getLogger()
        while not exists(self.shutdown_flag):
            self.handle_request()
        self.server_close()
        logger.info("Server closed.")

    def verify_request(self, request, client_address):
        """
        overwrite original verify_request to implement IP based access control
        """
        #--------------------<> internal function <>----
        def match(rule, ip):
            #check if ip is in the range
            try:
                #e.g.: 219.142.73.80
                if ip == rule: return True
                #e.g.: 219.142.73.*
                if rule.endswith("*") and ip.startswith(rule[:-1]): return True
                #e.g.: 219.142.73.0/20
                vip, bits = rule.split("/")
                vip_str = "".join(["%02x" % int(x) for x in vip.split(".")])
                ip_str = "".join(["%02x" % int(x) for x in ip.split(".")])
                q, r = divmod(int(bits), 4)
                if vip_str[:q] == ip_str[:q]:
                    if int(vip_str[q], 16) >> (4 - r) == int(ip_str[q], 16) >> (4 - r):
                        return True
            except: #in case there are errors in rules
                pass
        #--------------------<> main processing <>----
        #check ip access control rules, sequence: allow=>deny
        (ip, port) = client_address

        # always allow connect from localhost
        if ip == '127.0.0.1': return True

        config = get_config()
        #allow rules
        allow = False
        allow_rules = config.get("security", "allow").split(",")
        for rule in allow_rules:
            if match(rule.strip(), ip):
                allow = True
                break
        if not allow: return False

        #deny rules
        deny = False
        deny_rules = config.get("security", "deny").split(",")
        for rule in deny_rules:
            if match(rule.strip(), ip):
                deny = True
                break
        if deny: return False

        #nothing wrong, finally
        return True

#----------------------------------------------------------------------
def start_server():
    logger = logging.getLogger()
    try:
        #hostname = socket.gethostname()
        #here we still listen on all available address,
        #because if we install agent inside vm, and gethostname()
        #may simply be localhost
        config = get_config()
        server = OVSAgentServer(
                    ('', int(config.get("agent", "port")))
                    , AgentXMLRPCRequestHandler
                    , logRequests=False
                )

        #setup document
        server.set_server_title("OVSAgentServer Document")
        server.set_server_name("OVSAgentServer")

        #server start serving...
        logger.info("OVSAgentServer start serving!")
        try:
            server.serve_forever()
        except KeyboardInterrupt, e: #Ctrl+C
            errmsg = fmt_err_msg(e)
            logger.info("OVSAgentServer: server stopped!")
            logger.info("OVSAgentServer: server stopped!=>%s" % errmsg)
            server.server_close()
            sys.exit(0)
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("OVSAgentServer.start_server(): error. => %s" % errmsg)
        sys.exit(1)

#----------------------------------------------------------------------
def main(front):
    pid_filename = basename(__file__).split(".")[0]
    if not front:
        daemonize(pid_filename)
    init_signal_handler()
    init_log("client")

    #create the aliveness lock for the agent server
    from OVSConstants import AGENT_LIVENESS_LOCK
    if exists(AGENT_LIVENESS_LOCK):
        os.remove(AGENT_LIVENESS_LOCK)
    fd = os.open(AGENT_LIVENESS_LOCK, os.O_CREAT)
    try:
        from fcntl import flock, LOCK_EX, LOCK_NB
        flock(fd, LOCK_EX | LOCK_NB)
    except Exception, e:
        logging.error("OVSAgentServer.main(): cannot create lock for agent aliveness! => %s : %s" % (e.__class__, e))
        sys.exit(1)

    #remove shutdown mode flag
    if exists(AGENT_SHUTDOWN_MODE_FLAG):
        os.remove(AGENT_SHUTDOWN_MODE_FLAG)

    #overide some of the system default modules, using the compat version
    execfile(join(dirname(__file__), "OVSCompat.py"))

    #to set dom0 weight 512
    try:
        from OVSCommons import run_cmd
        run_cmd(args = ['xm', 'sched-credit', '-d', '0', '-w', '512'])
    except:
        pass

    start_server()

if __name__ == "__main__":
    front = False
    if len(sys.argv) > 1:
        if sys.argv[1] == 'front':
            front = True
    main(front)

