#!/usr/bin/python

#^#
# OVSCDS.py
#
# CDS -- Concurrent Data Store, a term borrowed from berkeley db.
# but in fact, in our case, the db access is serialized.
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import string
from time import sleep
from os.path import dirname, join, exists
from os import makedirs
from bsddb import dbshelve

from OVSTimer import Timer
from OVSFileLock import FileLock
from OVSException import *
from OVSConstants import CONFIG_DIR, CONFIG_DIR_CLUSTER

#----------------------------------------------------------------------
class CDSLockTimeout(OVSException):
    pass

class Timeout(Exception):
    pass

class CDS(object):
    #Concurrent Data Store
    #FIXME: here we use a simple file lock to serialize db access
    # make sure close the CDS as soon as possible
    # & avoid any deadlock, for e.g. donnot raise Exception before close()
    def __init__(self, *args, **kwargs):
        # timeout and interval support
        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
        else:
            self.timeout = 30
        if "interval" in kwargs:
            self.interval = kwargs["interval"]
        else:
            self.interval = 0.1

        #support db name list, e.g. "srv,vm"
        l = []
        for e in args:
            l.extend(map(string.strip, e.split(",")))
        db_names, flags = [], []
        for e in l:
            if e.isupper():
                flags.append(e)
            else:
                db_names.append(e)

        #set flags
        #lock db in share mode
        FLAG_LOCK_SHARE = 'LOCK_SH' in flags
        #access local db that is /etc/ovs-agent/db/*
        FLAG_LOCAL      = "LOCAL_DB" in flags

        self.db_names = list(set(db_names))
        #avoid deadlock
        self.db_names.sort()
        lock_base = join(CONFIG_DIR, "db")
        if not FLAG_LOCAL:
            db_base = join(CONFIG_DIR_CLUSTER, "db")
        else:
            db_base = join(CONFIG_DIR, "db")
        if not exists(db_base):
            makedirs(db_base)
        self.lock_timeout = False

        #lock db
        self.lock = {}
        lck_idx = 0
        for (idx, db_name) in enumerate(self.db_names):
            lock_file = join(lock_base, "%s.lock" % (db_name,) )

            # Create the lock file if it is absent.
            if not exists(lock_file):
                open(lock_file, 'w').close()

            self.lock[db_name] = FileLock(lock_file)
            timer = Timer(self.timeout, self.interval)
            while timer.wait():
                if FLAG_LOCK_SHARE:
                    ret = self.lock[db_name].acquire("share")
                else:
                    ret = self.lock[db_name].acquire()
                if ret:
                    break
            if timer.timeout():
                lck_idx = idx
                locker = 'unknown'
                fd = open(lock_file, 'r')
                try:
                    locker = fd.readline()
                finally:
                    fd.close()
                #clean up
                for idx in range(0, lck_idx + 1):
                    self.lock[self.db_names[idx]].release()

                raise CDSLockTimeout(ERR_CDS_LOCK_TIMOUT, {
                        "lock_file": lock_file, "locker": locker})

            db_file = join(db_base, "%s.db" % db_name)
            dbs = dbshelve.open(db_file)
            setattr(self, "db_%s" % db_name, dbs)

        self._opened = True #sentinel

    def close(self):
        if hasattr(self, "_opened") and self._opened: #so that call close() multitimes is ok in a process
            self._opened = False
            for db_name in self.db_names:
                try:
                    try:
                        getattr(self, "db_%s" % db_name).close()
                        delattr(self, "db_%s" % db_name)
                    finally:
                        self.lock[db_name].release()
                except:
                    pass

    def __del__(self):
        self.close()

#----------------------------------------------------------------------
def db_dump(db_name):
    d = {}
    cds = CDS(db_name)
    try:
        db_name = [e.strip() for e in db_name.split(",") if not e.isupper()][0]
        d.update(getattr(cds, "db_%s" % db_name))
    finally:
        cds.close()

    return d

#query
def db_load(db_name, key, get=False):
    value = None
    cds = CDS(db_name)
    try:
        #filter flags
        db_name = [e.strip() for e in db_name.split(",") if not e.isupper()][0]
        db = getattr(cds, "db_%s" % db_name)
        if get:
            value = db.get(key)
        else:
            value = db[key]
    finally:
        cds.close()

    return value

#insert
def db_save(db_name, key, value):
    cds = CDS(db_name)
    try:
        #filter flags
        db_name = [e.strip() for e in db_name.split(",") if not e.isupper()][0]
        db = getattr(cds, "db_%s" % db_name)
        db[key] = value
    finally:
        cds.close()

#delete
def db_remove(db_name, key):
    cds = CDS(db_name)
    try:
        db_name = [e.strip() for e in db_name.split(",") if not e.isupper()][0]
        db = getattr(cds, "db_%s" % db_name)
        if key in db:
            del db[key]
    finally:
        cds.close()

#update
def db_update(db_name, key, value):
    if not isinstance(value, dict):
        raise Exception("db_update: value must be of dict type. value=%s" % repr(value))
    cds = CDS(db_name)
    try:
        #filter flags
        db_name = [e.strip() for e in db_name.split(",") if not e.isupper()][0]
        db = getattr(cds, "db_%s" % db_name)
        if key in db:
            if not isinstance(db[key], dict):
                raise Exception("db_update: value must be of dict type. db[%s]=%s " % (
                        key, repr(db[key])))
            _value = db[key]
            _value.update(value)
            db[key] = _value
        else:
            db[key] = value
    finally:
        cds.close()

