#!/usr/bin/env python

#^#
# OVSCommons.py
#
# common function used in agent.
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import zlib
import sys
import string
import re
import random
import popen2
import os
import time, signal
from subprocess import Popen, PIPE
from os.path import abspath, exists, isfile, isdir, join, basename, dirname, islink
from OVSConfig import get_config

#compatible import
from OVSServerProxy import *
from OVSProtocol import *
from OVSParser import *
from OVSMultiprocess import *
from OVSExceptionLookup import *
from OVSException import *
from OVSConstants import *
from OVSCDS import *

#----------------------------------------------------------------------
def marshall_string(s):
    """
    compress with zlib, then encode base64
    """
    return zlib.compress(s).encode("base64").replace("\n", "")

def unmarshall_string(s):
    """
    decode base64, then decompress with zlib
    """
    return zlib.decompress(s.decode("base64"))

#----------------------------------------------------------------------

def timeout_command(command, timeout=10):
    """command with timeout (seconds)"""
    start = time.time()
    p = popen2.Popen3(command, True)
    while p.poll() == -1:
        time.sleep(0.5)
        now = time.time()
        if now - start > timeout:
            os.kill(p.pid, signal.SIGTERM)
            os.waitpid(-1, os.WNOHANG)
            raise Exception('timeout_command() timed out!')
    if p.wait() != 0:
        raise Exception('%s => %s' % (command, p.childerr.read()))
    return p.fromchild.read()

def run_cmd(args):
    """
    run a cmd and catch its output, if run into any problem raise exception
    """
    cmdlist = args
    if not isinstance(cmdlist, list):
        raise Exception("run_cmd() only accepts list!")
    p = popen2.Popen3(cmdlist, True)
    if p.wait() != 0:
        raise Exception('%s => %s' % (cmdlist, p.childerr.read()))
    return p.fromchild.read()

def old_run_cmd(cmdstr):
    """
    *** This method is deprecated!!! ***
    run a cmd and catch its output, if run into any problem raise exception
    """
    #possible cmd injection, like sql injection
    if not isinstance(cmdstr, str):
        raise Exception("old_run_cmd() only accepts str!")
    p = popen2.Popen3(cmdstr, True)
    if p.wait() != 0:
        raise Exception('%s => %s' % (cmdstr, p.childerr.read()))
    return p.fromchild.read()

def run_cmd_bg(cmd):
    #run cmd in background without wait.
    #NOTE: even implemented as a daemon, redirect 0,1,2 is still necessary or
    #you will not get a clean ssh disconnect
    devnull = open(os.devnull, "rw")
    p = Popen(cmd.split(), stdin=devnull, stdout=devnull, stderr=devnull, close_fds=True)

def setprocname(name):
    #http://nkour.wordpress.com/2006/08/30/change-process-name-in-python-posix-only-using-dl-module/
    import dl
    libc = dl.open("/lib/libc.so.6")
    libc.call('prctl', 15, name, 0, 0, 0)

#------------------------------------------------------------------------------
def well_behavior(decorator):
    """This decorator can be used to turn simple functions
    into well-behaved decorators, so long as the decorators
    are fairly simple. If a decorator expects a function and
    returns a function (no descriptors), and if it doesn't
    modify function attributes or docstring, then it is
    eligible to use this. Simply apply @well_behavior to
    your decorator and it will automatically preserve the
    docstring and function attributes of functions to which
    it is applied."""
    def new_decorator(f):
        g = decorator(f)
        g.__name__ = f.__name__
        g.__doc__ = f.__doc__
        g.__dict__.update(f.__dict__)
        return g
    # Now a few lines needed to make well_behavior itself
    # be a well-behaved decorator.
    new_decorator.__name__ = decorator.__name__
    new_decorator.__doc__ = decorator.__doc__
    new_decorator.__dict__.update(decorator.__dict__)
    return new_decorator

#@well_behavior
#def exposed(func):
#    #---------------------------------------
#    def wrapper(*args, **kwargs):
#        return func(*args, **kwargs)
#    #---------------------------------------
#    wrapper.exposed = True
#    wrapper.orig_func = func
#    return wrapper
def exposed(func):
    func.exposed = True
    return func

#@well_behavior
#def interconnect(func):
#    #---------------------------------------
#    def wrapper(*args, **kwargs):
#        return func(*args, **kwargs)
#    #---------------------------------------
#    wrapper.interconnect = True
#    wrapper.orig_func = func
#    return wrapper
#
def interconnect(func):
    func.interconnect = True
    return func

#------------------------------------------------------------------------------
def randomMAC():
    """Generate a random MAC address.

    Uses OUI (Organizationally Unique Identifier) 00-16-3E, allocated to
    Xensource, Inc. The OUI list is available at
    http://standards.ieee.org/regauth/oui/oui.txt.

    The remaining 3 fields are random, with the first bit of the first
    random field set 0.

    @return: MAC address string
    """
    mac = [ 0x00, 0x16, 0x3e,
            random.randint(0x00, 0x7f),
            random.randint(0x00, 0xff),
            random.randint(0x00, 0xff) ]
    return ':'.join(map(lambda x: "%02x" % x, mac))

def get_uuid():
    from xen.xend import uuid
    uuid.uuidFactory = uuid.getUuidUuidgen

    return uuid.createString()

def normal_uuid(uuid):
    return uuid.strip().replace("-", "").upper()

def format_uuid(uuid):
    _uuid = uuid.strip().replace("-", "").lower()
    return "-".join([_uuid[0:8], _uuid[8:12], _uuid[12:16], _uuid[16:20], _uuid[20:]])

def get_hash(s):
    import md5
    h = md5.new()
    h.update(s)
    return h.hexdigest()
#------------------------------------------------------------------------------
def host_alive(ip):
    """Determine whether the given host is alive."""
    status = os.popen("ping -q -c 2 %s" % ip).close()
    return not status

def get_host_by_name(hostname):
    ptn_ip = re.compile(r"\d+\.\d+\.\d+\.\d+")
    if ptn_ip.match(hostname):
        return hostname

    ip = ""
    try:
        import socket
        #NOTE: if return value is not constant, we treat it as configuration error.
        ip = socket.gethostbyname(hostname)
        if ip.startswith("127") and not hostname.startswith("localhost"):
            stdout = run_cmd("host %s" % hostname)
            l = [ line.split()[-1] for line in stdout.splitlines()]
            l = [_ip for _ip in l if ptn_ip.match(_ip)]
            l.sort()
            if l:
                ip = l[0]
    except:
        pass
    return ip

def get_my_ips():
    ptn = re.compile(r"inet addr:([.0-9]+)")
    lines = run_cmd(["/sbin/ifconfig"])
    ll = [] #list local
    for line in lines.splitlines():
        m = ptn.search(line)
        if m:
            ip = m.group(1)
            ll.append(ip)
    return ll

#get my cluster ip -- the ip alive and in /etc/ocfs2/cluster.conf
def get_my_member_ip():
    ll = get_my_ips()
    d = parse_cluster_conf()
    lc = [node["ip_address"] for node in d["nodes"]]
    l = list(set(ll).intersection(set(lc)))
    if len(l) > 0:
        return l[0]

#------------------------------------------------------------------------------
def chk_vm_path(vm_path):
    """Check path agaist vm operation

    NOTE: Deprecated!!
    """
    return True


def chk_vm_name(vm_name):
    if "'" in vm_name:
        raise Exception("single quotation mark is not allowed in VM name!")


def check_shutdown_mode():
    from OVSConstants import AGENT_SHUTDOWN_MODE_FLAG
    if exists(AGENT_SHUTDOWN_MODE_FLAG):
        raise Exception("operation cannot be invoked when agent in shutdown mode!")

def sync_file_attr(path):
    """
    NOTE: Deprecated!!
    """
    #[bugz 6759], sync file attribute by lock/unlock
    if not isfile(path):
        return
    try:
        from fcntl import lockf, LOCK_EX, LOCK_UN
        fd = os.open(path, os.O_RDWR)
        lockf(fd, LOCK_EX)
        lockf(fd, LOCK_UN)
        os.close(fd)
    except:
        pass

#------------------------------------------------------------------------------
def strings_common_head(l):
    """
    @l: list of strings
    @return:
    """
    r = []
    for idx, c in enumerate(l[0]):
        same = True
        for s in l[1:]:
            if s[idx] != c:
                same = False
        if not same:
            break
        r.append(c)

    return "".join(r)

#------------------------------------------------------------------------------
from datetime import datetime
import shutil
def ovs_backup(path, mode="move", symlinks=False):
    if not exists(path):
        raise OVSException(ERR_INTERNAL_ERROR,
                argd={"errmsg": "Cannot backup path '%s', not exists." % path})

    # umount the path if necessary
    try:
        run_cmd(["umount", "-fl", path])
    except:
        pass

    path = path.rstrip("\\/")
    backup_path = "%s.%s.orig" % (
            path, datetime.strftime(datetime.now(), "%Y%m%d-%H%M%S"))
    if mode == "move":
        shutil.move(path, backup_path)
    else: #copy
        if symlinks and islink(path):
            linkto = os.readlink(path)
            os.symlink(linkto, backup_path)
        else:
            if isfile(path):
                shutil.copyfile(path, backup_path)
            else:
                if not exists(backup_path):
                    os.makedirs(backup_path)
                shutil.copytree(path, backup_path, symlinks=symlinks)
    return backup_path

