#!/usr/bin/env python
#^#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import string
import re
import os
import logging
from time import sleep
from os.path import exists, basename
from os import O_CREAT, O_RDWR, O_NONBLOCK
import threading

from OVSXMasterFailover import i_am_the_master
from OVSXCluster import clusterm_status_heartbeat
from OVSSiteRMVM import get_live_vm_info
from OVSProtocol import *
from OVSLog import init_log
from OVSDaemon import daemonize
from OVSDB import db_update_srv, db_get_srv
from OVSConstants import *
from OVSConfig import init_config, get_config
from OVSCommons import *
from OVSPolicyServer import execute_policy
from OVSServerProxy import ServerProxy

def check_server_status():
    logger = logging.getLogger()

    trigger_recovery = False
    #check ocfs2 cluster status
    try:
        rs = clusterm_status_heartbeat(keep_silence=True)
        status = GETVAL(rs)
        if not status == HEARTBEAT_STATUS_ACTIVE:
            return trigger_recovery
    except:
        return trigger_recovery

    #check debugfs status
    livenodes_file = "/sys/kernel/debug/o2hb/livenodes"
    try:
        if not exists(livenodes_file):
            #cmd = "mount -t debugfs debugfs /sys/kernel/debug"
            cmdlist = ["mount", "-t", "debugfs",  "debugfs", "/sys/kernel/debug"]
            run_cmd(args=cmdlist)
            if not exists(livenodes_file):
                raise Exception("debugfs cannot be mounted!")
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("check_server_status: error. => %s" % errmsg)
        return trigger_recovery

    #main process
    try:
        #live nodes
        d = parse_cluster_conf()
        members = dict([(node["ip_address"], node["number"]) for node in d["nodes"]])
        f = open(livenodes_file)
        livenodes = f.read().split()
        f.close()
        live_nodes_ip = [ ip for ip, num in members.items() if num in livenodes ]

        #all nodes
        d = db_dump("srv")
        all_nodes_ip = [ v["ip"] for k, v in d.items() ]
        my_ip = get_my_member_ip()
        if my_ip in live_nodes_ip:
            all_nodes_ip.remove(my_ip)

        for ip in all_nodes_ip:
            try:
                try:
                    srv = db_get_srv(ip)
                except:
                    #member in ocfs2 cluster but not in the same pool
                    continue

                if ip not in live_nodes_ip:
                    if "real_status" not in srv or srv["real_status"] != "DEAD":
                        db_update_srv(ip, {"real_status": "DEAD", "status": "UNKNOWN"})
                        logger.debug("check_server_status: srv('%s') goes DEAD!" % ip)
                        trigger_recovery = True
                else:
                    if "real_status" not in srv or srv["real_status"] != "ALIVE":
                        db_update_srv(ip, {"real_status":"ALIVE"})
                        logger.debug("check_server_status: srv('%s') comes to ALIVE!" % ip)
                        trigger_recovery = True

                        #check agent connection
                        try:
                            sp = ServerProxy(srv["agt_url"])
                            rs = sp.echo("whatever")
                            if rs.startswith("success"):
                                db_update_srv(ip, {"status": "RUNNING"})
                        except Exception, e:
                            errmsg = fmt_err_msg(e)
                            logger.error("check_server_status: ip('%s') check agent connection error. => %s" % (ip, errmsg))

            except Exception, e:
                errmsg = fmt_err_msg(e)
                logger.error("check_server_status: ip('%s') error. => %s" % (ip, errmsg))
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("check_server_status: error. =>%s" % errmsg)

    return trigger_recovery


class Worker(threading.Thread):
    def run(self):
        logger = logging.getLogger()
        logger.info("Policy worker thread starts.")
        execute_policy()
        logger.info("Policy worker thread ends.")


def recover_process():
    #recover process according to policy
    Worker().run()


def run():
    logger = logging.getLogger()
    cnt = 0
    while True:
        try:
            init_config() #let the config modification take effect immediate
            config = get_config()

            srv_monitor_interval, vm_status_sync_interval, collect_vm_status = 10, 60, False
            try:
                srv_monitor_interval = int(config.get("ha", "server.monitor.interval"))
                vm_status_sync_interval = int(config.get("ha", "policy.execute.interval"))
            except Exception, e:
                errmsg = fmt_err_msg(e)
                logger.error("OVSMonitorServer.run(): error.=>%s" % errmsg)

            cnt_collect_vm_status = int(vm_status_sync_interval / srv_monitor_interval)
            cnt += 1
            if cnt > cnt_collect_vm_status:
                cnt = 0

            if i_am_the_master():
                if vm_status_sync_interval < srv_monitor_interval \
                        or cnt == cnt_collect_vm_status:
                    #cannot collect vm status more frequently than collect server status
                    collect_vm_status = True

                trigger_recovery = False
                pool_ha_enable = db_load("cluster", "pool_ha_enable", get=True)
                if pool_ha_enable:
                    trigger_recovery = check_server_status()

                if trigger_recovery or collect_vm_status:
                    get_live_vm_info()
                if trigger_recovery and config.get("ha", "policy.execute").startswith("enable"):
                    recover_process()

        except Exception, e:
            errmsg = fmt_err_msg(e)
            logger.error("OVSMonitorServer.run(): error. => %s" % errmsg)

        #sleep and next cycle ...
        try:
            sleep(srv_monitor_interval)
        except Exception, e:
            errmsg = fmt_err_msg(e)
            logger.error("OVSMonitorServer.run(): sleep error. => %s" % errmsg)

def main():
    pid_filename = basename(__file__).split(".")[0]
    daemonize(pid_filename)

    init_log("client")
    run()

if __name__ == "__main__":
    main()
