#!/usr/bin/env python

import xmlrpclib
import socket
try:
    from multiprocessing import Process, Manager
except ImportError:
    from processing import Process, Manager

from OVSServerProxy import *
from OVSProtocol import *
from OVSExceptionLookup import *
from OVSException import *
from OVSCDS import *

def servers_do(operation, param, select_type="db", server_list=None, timeout=None, fork_max=32, ping_agent_first=False):
    """Do operation on multiple servers with multiprocess.

    @param select_type, server_list:
        select_type in ["all", "include", "exclude", "nondb"]
            db  : do operation on servers that in db.
                  server_list is not used.
            db+l: do operation on servers that in db and in server_list.
                  server_list := [<(hostname, agent_passwd)>]
            db-l: do operation on servers that in db  but not in server_list.
                  server_list := [<hostname>]
            db&l: do operation on servers that in db and also in server_list.
                  server_list := [<hostname>]
            l   : do operation on servers in server_list only.
                      server_list := [<(hostname, agent_passwd)>]
    @param timeout:
        timeout in seconds. default to None, which means no timeout.
    @param fork_max:
        max number of concurrent processes.
    @param ping_agent_first:
        if True, check whether the agent is alive on the server before
        let the server do the operation and if agent is not alive,
        ignore this server.
        use case: when the operation's time consumption might vary in a large range,
        it is not easy to set an acurate timeout. you may want to filter the timeout
        caused by agent down first.

    @return: [<(hostname, rs)>], rs of "agent data interchange protocol 2v"
    """
    #------------------------------------------------------------------
    def process(result_dict, job_id, sp, operation, param, timeout=None):
        """
        @param result_dict: update this dict when finish processing.
        @param job_id: hostname is used as job_id
        @param sp, operation, param: a job description and will be executed as
                getattr(sp, operation)(*param)
        @param timeout:
                process the job within timeout.
        """
        orig_timeout = socket.getdefaulttimeout()
        socket.setdefaulttimeout(timeout)
        try:
            try:
                rs = getattr(sp, operation)(*param)
                rs = JSON(rs)
            except socket.timeout:
                raise OVSException(ERR_AGENT_CONNECTION)
            except socket.error, e:
                (code, errmsg) = e
                if code == 104 or code == 111:
                    raise OVSException(ERR_AGENT_CONNECTION)
                elif code == 113 or code == -5:
                    raise OVSException(ERR_SERVER_CONNECTION)
                else:
                    raise e
            except xmlrpclib.ProtocolError:
                raise OVSException(ERR_AGENT_PASSWD_INCORRECT)
        except Exception, e:
            rs = FAIL(exception=e)

        socket.setdefaulttimeout(orig_timeout)

        result_dict[job_id] = rs

    #------------------------------------------------------------------
    def multiprocess(jobs):
        """
        @param jobs: [<(result_dict, job_id, sp, operation, param, timeout)>]
        """
        rs_dict = manager.dict()
        while jobs:
            count = 0
            p_list = []
            while count < fork_max and jobs:
                job = jobs.pop()
                p = Process(target=process, args=job)
                p.start()
                p_list.append(p)
                count += 1
            for p in p_list:
                p.join()
    #------------------------------------------------------------------

    #get proxy for each server
    sp_dict = {} #{<hostname:ServerProxy>}
    if "db" in select_type:
        servers = {}
        cds = CDS("srv")
        try:
            servers.update(cds.db_srv)
        finally:
            cds.close()
        for hostname, srv in servers.items():
            sp_dict[hostname] = ServerProxy(srv["agt_url"])

    if select_type == "db-l":
        if server_list:
            for hostname in server_list:
                if hostname in sp_dict:
                    del sp_dict[hostname]

    if select_type == "db+l":
        if server_list:
            _sp_dict = get_server_proxy_from_list(server_list)
            sp_dict.update(_sp_dict)

    if select_type == "db&l":
        sp_dict = dict([(hostname, sp) for (hostname, sp) in sp_dict.items() if hostname in server_list])

    if select_type == "l":
        if server_list:
            sp_dict = get_server_proxy_from_list(server_list)

    sp_list = sp_dict.items()

    #server process for sharing state between processes.
    manager = Manager()

    #ping agent fist
    bad_agent_result_list = []
    if ping_agent_first:
        result_dict = manager.dict()
        jobs = [(result_dict, hostname, sp, "echo", (hostname,), 2) for hostname, sp in sp_list]
        multiprocess(jobs)

        for hostname, rs in result_dict.items():
            try:
                GETVAL(rs)
            except Exception, e:
                bad_agent_result_list.append((hostname, DJSON(FAIL(exception=e))))
                del sp_dict[hostname]

    #process the real jobs
    sp_list = sp_dict.items()
    result_dict = manager.dict()
    jobs = [(result_dict, hostname, sp, operation, param, timeout) for hostname, sp in sp_list]
    multiprocess(jobs)

    #deal with the result
    #l  := [<(hostname, rs)>]
    #rs := {<key:value>}
    #key:= "status" | "value" | "message"
    l = []
    for hostname, rs in result_dict.items():
        l.append((hostname, DJSON(rs)))
    if ping_agent_first:
        l.extend(bad_agent_result_list)

    return l

def servers_do_wrapper(operation, param, logger=None, **argd):
    warning = False
    if "warning" in argd:
        warning = argd["warning"]
        del argd["warning"]

    rs_list = servers_do(operation, param, **argd)
    err_list = []
    for hostname, rs in rs_list:
        if rs["status"] == OP_STATUS_FAIL:
            err_list.append((hostname, rs))
            if logger:
                logger.error("%s: srv('%s') => %s" % (
                        operation, hostname, JSON(rs)))
    if not warning and err_list:
        raise OVSException(ERR_DO_ON_SERVERS, argd={
                "operation": operation,
                "hostnames": ",".join([hostname for hostname, rs in err_list])})

    return err_list

def get_server_list_with_conn_err(logger=None):
    server_list = []
    operation, param = "echo", ("whatever",)
    err_srv_rs = servers_do_wrapper(operation, param, logger, warning=True, ping_agent_first=True)
    for hostname, rs in err_srv_rs:
        server_list.append(hostname)
    return server_list

def get_server_list_with_local_conf():
    server_list = []
    operation = "get_agent_conf_mode"
    param = ()
    rs_list =  servers_do(operation, param)
    for hostname, rs in rs_list:
        if rs["status"] == OP_STATUS_SUCC:
            if GETVAL(rs)["mode"].encode("utf-8") == "local":
                server_list.append(hostname)

    return server_list
