#!/usr/bin/env python

try:
    import json
except ImportError:
    import simplejson as json

from OVSException import *
from OVSExceptionLookup import *

#agent communication protocol v1
def parse_result(result):
    """
    all return result following same convension, (our own marshal mechanism):
    success                         #if success completed without return value
    success:key=value;key=value...  #if success completed with return result
    failed:errmsg                   #failed with error message
    """
    if result.startswith("failed"): return

    #e.g. result="success:key=value;key2=value2..."
    d = {}
    result = result[8:].split(";")
    if len(result) == 1 and result[0].strip() == "":
        return d
    for kv in result:
        k,v = kv.split("=")
        d[k] = v

    return d

#agent communication protocol v2
OP_STATUS_SUCC = "SUCC"
OP_STATUS_FAIL = "FAIL"
OP_STATUS_WARN = "WARN"

def SUCC(value=None):
    """
    @param value: value to be returned.

    @return {
        "status"    : OP_STATUS_SUCC,
        "value"     : value,
    } serialized with json.dumps()
    """

    rs = {
        "status"    : OP_STATUS_SUCC,
    }

    if value:
        rs["value"] = value

    return json.dumps(rs)

def FAIL(errcode=ERR_RAW_MESSAGE, argd=None, exception=None, error=None):
    """
    @param errcode: code defined in OVSExceptionLookup
    @param argd: parameter dict to format error message template
    @param exception: if not None, error

    @return {
        "status"    : OP_STATUS_WARN,
        "value"     : value,
        "error"     : {"errcode":errcode, "errmsg":errmsg},
    } serialized with json.dumps()
    """
    if isinstance(exception, Exception):
        if isinstance(exception, OVSException):
            _error = exception.value
        else:
            _error = {
                "errcode"    : ERR_INTERNAL_ERROR,
                "errmsg"     : "%s: %s" % (exception.__class__.__name__, exception),
            }
    elif error:
        _error = error
    else:
        _error = OVSException(errcode, argd).value

    rs = {
        "status"    : OP_STATUS_FAIL,
        "value"     : None,
        "error"     : _error,
    }
    return json.dumps(rs)

def WARN(value=None, errcode=ERR_RAW_MESSAGE, argd=None, exception=None, error=None):
    """
    @param errcode: code defined in OVSExceptionLookup
    @param argd: parameter dict to format error message template
    @param exception: if not None, error

    @return {
        "status"    : OP_STATUS_WARN,
        "value"     : value,
        "error"     : {"errcode":errcode, "errmsg":errmsg},
    } serialized with json.dumps()
    """
    if exception:
        if isinstance(exception, OVSException):
            _error = exception.value
        else:
            errmsg = "<%s: %s>" % (exception.__class__.__name__, exception)
            _error = OVSException(ERR_RAW_MESSAGE, {"errmsg": errmsg}).value
    elif error:
        _error = error
    else:
        _error = OVSException(errcode, argd).value


    rs = {
        "status"    : OP_STATUS_WARN,
        "value"     : value,
        "error"     : _error,
    }
    return json.dumps(rs)

#compat with protocol v1
def JSON(rs):
    if not isinstance(rs, str):
        rs = json.dumps(rs)

    elif rs.startswith("success"):
        try:
            value = parse_result(rs)
        except:
            #e.g. ha_generate_ocfs2_cluster_conf
            value = rs[8:]
        rs = {
            "status"    : OP_STATUS_SUCC,
            "value"     : value
        }
        rs = json.dumps(rs)

    elif rs.startswith("failed"):
        rs = {
            "status"    : OP_STATUS_FAIL,
            "value"     : None,
            "error"     : OVSException(ERR_RAW_MESSAGE, {"errmsg": rs[7:]}).value
        }
        rs = json.dumps(rs)

    return rs

def DJSON(rs):
    return json.loads(rs)

def GETVAL(rs):
    rs = DJSON(JSON(rs))
    #print rs
    if rs["status"] in (OP_STATUS_SUCC, OP_STATUS_WARN):
        if "value" in rs:
            return rs["value"]
        else:
            return None
    else:
        raise OVSException(error=rs["error"])

