#!/usr/bin/env python
#^#
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import pickle
import os
import logging
from os.path import isfile, islink, isdir, exists, dirname, join, basename
from glob import glob

from OVSServerProxy import ServerProxy
from OVSCommons import *
from OVSCDS import *
from OVSMultiprocess import *
from OVSConstants import *
from OVSXMasterFailover import clusterm_cleanup_master_env
from OVSXCluster import clusterm_teardown
from OVSSiteHA import *

@exposed
def cluster_get_master():
    """
    Get the hostname of cluster master.
    @return "success:master_hostname=<master_hostname>" if success else "failed:<error message>"
    """
    try:
        master = db_load("cluster", "master_hostname", True)
        if not master: master=""
        return "success:master_hostname=%s" % master
    except Exception, e:
        errmsg = fmt_err_msg(e)
        return "failed:%s" % errmsg


@exposed
def cluster_get_info():
    logger = logging.getLogger("ovs.query")
    try:
        d = {}
        cds = CDS("cluster")
        try:
            d.update(cds.db_cluster)
        finally:
            cds.close()

        rs = SUCC(d)
        logger.info("cluster_get_info: success. => %s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("cluster_get_info: error. => %s" % errmsg)
        return FAIL(exception=e)


@exposed
def cluster_sync_agent_config(src_sr_uuid, dst_sr_uuid, config_sub_path="db,passwdfile"):
    """
    @param src_sr_uuid: src base dir is $src_sr.mountpoint/.ovs-agent
        if not sr_src_uuid: /etc/ovs-agent
    @param dst_sr_uuid: dst base dir is $dst_sr.mountpoint/.ovs-agent
    @param config_sub_path: subpath in config dir, e.g. ["db", "cert", "passwdfile"]
    """
    logger = logging.getLogger("ovs.operation")
    try:
        if not src_sr_uuid:
            src_base_dir = CONFIG_DIR
        else:
            sr = db_load("sr", src_sr_uuid)
            sr.initialize()
            src_base_dir = join(sr.mountpoint, CONFIG_DIR_CLUSTER_BASENAME)

        if not src_sr_uuid:
            sr = db_load("LOCAL_DB,sr", dst_sr_uuid)
        else:
            sr = db_load("sr", dst_sr_uuid)
        sr.initialize()
        dst_base_dir = join(sr.mountpoint, CONFIG_DIR_CLUSTER_BASENAME)

        if islink(dst_base_dir):
            os.remove(dst_base_dir)

        l = []
        for _path in config_sub_path.split(","):
            src = join(src_base_dir, _path)
            dst = join(dst_base_dir, _path)
            if not exists(src):
                raise OVSException(ERR_SYNC_AGENT_CONFIG, argd={"src_path": src, "dst_path": dst})

            if isdir(src):
                src = "%s/" % src.rstrip("/\\")
                dst = "%s/" % dst.rstrip("/\\")
                if not exists(dst):
                    os.makedirs(dst)

            else:
                src = src.rstrip("/\\")
                dst = dst.rstrip("/\\")
                if not exists(dirname(dst)):
                    os.makedirs(dirname(dst))

            cmd = ['rsync', '-a', src, dst]
            l.append(cmd)

        for cmd in l:
            run_cmd(cmd)

        rs = SUCC()
        logger.info("cluster_sync_agent_config: src_sr='%s', dst_sr='%s', config_sub_path='%s'" % (src_sr_uuid, dst_sr_uuid, config_sub_path))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("cluster_sync_agent_config: => %s" % errmsg)
        return FAIL(exception=e)


from OVSXCluster import clusterm_init_root_sr
@exposed
def cluster_init_root_sr(sr_uuid, migrate_data=False, construct_from_local=False):
    """
    @param construct_from_local:
            if True:
                sync db from local dir. useful when
                1) create a new pool.
                2) previous cluster root is destroyed.
            else:
                migrate current cluster root to new sr.

    NOTE:
        1) first setup a cluster.
            sr_uuid, migrate_data=True, construct_from_local=True
        2) reassign cluster root to a new sr.
            sr_uuid, migrate_data=True, construct_from_local=False
        3) normal sync from master node.
            sr_uuid, migrate_data=False, construct_from_local=False
    """
    logger = logging.getLogger("ovs.operation")
    try:
        #1) migrate data
        orig_root_sr_uuid = ""
        if migrate_data:
            #find out orig_root_sr_uuid
            if not construct_from_local:
                rs = cluster_get_info()
                rs = GETVAL(rs)
                orig_root_sr_uuid = rs.get("cluster_root_sr_uuid")
                if not orig_root_sr_uuid:
                    raise OVSException(ERR_NO_CLUSTER_ROOT_SR)
                orig_root_sr_uuid = orig_root_sr_uuid.encode("utf-8")
                if orig_root_sr_uuid == sr_uuid:
                    migrate_data = False

        if migrate_data:
            rs = cluster_sync_agent_config(orig_root_sr_uuid, sr_uuid)
            rs = DJSON(rs)
            if rs["status"] == OP_STATUS_FAIL:
                raise OVSException(error=rs["error"])

        #2) mount root sr and sync to members
        if construct_from_local:
            sr = db_load("LOCAL_DB,sr", sr_uuid)
        else:
            sr = db_load("sr", sr_uuid)
        root_sro = pickle.dumps(sr)

        rs = clusterm_init_root_sr(root_sro)
        rs = DJSON(rs)
        if rs["status"] == OP_STATUS_FAIL:
            raise OVSException(error=rs["error"])

        exclude = get_server_list_with_conn_err()
        hostname = db_load("LOCAL_DB,node", "registered_hostname", get=True)
        if hostname:
            hostname = hostname.encode("utf-8")
            if not hostname in exclude: exclude.append(hostname)

        operation, param = "clusterm_init_root_sr", (root_sro,)
        servers_do_wrapper(operation, param, logger, select_type="db-l", server_list=exclude)

        #3) save config
        if migrate_data:
            db_save("cluster", "cluster_root_sr_uuid", sr_uuid)
            db_save("cluster", "cluster_root_sr_uuid_orig", orig_root_sr_uuid)

        rs = SUCC({"cluster_root_sr_uuid": sr_uuid})
        logger.info("cluster_init_root_sr: sr('%s') -> sr('%s') => %s" % (
                orig_root_sr_uuid, sr_uuid, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("cluster_init_root_sr: try to move cluster root to sr('%s') construct_from_local=%s => %s" % (
                sr_uuid, repr(construct_from_local), errmsg))
        return FAIL(exception=e)


@interconnect
def get_hostname_ip(public_hostname):
    hostname = socket.gethostname()
    ip = get_host_by_name(public_hostname)
    return (hostname, ip)


@exposed
def cluster_generate_ocfs2_cluster_conf(single_node=False):
    """
    @return: <content of generated '%s' generated according to servers in db.>
    """ % OCFS2_CLUSTER_CONF
    #------------------------------------------
    def _read_value(filename):
        f = open(filename, "r")
        r = f.read().strip()
        f.close()
        return r

    def next_slot(num_in_cluster):
        """Find next available slot for ocfs2 node

        @num_in_cluster: num slot already used by the cluster.
        """
        slot = 0
        while slot in num_in_cluster:
            slot += 1
        num_in_cluster.append(slot)

        return slot

    def _get_host_sp():
        l = []
        exclude = get_server_list_with_conn_err()
        cds = CDS("srv")
        try:
            l = [ (hostname, ServerProxy(srv["agt_url"])) for (hostname, srv) in cds.db_srv.items() if hostname not in exclude]
        finally:
            cds.close()
        return l

    #------------------------------------------
    logger = logging.getLogger("ovs.operation")
    try:
        #global scope var
        #cluster_name   : <cluster_name>
        #node_dict      : {<ip>:<node dict>}

        #1. consider setting in config file
        cluster_config_file = OCFS2_CLUSTER_CONF
        if exists(cluster_config_file):
            conf_dict = parse_ocfs2_cluster_conf(cluster_config_file)
            node_dict = dict([ (node["ip_address"], node) for node in conf_dict["nodes"] ])
            if conf_dict["cluster"].get("name"):
                cluster_name = conf_dict["cluster"]["name"]
            else:
                cluster_name = "ocfs2" #default
        else:
            conf_dict = {"node_dict":[], "cluster":{}}
            cluster_name = "ocfs2" #default
            node_dict = {}

        #2. consider runtime setting that already in configfs
        ptn = "/sys/kernel/config/cluster/%s/node/*" % cluster_name
        l = glob(ptn)
        for node_dir in l:
            node = {}
            node["ip_address"]  = _read_value(join(node_dir, "ipv4_address"))
            node["ip_port"]     = _read_value(join(node_dir, "ipv4_port"))
            node["number"]      = _read_value(join(node_dir, "num"))
            node["name"]        = basename(node_dir)
            node["cluster"]     = cluster_name
            node["in"] = True
            node_dict[node["ip_address"]] = node

        #num of node_dict in the cluster
        num_in_cluster  = [ int(node["number"]) for ip, node in node_dict.items()]

        # generate hostname list that should be include in cluster.conf
        l = []
        host_sp = _get_host_sp()
        for host, sp in host_sp:
            hostname_ip = sp.get_hostname_ip(host)
            l.append(hostname_ip)

        # single node cluster
        if single_node and not l:
            my_hostname = socket.gethostname()
            my_ip = get_host_by_name(my_hostname)
            l.append((my_hostname, my_ip))
        else:
            # move myself to the first entry
            my_hostname = socket.gethostname()
            for hostname_ip in l:
                if hostname_ip[0] == my_hostname:
                    l.remove(hostname_ip)
                    l.insert(0, hostname_ip)
                    break

        for hostname,ip in l:
            if not ip in node_dict:
                _d = {"in": True}
                _d["ip_port"]       = "7777"
                _d["ip_address"]    = ip
                _d["number"]        = next_slot(num_in_cluster)
                _d["name"]          = hostname
                _d["cluster"]       = cluster_name
                node_dict[ip] = _d
            node_dict[ip]["in"] = True

        #generate cluster.conf based on node_dict and cluster_name
        #sort by 'number'
        #filter server not in the pool.
        l = [ (int(node["number"]), node) for node in node_dict.values()
                    if node.get("in") ]
        l.sort()
        lines = []

        for (_number, node) in l:
            lines.append("node:\n")
            lines.append("\tip_port     = %s\n" % node["ip_port"])
            lines.append("\tip_address  = %s\n" % node["ip_address"])
            lines.append("\tnumber      = %s\n" % str(node["number"]))
            lines.append("\tname        = %s\n" % node["name"])
            lines.append("\tcluster     = %s\n" % node["cluster"])
            lines.append("\n")
        lines.append("cluster:\n")
        lines.append("\tnode_count  = %d\n" % len(l))
        lines.append("\tname        = %s\n" % cluster_name)
        lines.append("\n")
        cluster_conf = "".join(lines)

        rs = SUCC(cluster_conf)
        logger.info("cluster_generate_ocfs2_cluster_conf: =>\n%s" % cluster_conf)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("cluster_generate_ocfs2_cluster_conf: error. => %s" % errmsg)
        return FAIL(exception=e)


from OVSXCluster import clusterm_set_ocfs2_cluster_conf

@exposed
def cluster_prepare_ocfs2_conf(single_node=False):
    logger = logging.getLogger("ovs.operation")
    try:
        #generate OCFS2_CLUSTER_CONF
        rs = cluster_generate_ocfs2_cluster_conf(single_node)
        cluster_conf = GETVAL(rs)

        #update OCFS2_CLUSTER_CONF
        clusterm_set_ocfs2_cluster_conf(cluster_conf)

        exclude = get_server_list_with_conn_err()
        #nodes update OCFS2_CLUSTER_CONF
        operation = "clusterm_set_ocfs2_cluster_conf"
        param = (cluster_conf,)
        servers_do_wrapper(operation, param, logger, select_type="db-l", server_list=exclude)

        logger.info("cluster_prepare_ocfs2_conf: success.")
        return SUCC()
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("cluster_perpare_ocfs2_conf: failed. => %s" % errmsg)
        return FAIL(exception=e)


@exposed
def cluster_start_o2cb_service(ignore_connection_error=True):
    """Start the o2cb cluster stack on all nodes in the pool.

    @param ignore_connection_error: if the server cannot be connected, ignore this server.

    NOTE:
        1) Generate %s from db
        2) Propagate to all the nodes in the pool.
        3) Start/Reload o2cb service on all nodes.
    """ % OCFS2_CLUSTER_CONF
    logger = logging.getLogger("ovs.operation")
    try:
        #generate OCFS2_CLUSTER_CONF
        rs = cluster_generate_ocfs2_cluster_conf()
        cluster_conf = GETVAL(rs)

        exclude = []
        if ignore_connection_error:
            exclude = get_server_list_with_conn_err()

        #nodes update OCFS2_CLUSTER_CONF
        operation = "clusterm_set_ocfs2_cluster_conf"
        param = (cluster_conf,)
        servers_do_wrapper(operation, param, logger, select_type="db-l", server_list=exclude)

        #nodes start o2cb service
        operation = "clusterm_start_o2cb_service"
        param = ()
        servers_do_wrapper(operation, param, logger, select_type="db-l", server_list=exclude)

        rs = SUCC()
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("cluster_start_o2cb_service: => %s" % errmsg)
        return FAIL(exception=e)


def validate_heartbeat_file(hb_file):
    logger = logging.getLogger("ovs.operation")
    try:
        cmd = "debugfs.ocfs2 %s <<< stats 2>/dev/null | grep 'Flags:'" % hb_file
        output = old_run_cmd(cmd)
        if "Valid System Superblock" in output:
            cmd = "debugfs.ocfs2 %s <<< 'stat //heartbeat' 2>/dev/null | grep 'Flags:'" % hb_file
            output = old_run_cmd(cmd)
            if "Valid System Heartbeat" in output:
                logger.debug("validate_heartbeat_file: success.")
                return
    except:
        pass
    # remove hb
    logger.debug("validate_heartbeat_file: remove hb file.")
    cmd = [join(AGENT_DIR, "utils/do_heartbeat_nfs.sh"), '--action', 'remove', '--heartbeatfile', hb_file]
    run_cmd(cmd)


@exposed
def cluster_start_heartbeat(root_sr_uuid="", ignore_connection_error=True):
    """Start heartbeat among the cluster.
    """
    logger = logging.getLogger("ovs.operation")
    try:
        if not root_sr_uuid:
            root_sr_uuid = db_load("cluster", "cluster_root_sr_uuid")

        #create and initialize heartbeat it.
        root_sr = db_load("sr", root_sr_uuid)
        root_sr.initialize()

        if root_sr.sp.fs_type != "ocfs2.cluster":
            hb_file = join(root_sr.mountpoint, SR_HEARTBEAT_FILE)

            # validate hb file
            if exists(hb_file):
                validate_heartbeat_file(hb_file)

            cmd = [join(AGENT_DIR, "utils/do_heartbeat_nfs.sh"), '--action', 'create', '--heartbeatfile', hb_file]
            try:
                run_cmd(cmd)
            except Exception, e:
                cmd = [join(AGENT_DIR, "utils/do_heartbeat_nfs.sh"), '--action', 'remove', '--heartbeatfile', hb_file]
                run_cmd(cmd)
                raise e

        exclude = []
        if ignore_connection_error:
            exclude = get_server_list_with_conn_err()

        operation, param = "clusterm_start_heartbeat", (root_sr_uuid,)
        servers_do_wrapper(operation, param, logger, select_type="db-l", server_list=exclude)

        rs = SUCC()
        logger.info("cluster_start_heartbeat: => %s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("cluster_start_heartbeat: => %s" % errmsg)
        return FAIL(exception=e)


@exposed
def cluster_stop_heartbeat(root_sr_uuid="", ignore_connection_error=True):
    """Stop heartbeat among the cluster.
    """
    logger = logging.getLogger("ovs.operation")
    try:
        if not root_sr_uuid:
            root_sr_uuid = db_load("cluster", "cluster_root_sr_uuid")

        exclude = []
        if ignore_connection_error:
            exclude = get_server_list_with_conn_err()

        operation, param = "clusterm_stop_heartbeat", (root_sr_uuid,)
        servers_do_wrapper(operation, param, logger, select_type="db-l", server_list=exclude)

        rs = SUCC()
        logger.info("cluster_stop_heartbeat: => %s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("cluster_stop_heartbeat: => %s" % errmsg)
        return FAIL(exception=e)


from OVSXCluster import _check

def cluster_clear_ocfs2_config():
    logger = logging.getLogger("ovs.operation")
    try:
        exclude = get_server_list_with_conn_err()
        cluster_conf = ""
        operation = "clusterm_set_ocfs2_cluster_conf"
        param = (cluster_conf,)
        servers_do_wrapper(operation, param, logger, select_type="db-l", server_list=exclude)

        logger.info("cluster_clear_ocfs2_config: success.")
        return SUCC()
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("cluster_clear_ocfs2_config: failed. => %s" % errmsg)
        return FAIL(exception=e)

@exposed
def cluster_init_sr():
    logger = logging.getLogger("ovs.operation")
    try:
        exclude = get_server_list_with_conn_err()
        operation = "clusterm_init_sr"
        param = ()
        servers_do_wrapper(operation, param, logger, select_type="db-l", server_list=exclude)

        logger.info("cluster_init_sr: success.")
        return SUCC()
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("clusterm_init_sr: failed. => %s" % errmsg)
        return FAIL(exception=e)


@exposed
def cluster_precheck(single_node=False, ha_enable=True):
    """
    Check cluster prerequisite before setup.
    @param single_node: single node cluster?
    @return "success" if success, else "failed:<error message>"
    """
    logger = logging.getLogger("query")
    try:
        # 1) cluster root sr
        cluster_root_sr_uuid = db_load("cluster", "cluster_root_sr_uuid_new", get=True)
        if not cluster_root_sr_uuid:
            raise Exception("No root storage repository! Please setup storage repositories with repos.py before server pool creation.")

        # 2) check sr types
        if ha_enable or not single_node:
            supported_types = ["nfs", "ocfs2.cluster"]
            d = db_dump("sr")
            for sr_uuid, sr in d.items():
                if not sr.sp.fs_type in supported_types:
                    raise Exception("SR '%s' not supported: type '%s' not in %s" % (sr.sp.fs_spec, sr.sp.fs_type, supported_types))

        logger.info("cluster_precheck: success. single_node=%s, ha_enable=%s" % (single_node, ha_enable))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("cluster_precheck: faild. => %s" % errmsg)
        return "failed:%s" % errmsg


@exposed
def cluster_check_prerequisite(hosts="", ha_enable=False):
    """
    Check cluster prerequisite before setup:
        *) oracle vm version
        *) agent version
        *) cpu compatibility
        *) hostname/IP
        *) storage repositories
        *) ocfs2 configuration

    @param hosts    hostname:agent_passwd[,hostname:agent_passwd]
    @return         "success" if success, else "failed:<error message>"
    """
    logger = logging.getLogger("operation")
    try:
        if hosts == "":
            d = db_dump("srv")
            if len(d) == 0:
                raise Exception("No server information.")

        rs = []
        rs.append(("ha_check_oracle_vm_version", ha_check_oracle_vm_version(hosts)))
        rs.append(("ha_check_agent_version", ha_check_agent_version(hosts)))
        rs.append(("ha_check_cpu_compatibility", ha_check_cpu_compatibility(hosts)))
        rs.append(("ha_check_hostname_ip", ha_check_hostname_ip(hosts)))
        rs.append(("ha_check_storage_repos", ha_check_storage_repos(hosts, ha_enable)))
        rs.append(("ha_check_ocfs2_config", ha_check_ocfs2_config(hosts)))

        failed = [(a, b) for (a, b) in rs if not b.startswith("success")]
        if failed:
            msg = "\n".join(["\n".join(t) for t in failed])
            raise Exception(msg)

        logger.info("cluster_check_prerequisite: success. ha_enable=%s" % ha_enable)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("cluster_check_prerequisite: faild. => %s" % errmsg)
        return "failed:%s" % errmsg

# alias
ha_check_cluster_prerequisite = cluster_check_prerequisite


def _cluster_setup(ha_enable=True):
    """Setup the cluster when
    1) HA is enabled or disabled
    2) Nodes change

    @return     "success" if success, else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        # start o2cb service
        ret = cluster_start_o2cb_service()
        _check(ret)

        init_root = db_load("cluster", "init_root", get=True)
        init_root = init_root or False
        local_db = db_load("LOCAL_DB,node", "local_db")
        ha_enable = ha_enable and True or False

        d = db_dump("srv")
        if len(d) == 0:
            raise Exception("No server information.")

        # cluster precheck
        ret = cluster_check_prerequisite(ha_enable=ha_enable)
        _check(ret)

        # init root sr
        cluster_root_sr_uuid = db_load("cluster", "cluster_root_sr_uuid_new")
        ret = cluster_init_root_sr(cluster_root_sr_uuid.encode("utf-8"), init_root, local_db)
        _check(ret)

        # init all sr
        ret = cluster_init_sr()
        _check(ret)

        # start/stop cluster heartbeat
        if ha_enable:
            ret = cluster_start_heartbeat()
            _check(ret)
        else:
            # get the original ha_enable flag
            flag = db_load("cluster", "pool_ha_enable", get=True)
            if flag:
                ret = cluster_stop_heartbeat()
                _check(ret)
                ret = clusterm_cleanup_master_env()
                _check(ret)

        # sync dlm locks
        if ha_enable:
            ha_sync_dlm_lock()
        else:
            ha_remove_dlm_lock()

        # save configuration
        db_save("cluster", "pool_ha_enable", ha_enable)
        db_save("cluster", "init_root", False)

        logger.info("cluster_setup: success. ha_enable=%s" % ha_enable)
        return "success"

    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("cluster_setup: failed. => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def cluster_setup(ha_enable=True):
    # set timeout to 600 seconds
    logger = logging.getLogger("ovs.operation")
    try:
        cds = CDS("cluster_setup", timeout=600)
        try:
            return _cluster_setup(ha_enable)
        finally:
            cds.close()
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("cluster_setup: failed. => %s" % errmsg)
        return "failed:%s" % errmsg

# alias
set_pool_ha_enable = cluster_setup

@exposed
def cluster_cleanup_vip():
    """Cleanup cluster vip."""
    logger = logging.getLogger("ovs.operation")
    try:
        d = db_dump("srv")
        if d:
            exclude = get_server_list_with_conn_err()
            operation = "clusterm_cleanup_master_env"
            param = ()
            servers_do_wrapper(operation, param, logger, select_type="db-l", server_list=exclude)
        else:
            # single node, no server registered
            ret = clusterm_cleanup_master_env()
            _check(ret)

        db_remove("cluster", "master_vip")

        logger.info("cluster_cleanup_vip: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("cluster_cleanup_vip: failed. => %s" % errmsg)
        return "failed:%s" % errmsg


@exposed
def cluster_teardown(keep_sr=True):
    """ Remove the cluster configuration.
        1) cleanup vip
        2) remove all sr
        3) purge db
        4) ...
    """
    logger = logging.getLogger("ovs.operation")
    try:
        # 1)
        ret = cluster_cleanup_vip()
        _check(ret)

        # 2)
        if keep_sr:
            cluster_root_sr_uuid = db_load("cluster", "cluster_root_sr_uuid_new", get=True)
            # backup
            src_dir = "%s/.ovs-agent/db" % OVS_BASE
            dst_dir = "/tmp"
            db_bak = ["sr.db", "sp.db"]
            for db in db_bak:
                src = join(src_dir, db)
                dst = join(dst_dir, db)
                shutil.copy(src, dst)

        # 3)
        d = db_dump("srv")
        if d:
            exclude = get_server_list_with_conn_err()
            operation = "clusterm_teardown"
            param = ()
            servers_do_wrapper(operation, param, logger, select_type="db-l", server_list=exclude)
        else:
            # single node, no server registered
            ret = clusterm_teardown()
            _check(ret)

        # 4)
        if keep_sr:
            if cluster_root_sr_uuid:
                db_save("cluster", "cluster_root_sr_uuid_new", cluster_root_sr_uuid)
                db_save("cluster", "init_root", True)

            for db in db_bak:
                src = join(src_dir, db)
                dst = join(dst_dir, db)
                shutil.move(dst, src)

        # 5)
        ha_remove_dlm_lock()

        logger.info("cluster_teardown: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("cluster_teardown: failed. => %s" % errmsg)
        return "failed:%s" % errmsg
