#!/usr/bin/env python
#^#
# OVSSiteDB.py
#
# <description>
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import socket
import re
import logging
from os.path import join, basename, exists

from OVSServerProxy import *
from OVSMultiprocess import *
from OVSExceptionLookup import *
from OVSException import *
from OVSDB import db_get_srv, db_update_srv, db_del_srv
from OVSConstants import *
from OVSCommons import *
from OVSCDS import *


#-----------------------------------------------------------------------
@exposed
def register_server(hostname, srv_role, is_shared=True,
        agt_username=DEFAULT_AGENT_USERNAME,
        agt_passwd=DEFAULT_AGENT_PASSWD,
        agt_port=DEFAULT_AGENT_PORT,
        ssl_enabled=True):
    """Register server.

    @param srv_role: "site", "xen" or "utility"
    @return: "success" if success else "failed:<error message>"

    NOTE:
      when register a server, agent on the server must be up and running
    for collecting server infomation
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.debug("register_server: hostname=%s, srv_role=%s, is_shared=%s, agt_username=%s, agt_port=%s, ssl_enable=%s" % \
                (hostname, srv_role, repr(is_shared), agt_username, agt_port, repr(ssl_enabled)))

        # agt url
        agt_url = "http%s://%s:%s@%s:%s" % (ssl_enabled and "s" or "", agt_username, agt_passwd, hostname, agt_port)

        #check if master server already registered.
        master_weight = 0
        srv_role = srv_role.lower()
        if "site" in srv_role:
            cds = CDS("srv")
            try:
                candidates = [_hostname for (_hostname, srv) in cds.db_srv.items()
                        if (srv.has_key("srv_role") and "site" in srv["srv_role"])]
            finally:
                cds.close()

            if len(candidates) > 1:
                raise Exception("More than one server pool master exists: %s" % repr(candidates))
            if len(candidates) == 1 and candidates[0] != hostname:
                raise Exception("Master pool server already exists: %s" % candidates[0])
            # master server
            master_weight = 1
            # save master hostname
            db_save("cluster", "master_hostname", hostname)
            # save master agt url
            db_save("cluster", "master_agt_url", agt_url)

        old_srv = None
        #check if server with same ip is already registered.
        ip = get_host_by_name(hostname)
        cds = CDS("srv", "srvp")
        try:
            for _hostname in cds.db_srv:
                if _hostname != hostname and get_host_by_name(_hostname) == ip:
                    #raise Exception("srv('%s') with ip('%s') is already registered." % (_hostname, ip))
                    #reuse what has already been in db
                    hostname = _hostname
                    break
            cds.db_srvp[ip] = hostname

            if hostname in cds.db_srv:
                old_srv = cds.db_srv[hostname]
        finally:
            cds.close()

        srv = {
            #is in shared server pool, vm can be balanced to this server.
            "is_shared": (is_shared and True or False), #FIXME: not used
            "srv_role": srv_role,
            "ip": ip,
            "agt_addr": hostname,
            "agt_username": agt_username,
            "agt_passwd": agt_passwd,
            "agt_port": agt_port,
            "agt_url": agt_url,
            "master_weight": master_weight,
            }

        sp = ServerProxy(srv["agt_url"])
        rs = sp.sys_os_info()
        if not rs.startswith("success"):
            raise Exception(rs)
        rs = parse_result(rs)
        srv["arch"] = rs["arch"]
        srv["cpu_num"] = int(rs["cpu_num"])
        srv["cpu_speed"] = float(rs["bogomips"])
        srv["mem_total"] = int(rs["mem_total"])
        srv["status"] = "RUNNING"
        try:
            from OVSServerSelection import _get_agent_version
            srv["agent_version"] = _get_agent_version(srv["agt_url"])
        except: pass
        if "hostname" in rs:
            if rs["hostname"].startswith("localhost"):
                raise Exception("hostname cannot be 'localhost*'!")
            srv["hostname"] = rs["hostname"]
        if "xen" in srv_role:
            try:
                rs = sp.xen_hypervisor_info()
                if rs.startswith("success"):
                    rs = parse_result(rs)
                    srv["vcaps"] = rs["vcaps"]
                    srv["mem_free"] = rs["mem_free"]
                else:
                    raise Exception(rs)
            except Exception, e:
                errmsg = fmt_err_msg(e)
                logger.warning("register_server: warning. hostname=%s => %s" % \
                    (hostname, errmsg))

        sp.register_hostname(hostname)

        #add hostname to ocfs2 cluster
        #FIXME: cluster_start_o2cb_service()
        if old_srv:
            old_srv.update(srv)
            srv = old_srv
        db_save("srv", hostname, srv)
        logger.info("register_server: success. %s->%s" % (hostname, srv_role))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("register_server: failed! %s->%s => %s" \
                % (hostname, srv_role, errmsg))
        return "failed:%s" % (errmsg,)

@exposed
def unregister_server(hostname):
    """Unregister a server from site

    @return: "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        host = db_load("srv", hostname, get=True)
        if not host:
            raise Exception("This server is not registered.")

        sp = ServerProxy(host["agt_url"])
        rs = sp.unregister_hostname()
        if not rs.startswith("success"):
            raise Exception(rs)

        if "site" in host["srv_role"]:
            cluster_root_sr_uuid = db_load("cluster", "cluster_root_sr_uuid_new", get=True)
            # backup
            src_dir = "%s/.ovs-agent/db" % OVS_BASE
            dst_dir = "/tmp"
            db_bak = ["sr.db", "sp.db"]
            for db in db_bak:
                src = join(src_dir, db)
                dst = join(dst_dir, db)
                shutil.copy(src, dst)

        ret = sp.clusterm_teardown()
        if not rs.startswith("success"):
            raise Exception(rs)

        if "site" in host["srv_role"]:
            if cluster_root_sr_uuid:
                db_save("cluster", "cluster_root_sr_uuid_new", cluster_root_sr_uuid)
                db_save("cluster", "init_root", True)

            for db in db_bak:
                src = join(src_dir, db)
                dst = join(dst_dir, db)
                shutil.move(dst, src)

        #FIXME: reassign master ip to another server.
        db_del_srv(hostname)

        #remove hostname from ocfs2 cluster
        #FIXME: cluster_start_o2cb_service()

        logger.info("unregister_server: success. srv('%s') removed." % hostname)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("unregister_server: failed. srv('%s') => %s" % \
                (hostname, errmsg))
        return "failed:%s" % errmsg


#-----------------------------------------------------------------------
@exposed
def is_registered():
    """Check whether a server is registered or not."""
    logger = logging.getLogger("ovs.query")
    try:
        hostname = db_load("LOCAL_DB,node", "registered_hostname", get=True)
        if hostname: rs = "true"
        else: rs = "false"
        logger.info("is_registered: success. => %s" % rs)
        return "success:%s" % rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("is_registered: failed. => %s" % errmsg)
        return "failed:%s" % errmsg


@exposed
def get_srv_agent_status(hostname):
    """Check server agent aliveness.
    @return: "success:echo=<msg>" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.query")
    try:
        srv = db_get_srv(hostname)
        sp = ServerProxy(srv['agt_url'])
        rs = sp.echo('hello')
        logger.info("get_srv_agent_status: agent('%s')=> %s" % (hostname, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        try:
            db_update_srv(hostname, {"status" : "UNKNOWN"})
        except:
            pass
        logger.error("get_srv_agent_status: error. agent('%s')=>%s" % \
                (hostname, errmsg))
        return "failed:%s" % errmsg


@exposed
def set_shutdown_mode(ips, status=True):
    """Register the server to enter shutdown mode or not

    @param ips: "," seperated list of the ips assigned to a server.
    """
    logger = logging.getLogger("ovs.operation")
    try:
        status = bool(status)
        hostname = ""
        for ip in ips.split(","):
            try:
                srv = db_get_srv(ip)
                hostname = srv.get("hostname")
                db_update_srv(hostname, {"shutdown_mode": status})
                break
            except:
                pass
        if hostname:
            logger.info("set_shutdown_mode: srv('%s')=%s" % (hostname, repr(status)))
        else:
            logger.warning("set_shutdown_mode: srv('%s')=%s" % (ips, repr(status)))

        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("enable_pool_ha:failed. =>%s" % errmsg)
        return "failed:%s" % errmsg


#----------------------------------------------------------------------
from OVSSiteCluster import cluster_cleanup_vip
@exposed
def register_master_vip(vip):
    """Register master vip for the pool
    """
    logger = logging.getLogger("ovs.operation")
    try:
        cluster_cleanup_vip()
        db_save("cluster", "master_vip", vip)

        logger.info("register_master_vip: success. vip=%s" % vip)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("register_master_vip: failed. vip=%s => %s" % (vip, errmsg))
        return "failed: %s" % errmsg


@exposed
def get_master_vip():
    """Get master vip for the pool
    """
    logger = logging.getLogger("ovs.query")
    try:
        try:
            vip = db_load("cluster", "master_vip")
        except KeyError:
            vip = ""

        logger.info("get_master_vip: success. vip=%s" % vip)
        return "success:vip=%s" % vip
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_master_vip: failed. => %s" % errmsg)
        return "failed: %s" % errmsg

@exposed
def validate_master_vip(vip, base_ip):
    """validate master vip for the pool
        1. vip must be legal ip address
        2. vip should not be in use
        3. vip,base_ip are in the same network
    """
    def _legal_ipaddr(ip):
        import socket
        try:
            socket.inet_aton(ip)
            return True
        except:
            return False

    def _get_network(ip, mask):
        args = ["ipcalc", "-n", ip, mask]
        output = run_cmd(args).strip()
        if not output.startswith("NETWORK="):
            raise Exception("Failed to get network.")
        return output[8:]

    logger = logging.getLogger("ovs.operation")
    try:
        if not _legal_ipaddr(vip):
            raise Exception("VIP %s is not legal IP address." % vip)
        if not _legal_ipaddr(base_ip):
            raise Exception("Base IP %s is not legal IP address." % base_ip)
        if host_alive(vip):
            raise Exception("VIP %s is in use." % vip)

        cmd = "ifconfig | grep '%s ' | awk -F: '{print $4}'" % base_ip
        netmask = old_run_cmd(cmd).strip()
        if netmask == "":
            raise Exception("Failed to get netmask for base ip: %s" % base_ip)

        try:
            network1 = _get_network(base_ip, netmask)
        except:
            raise Exception("Failed to get network value: base_ip=%s, netmask=%s" % (base_ip, netmask))
        try:
            network2 = _get_network(vip, netmask)
        except:
            raise Exception("Failed to get network value: vip=%s, netmask=%s" % (vip, netmask))

        if network1 != network2:
            raise Exception("VIP %s and base IP %s are not in the same network." % (vip, base_ip))

        logger.info("validate_master_vip: success. vip=%s, base_ip=%s" % (vip, base_ip))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("validate_master_vip: failed. vip=%s, base_ip=%s => %s" % (vip, base_ip, errmsg))
        return "failed:%s" % errmsg


@exposed
def get_master_ip():
    """Get current master ip
    """
    logger = logging.getLogger("ovs.query")
    try:
        try:
            ip = db_load("cluster", "master_ip")
        except KeyError:
            try:
                hostname = db_load("cluster", "master_hostname")
                ip = get_host_by_name(hostname)
            except KeyError:
                raise Exception("master_ip has not been registered yet.")

        logger.info("get_master_ip: success. ip=%s" % ip)
        return "success:ip=%s" % ip
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_master_ip: failed. => %s" % errmsg)
        return "failed: %s" % errmsg


#----------------------------------------------------------------------
@interconnect
def register_hostname(hostname):
    """
    Tell the registered server what hostname is used in the cluster db.
    @param hostname: the registered hostname.
    """
    logger = logging.getLogger("ovs.operation")
    try:
        db_save("LOCAL_DB,node", "registered_hostname", hostname)
        ip = get_host_by_name(hostname)
        db_save("LOCAL_DB,node", "registered_ip", ip)

        rs = SUCC()
        logger.info("registered_hostname: hostname=%s => %s" % (hostname, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("registered_hostname: hostname=%s => %s" % (hostname, errmsg))
        return FAIL(exception=e)


@interconnect
def unregister_hostname():
    logger = logging.getLogger("ovs.operation")
    try:
        db_remove("LOCAL_DB,node", "registered_hostname")
        db_remove("LOCAL_DB,node", "registered_ip")

        logger.info("unregister_hostname: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("unregister_hostname: failed. => %s" % errmsg)
        return "failed:%s" % errmsg
