#!/usr/bin/python

import socket
import logging
from os.path import join, basename, exists

from OVSXXenVMConfig import xen_update_uuid, xen_get_vm_type
from OVSXUtility import find_real_path
from OVSServerSelection import collect_srv_stat
from OVSServerProxy import *
from OVSDB import db_get_srv, db_update_srv, db_del_srv
from OVSDB import db_update_vm, db_get_vm_path, db_get_real_vm_path
from OVSConstants import *
from OVSCommons import *
from OVSCDS import *
from OVSMultiprocess import *
from OVSProtocol import *

#-----------------------------------------------------------------------
@exposed
def register_vm(vm_path, vtech="xen",
        agt_username=DEFAULT_AGENT_USERNAME,
        agt_passwd=DEFAULT_AGENT_PASSWD,
        agt_port=DEFAULT_AGENT_PORT,
        packaged=False):
    """Register vm

    @return: "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)

        real_path = ""
        try:
            real_path = find_real_path(vm_path)
            vm_cfg_file = join(real_path, "vm.cfg")
            if exists(vm_cfg_file):
                xen_update_uuid(vm_cfg_file)
        except:
            # register non-existent vm
            pass

        d = {
            "vtech": vtech.lower().strip(),
            "agt_addr": "",
            "agt_username": agt_username,
            "agt_passwd": agt_passwd,
            "agt_port": agt_port,
            "real_path": real_path,
            "host_srv": "",
            "packaged": packaged,
            "status": "DOWN",
            "readlock": 0,
            "manual_shutdown": True
        }
        db_update_vm(vm_path, d)

        db_get_vm_path(basename(vm_path))

        #sync vm status
        get_live_vm_info()
        vm = db_load("vm", vm_path)

        #try to update the uuid for the vm
        if vm["status"] == "RUNNING":
            from OVSServerSelection import get_vm_host
            hostname, sp = get_vm_host(vm_path)
            rs = sp.xen_get_running_vm_uuid(vm_path)
            if rs.startswith("success"):
                xen_update_uuid(vm_cfg_file, parse_result(rs)["uuid"])

        logger.info("register_vm: success. vm('%s')" % vm_path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("register_vm: failed. vm('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def unregister_vm(vm_path):
    """Unregister vm

    @return: "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        db_remove("vm", vm_path)
        logger.info("unregister_vm: success. vm('%s')" % (vm_path,))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("unregister_vm: failed. vm('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

#----------------------------------------------------------------------
@exposed
def get_vm_hostserver(vm_path):
    """Get vm host server, if down, then hostname should be empty

    @return: "success:host_srv=<hostname>" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm = db_load("vm", vm_path)

        logger.info("get_vm_hostserver: success. vm('%s') => host_srv='%s'" % \
                (vm_path, vm['host_srv']))
        return "success:host_srv=%s" % vm["host_srv"]
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_hostserver: error. vm('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def set_vm_preferred_servers(vm_path, preferred_servers):
    """Set preferred servers for the vm to running on

    @param preferred_servers: preferred servers that the VM will be running on one of them,
        format: "<hostname>[,<hostname>]" e.g.
            "oops1.cn.oracle.com,oops2.cn.oracle.com"
    @return: "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        preferred_server_list = [ server.strip()
                                    for server in preferred_servers.strip().split(",")
                                    if server.strip() ]
        db_update_vm(vm_path, {"preferred_servers":preferred_server_list})

        logger.info("set_vm_preferred_servers:success. vm('%s')=>%s" % (
                vm_path, repr(preferred_server_list)))
        #TODO
        #how can remove the field:
        #preferred_servers while preferred_server_list = []?
        #since db_update_vm can only update ,but can not remove or pop
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("set_vm_preferred_servers:vm('%s') failed. => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def get_vm_preferred_servers(vm_path):
    """
    @return: "success:<string of hostnames>" if success else "failed:<error message>"
        example return value: "success:oops1.cn.oracle.com,oops2.cn.oracle.com,oops3.cn.oracle.com"
        NOTE:
            "success:" indicates no preferred servers information stored for specified VM.
    """
    logger = logging.getLogger("ovs.query")
    try:
        #just incase prefered_servers not exists or someone set it to None
        vm = db_load("vm", vm_path)
        l = vm.get("preferred_servers")
        preferred_servers = ""
        if l:
            preferred_servers = ",".join(l)

        logger.info("get_vm_preferred_servers:success. vm('%s')=>%s" % (
                vm_path, repr(preferred_servers)))
        return "success:%s" % preferred_servers
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_preferred_servers:vm('%s') failed. => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg


@exposed
def set_vm_policy(vm_path, policies, append=True):
    """Set vm policy

    @param policies: priority,policy_name[;priority,policy_name] priority in (-100, 100)

    NOTE:
        mainly for set HA vm autostart policy.
    """
    logger = logging.getLogger("ovs.operation")
    try:
        vm = db_load("vm", vm_path)
        if "policy" not in vm:
            vm["policy"] = []

        l = [e.strip().split(",") for e in policies.split(";") if e.strip()]
        l = [(int(priority), policy_name) for priority, policy_name in l]

        from OVSPolicy import policy
        for priority, policy_name in l:
            if policy_name not in policy:
                raise Exception("policy('%s') not in %s" % (policy_name, repr(policy.keys())))

        if not append:
            vm['policy'] = l
        else:
            vm["policy"] += l
        vm["policy"] = list(set(vm["policy"])) # remove duplicate
        db_save("vm", vm_path, vm)

        logger.info("set_vm_policy: vm('%s') policy=%s append=%s=> success." % (vm_path, repr(vm['policy']),repr(append)))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("set_vm_policy: failed. => %s" % errmsg)
        return "failed: %s" % errmsg

@exposed
def get_vm_policy(vm_path):
    """Get vm policy

    @return: "success:repr(vm['policy'])"
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm = db_load("vm", vm_path)
        if "policy" in vm:
            policy = repr(vm["policy"])
        else:
            policy = ""
        rs = "success:policy=%s" % policy

        logger.info("get_vm_policy: %s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_policy: failed. => %s" % errmsg)
        return "failed: %s" % errmsg

@exposed
def set_vm_status(vm_path, status):
    """
    @param status: valid status in ('DOWN', 'RUNNING', 'ERROR', ...)
    @return: "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        db_update_vm(vm_path, {"status":status})
        logger.info("set_vm_status:success. vm('%s')=>%s" % (vm_path, status))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("set_vm_status:vm('%s') failed. => %s" % \
                (vm_path, errmsg(e)))
        return "failed:%s" % errmsg(e)

@exposed
def get_vm_attr(vm_path, attr_name):
    """
    @return: "success:repr(vm[attr_name])"
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm = db_load("vm", vm_path)
        if attr_name in vm:
            attr = repr(vm[attr_name])
        else:
            attr = ""
        rs = "success:%s=%s" % (attr_name, attr)

        logger.info("get_vm_attr: %s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_vm_attr: failed. => %s" % errmsg)
        return "failed: %s" % errmsg

#-----------------------------------------------------------------------
def get_live_vms(vtech="xen"):
    """Query each host server for live vms, and update vm.db accordingly

    @return: [vm1, vm2,] => list of running vms,
    """
    logger = logging.getLogger("ovs.query")
    logger.info("get_live_vms: start...")

    vtech = vtech.lower().strip()
    collect_srv_stat()
    cds = CDS("srv")
    try:
        candidates = [hostname for (hostname, srv) in cds.db_srv.items() if (vtech in srv["srv_role"] and srv["status"] == "RUNNING")]
    finally:
        cds.close()

    operation = "%s_get_vms_name_vncport_macs" % vtech
    param = ()
    rs_list = servers_do(operation, param, select_type="db&l", server_list=candidates, timeout=15)

    d_host_vms = {}
    for hostname, rs in rs_list:
        if rs["status"] == OP_STATUS_FAIL:
            logger.error("%s: srv('%s') => %s" % (operation, hostname, JSON(rs)))
        else:
            d = GETVAL(rs)
            name_vncport_macs = d["vm_info"]
            if not name_vncport_macs:
                continue
            d_host_vms[hostname] = name_vncport_macs.split(",")

    #update db
    site_running_vms = []
    d_vm_in_doubt = {}
    cds = CDS("vm", "vmp")
    try:
        for hostname in d_host_vms:
            vm_infos = d_host_vms[hostname]
            for vm_info in vm_infos:
                vm_name, vncport, macs = vm_info.split("@")
                site_running_vms.append(vm_name)

                #updated vm info => vncport, host_srv
                if vncport: vncport = int(vncport)
                else: vncport = -1
                d = { "host_srv": hostname, "vnc_port": vncport }

                if vm_name in cds.db_vmp:
                    vm_path = cds.db_vmp[vm_name]
                    if vm_path in cds.db_vm:
                        vm = cds.db_vm[vm_path]
                        vm.update(d)
                        cds.db_vm[vm_path] = vm
                    else: #no entry in vm.db
                        d_vm_in_doubt[vm_name] = d
                else: #no entry in vmp.db
                    d_vm_in_doubt[vm_name] = d
    finally: cds.close()
    #import vm (manage vm not created by ovs manager)
    #FIXME: possible conflict:
    # if user startup a vm outside OVS_BASE but with same name to a vm ovs managed.
    for vm_name in d_vm_in_doubt:
        vm_info = d_vm_in_doubt[vm_name]
        vm_path = join(OVS_BASE, "running_pool/%s" % vm_name)
        try:
            real_path = find_real_path(vm_path)
            register_vm(vm_path, vtech)
            db_update_vm(vm_path, vm_info)
        except: #cannot get real_path
            if vm_name in site_running_vms:
                site_running_vms.remove(vm_name)

    logger.info("get_live_vms: rs=%s" % repr(site_running_vms))
    return site_running_vms

def sync_vm_status(site_running_vms, agent_restart=True):
    """Sync vm status based on info gethered from host servers

    @return: site_running_vms
    """
    logger = logging.getLogger("ovs.query")
    logger.info("sync_vm_status: start...")

    #rebuild index vmp
    if agent_restart:
        cds = CDS("vm", "vmp")
        try:
            for vm_path in cds.db_vm:
                vm_name = basename(vm_path)
                cds.db_vmp[vm_name] = vm_path
        finally:
            cds.close()

    if not agent_restart:
        #routine sync, agent works normal
        #if not in running list, if status is RUNNING then set to DOWN
        l_in_doubt = []
        cds = CDS("vm", "vmp")
        try:
            vm_name_non_exist = []
            for vm_name in cds.db_vmp:
                vm_path = cds.db_vmp[vm_name]
                try:
                    vm = cds.db_vm[vm_path]
                    if not isinstance(vm, dict):
                        del cds.db_vm[vm_path]
                        vm_name_non_exist.append(vm_name)
                        continue
                    if vm["status"] == "RUNNING" and vm_name not in site_running_vms:
                        l_in_doubt.append(vm_path)
                except KeyError:
                    vm_name_non_exist.append(vm_name)
            for vm_name in vm_name_non_exist:
                del cds.db_vmp[vm_name]
        finally: cds.close()

        #double check if the vm in doubt is really running or down
        for vm_path in l_in_doubt:
            try:
                vm = db_load("vm", vm_path)
                srv = db_get_srv(vm['host_srv'])
                if "real_status" not in srv \
                        or ("real_status" in srv and srv["real_status"] == "ALIVE"):
                    sp = ServerProxy(srv['agt_url'])
                    rs = getattr(sp, "%s_get_vm_names" % vm['vtech'])()
                    vm_names = parse_result(rs)["vm_names"].split(",")
                    vm_name = basename(vm_path)
                    if vm_name not in vm_names: #not in running list indeed
                        logger.debug("sync_vm_status: vm('%s') RUNNING -> DOWN!" % vm_path)
                        db_update_vm(vm_path, {"status" : "DOWN"})
                    else:
                        site_running_vms.append(vm_name)
                else: #server is DEAD, vm has gone with it
                    logger.debug("sync_vm_status: vm('%s') RUNNING -> DOWN!" % vm_path)
                    db_update_vm(vm_path, {"status" : "DOWN"})

            except Exception, e:
                logger.warning("sync_vm_status: vm('%s') => %s" % (vm_path, fmt_err_msg(e)))
        #sync DOWN -> RUNNING if in site_running_vms
        cds = CDS("vm", "vmp")
        try:
            for vm_name in site_running_vms:
                try:
                    vm_path = cds.db_vmp[vm_name]
                    vm = cds.db_vm[vm_path]
                    if vm['status'] == "DOWN":
                        vm['status'] = "RUNNING"
                        vm['manual_shutdown'] = False
                        cds.db_vm[vm_path] = vm
                except:
                    pass
        finally:
            cds.close()

    else:
        #called by agtctl startup
        #if in running list, set to RUNNING else DOWN
        cds = CDS("vm")
        try:
            for vm_path in cds.db_vm:
                try:
                    vm = cds.db_vm[vm_path]
                    vm_name = basename(vm_path)
                    if vm_name in site_running_vms:
                        vm["status"] = "RUNNING"
                        vm['manual_shutdown'] = False
                    else:
                        vm["status"] = "DOWN"
                        vm["readlock"] = 0
                    cds.db_vm[vm_path] = vm
                except: pass
        finally:
            cds.close()

    logger.info("sync_vm_status: rs=%s" % repr(site_running_vms))
    return site_running_vms

@exposed
def get_live_vm_info(agent_restart=False):
    """Get info of all live vms, update the database

    @return: "success:<string of host-info>" if success else "failed:<error message>"
    example of string of host-info: "host1=vm1:ip1:vnc_port1;host2=vm2:ip2:vnc_port2,vm3:ip3:vnc_port3"
    string of host-info can be an empty string; ip/vnc_port can be an empty string.
    """
    logger = logging.getLogger("ovs.query")
    #make it ready to be called by utlis/do_rpc.py
    agent_restart = bool(agent_restart)
    try:
        site_running_vms = get_live_vms()
        site_running_vms = sync_vm_status(site_running_vms, agent_restart)

        host_vm_info = {}
        cds = CDS("vm", "vmp")
        try:
            for vm_name in site_running_vms:
                try:
                    vm_path = cds.db_vmp[vm_name]
                    vm = cds.db_vm[vm_path]
                    hostname = vm['host_srv']
                    if hostname:
                        if vm["vnc_port"] < 0: vnc_port = ""
                        else: vnc_port = "%d" % vm["vnc_port"]
                        vm_ip = ""
                        if "ip" in vm: vm_ip = vm["ip"]
                        if hostname in host_vm_info:
                            host_vm_info[hostname].append("%s:%s:%s" % (vm_name, vm_ip, vnc_port))
                        else:
                            host_vm_info[hostname] = [ "%s:%s:%s" % (vm_name, vm_ip, vnc_port), ]
                except Exception, e:
                    logging.debug("get_live_vm_info: warning. vm('%s') => %s" % \
                            (vm_name, fmt_err_msg(e)))
        finally: cds.close()
        rs = ""
        if host_vm_info:
            rs = ";".join([ "%s=%s" % (hostname, ",".join(vm_info))
                    for hostname, vm_info in host_vm_info.items() ])
        logger.info("get_live_vm_info: success. vm info => %s" % rs)
        return "success:%s" % rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("get_live_vm_info: failed. => %s" % errmsg)
        return "failed:%s" % errmsg


#----------------------------------------------------------------------
@exposed
def truncate_db():
    """Truncate agent db.

    @return: "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        cds = CDS("vm", "vmp", "srv")
        try:
            cds.db_vm.truncate()
            cds.db_vmp.truncate()
            cds.db_srv.truncate()
        finally:
            cds.close()
        logger.info("truncate_db: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("truncate_db: failed. => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def update_manual_shutdown(vm_name):
    """update the manual shutdown flag

    @return: "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        vm_path = db_get_vm_path(vm_name)
        if not vm_path:
            logger.debug("update_manual_shutdown: vm('%s') not in ovs-agent db." % vm_name)
        else:
            d = db_load("vm", vm_path)
            if "vtype" in d:
                vtype = d["vtype"]
            else:
                real_vm_path = db_get_real_vm_path(vm_path)
                rs = xen_get_vm_type(real_vm_path)
                if not rs.startswith("success"):
                    raise Exception(rs)
                else:
                    vtype = parse_result(rs)["type"]
                    if vtype == "hvm":
                        d["vtype"] = "hvm"
                    else:
                        d["vtype"] = "xen"
            # only update pvm
            if vtype != "hvm":
                d["manual_shutdown"] = True
                db_update("vm", vm_path, d)

        logger.info("update_manual_shutdown: success. vm_name='%s'" % vm_name)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("update_manual_shutdown: failed. => %s" % errmsg)
        return "failed:%s" % errmsg
