#!/usr/bin/env python
#^#
# OVSSiteVM.py
#
# <description>
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#




import logging
from os.path import join
from OVSServerSelection import select_tgt_host, get_vm_host
from OVSServerProxy import ServerProxy
from OVSDB import db_get_vm, db_update_vm, db_chkset_vm_status, db_get_real_vm_path
from OVSDB import db_get_srv, db_update_srv
from OVSCommons import exposed, parse_result, fmt_err_msg, chk_vm_path
from OVSCDS import CDS

from OVSXUtility import utl_get_free_space
from OVSXXenVMConfig import xen_get_memory
from OVSException import OVSException
from OVSSiteVMConfig import get_vm_vdisks

from OVSWrappers import raise_if_not_success

def _check_free_mem(server_sp, vm_path):
    """
    Check if the server has enough free memory to run/restore the vm.

    @param server_sp: ServerProxy of the server
    @type server_sp: ServerProxy
    @param vm_path: vm path
    @type vm_path: string
    @raise Exception: if there is not enough free memory for the vm,
                        raise a Exception.
    """
    ret = server_sp.xen_mem_free()
    raise_if_not_success(ret)
    mem_free = int(parse_result(ret)['mem_free'])
    real_vm_path = db_get_real_vm_path(vm_path)
    ret = xen_get_memory(real_vm_path)
    raise_if_not_success(ret)
    memory = int(parse_result(ret)['memory'])

    if mem_free < memory:
        raise Exception('No enough free memory.')

@exposed
def start_vm(vm_path, dedicated_server=None):
    """
    start a vm
    dedicated_server => circumvent the balance mechanism

    @return         "success:host_server=<host_server>" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("start_vm: vm('%s') dedicated_server=%s start..." % (vm_path, repr(dedicated_server)))
        chk_vm_path(vm_path)

        db_chkset_vm_status(vm_path, ("DOWN", "ERROR"), "PROCESSING")
        try:
            vm = db_get_vm(vm_path)
            real_vm_path = db_get_real_vm_path(vm_path)
            ret = xen_get_memory(real_vm_path)
            raise_if_not_success(ret)
            memory = int(parse_result(ret)['memory'])

            if dedicated_server:
                host_srv = dedicated_server
                srv = db_get_srv(host_srv)
                sp = ServerProxy(srv['agt_url'])
            else:
                host_srv = ""
                cds = CDS("start_vm", timeout=30)
                try:
                    logger.debug("start_vm: vm('%s') server selection starts..." % vm_path)
                    (host_srv, sp) = select_tgt_host(vm_path)
                    mem_free = db_get_srv(host_srv)["mem_free"]
                    mfree = mem_free - memory
                    db_update_srv(host_srv, {"mem_free": mfree})
                    logger.debug("start_vm: vm('%s') server selection done." % vm_path)
                finally:
                    cds.close()

            logger.debug("start_vm: vm('%s') on srv('%s') ..." % (vm_path, host_srv))

            # doulbe check the free memory of the selected server
            _check_free_mem(sp, vm_path)

            #start vm
            paras = (real_vm_path,)
            rs = getattr(sp, "%s_start_vm" % vm["vtech"])(*paras)

            #check result & update db
            if rs.startswith("success"):
                db_update_vm(vm_path, {"host_srv": host_srv, "status": "RUNNING", "manual_shutdown": False})
            else:
                # in case start_vm() fails, do not free the allocated memory
                # ovs-agent will collect server statistics every 60 seconds
                raise Exception("return=>%s" % rs)

        except OVSException, e:
            if e.errcode in ("20000", "20001", "20003"):
                db_chkset_vm_status(vm_path, "", "DOWN")
            else:
                db_chkset_vm_status(vm_path, "", "ERROR")
            raise e
        except Exception, e:
            try:
                db_chkset_vm_status(vm_path, "", "ERROR")
                db_update_vm(vm_path, {"host_srv": host_srv})
            except:
                pass
            raise e

        rs = "success:host_server=%s" % host_srv
        logger.info("start_vm: success. vm('%s') => %s" % (vm_path, rs ))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("start_vm: failed! vm('%s') %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def stop_vm(vm_path):
    """
    stop a vm

    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("stop_vm: vm('%s') start..." % vm_path)
        chk_vm_path(vm_path)

        #let DOWN in status_before, since it may out sync with ovs manager.
        #if it already DOWN, let it success anyway.
        db_chkset_vm_status(vm_path, ("DOWN", "RUNNING", "SUSPEND", "ERROR"), "PROCESSING")
        try:
            try:
                (host_srv, sp) = get_vm_host(vm_path)
            except Exception, e:
                logger.warning("no srv selected for vm('%s')" % vm_path)
                db_update_vm(vm_path, {"host_srv":"", "status": "DOWN", "manual_shutdown": True})
                return "success"
            vm = db_get_vm(vm_path)
            real_vm_path = db_get_real_vm_path(vm_path)
            rs = getattr(sp, "%s_stop_vm"% vm["vtech"])(real_vm_path)

            #check result & update db
            logger.info("stop_vm: vm('%s') on srv('%s') =>%s" % (vm_path, host_srv, rs))
            if rs.startswith("success"):
                db_update_vm(vm_path, {"host_srv":"", "status": "DOWN", "manual_shutdown": True})
            else:
                raise Exception(rs)
        except Exception, e:
            db_chkset_vm_status(vm_path, "", "ERROR")
            raise e

        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("stop_vm: vm('%s') failed=> %s" % (vm_path, errmsg))
        return "failed:%s" % (errmsg, )

@exposed
def reset_vm(vm_path):
    """
    reset a vm (reboot)

    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("reset_vm: vm('%s')" % vm_path)
        chk_vm_path(vm_path)

        db_chkset_vm_status(vm_path, ("ERROR", "RUNNING"), "PROCESSING")
        try:
            (host_srv, sp) = get_vm_host(vm_path)
            vm = db_get_vm(vm_path)
            real_vm_path = db_get_real_vm_path(vm_path)
            rs = getattr(sp, "%s_reboot_vm" % vm["vtech"])(real_vm_path)

            #check result & update db
            logger.info("reset_vm: vm('%s') on srv('%s') => %s" % (vm_path, host_srv, rs))
            if rs.startswith("success"):
                db_update_vm(vm_path, {"status": "RUNNING"})
            else:
                raise Exception(rs)
        except Exception, e:
            db_chkset_vm_status(vm_path, "", "ERROR")
            raise e

        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("reset_vm: failed. vm('%s') %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def pause_vm(vm_path):
    """
    pause a vm

    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("pause_vm: vm('%s') start..." % vm_path)
        chk_vm_path(vm_path)

        db_chkset_vm_status(vm_path, ("ERROR", "RUNNING"), "PROCESSING")
        try:
            (host_srv, sp) = get_vm_host(vm_path)
            vm = db_get_vm(vm_path)
            real_vm_path = db_get_real_vm_path(vm_path)
            rs = getattr(sp, "%s_pause_vm"% vm["vtech"])(real_vm_path)

            #check result & update db
            logger.info("pause_vm: vm('%s') on srv('%s') => %s" % (vm_path, host_srv, rs))
            if rs.startswith("success"):
                db_update_vm(vm_path, {"status": "SUSPEND"})
            else:
                raise Exception(rs)
        except Exception, e:
            db_chkset_vm_status(vm_path, "", "ERROR")
            raise e

        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("pause_vm: failed. vm('%s')  %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def unpause_vm(vm_path):
    """
    unpause a vm

    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("unpause_vm: vm('%s') start..." % vm_path)
        chk_vm_path(vm_path)

        db_chkset_vm_status(vm_path, ("ERROR", "SUSPEND"), "PROCESSING")
        try:
            (host_srv, sp) = get_vm_host(vm_path)
            vm = db_get_vm(vm_path)
            real_vm_path = db_get_real_vm_path(vm_path)

            rs = getattr(sp, "%s_unpause_vm"% vm["vtech"])(real_vm_path)

            #check result & update db
            logger.info("unpause_vm: vm('%s') on srv('%s') => %s" % (vm_path, host_srv, rs))
            if rs.startswith("success"):
                db_update_vm(vm_path, {"status": "RUNNING"})
            else:
                raise Exception(rs)
        except Exception, e:
            db_chkset_vm_status(vm_path, "", "ERROR")
            raise e

        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("unpause_vm: failed. vm('%s')  %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def save_vm(vm_path, save_file=None):
    """
    save a vm

    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("save_vm: vm('%s') start..." % vm_path)
        chk_vm_path(vm_path)
        db_chkset_vm_status(vm_path, "RUNNING", "PROCESSING")
        try:
            (host_srv, sp) = get_vm_host(vm_path)
            vm = db_get_vm(vm_path)
            real_vm_path = db_get_real_vm_path(vm_path)

            # check if there is enough space
            rs = utl_get_free_space(real_vm_path)
            if not rs.startswith("success"):
                raise Exception(rs)
            free_space = int(parse_result(rs)["free_space"])
            rs = xen_get_memory(real_vm_path)
            if not rs.startswith("success"):
                raise Exception(rs)
            memory = int(parse_result(rs)["memory"])
            if free_space < memory:
                raise Exception("There is not enough space.")

            if not save_file:
                save_file = join(real_vm_path, "vm.save")
            rs = getattr(sp, "%s_save_vm"% vm["vtech"])(real_vm_path, save_file)

            #check result & update db
            logger.info("save_vm: vm('%s') %s" % (vm_path, rs))
            if rs.startswith("success"):
                db_update_vm(vm_path, {"status": "DOWN", "manual_shutdown": True})
            else:
                raise Exception(rs)
        except Exception, e:
            db_chkset_vm_status(vm_path, "", "ERROR")
            raise e

        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("save_vm: failed. vm('%s')  %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def restore_vm(vm_path, dedicated_server=None, save_file=None):
    """
    restore a vm

    @return         "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("restore_vm: vm('%s') dedicated_server=%s start..." % (vm_path, repr(dedicated_server)))
        chk_vm_path(vm_path)

        db_chkset_vm_status(vm_path, "DOWN", "PROCESSING")
        try:
            if dedicated_server:
                host_srv = dedicated_server
                srv = db_get_srv(host_srv)
                sp = ServerProxy(srv['agt_url'])
            else:
                (host_srv, sp) = select_tgt_host(vm_path)

            vm = db_get_vm(vm_path)
            real_vm_path = db_get_real_vm_path(vm_path)
            if not save_file:
                save_file = join(real_vm_path, "vm.save")
            else:
                save_file = join(real_vm_path, save_file)

            _check_free_mem(sp, vm_path)

            rs = getattr(sp, "%s_restore_vm"% vm["vtech"])(save_file)

            #check result & update db
            logger.info("restore_vm: vm('%s') from '%s' on srv('%s') => %s" % \
                    (vm_path, save_file, host_srv ,rs))
            if rs.startswith("success"):
                db_update_vm(vm_path, {"status": "RUNNING", "host_srv": host_srv, "manual_shutdown": False})
            else:
                raise Exception(rs)

        except OVSException, e:
            from OVSExceptionLookup import ERRGRP_SERVER_SELECTION
            if e.iskindof(ERRGRP_SERVER_SELECTION):
                db_chkset_vm_status(vm_path, "", "DOWN")
            else:
                db_chkset_vm_status(vm_path, "", "ERROR")
            raise e
        except Exception, e:
            db_chkset_vm_status(vm_path, "", "ERROR")
            raise e

        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("restore_vm: failed. vm('%s')  %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def select_migration_target(vm_path):
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("select_migration_target: vm('%s') start..." % vm_path)
        chk_vm_path(vm_path)

        vm = db_get_vm(vm_path)
        srv = db_get_srv(vm['host_srv'])
        sp = ServerProxy(srv['agt_url'])

        (tgt_host, tgt_sp) = select_tgt_host(vm_path,
                            any = False,
                            excludes = (vm['host_srv'], ),
                            with_attr = {"arch": srv["arch"]})

        logger.info("select_migration_target: vm('%s') from '%s'->'%s'" % \
                    (vm_path, vm["host_srv"], tgt_host))
        return "success:target_host=%s" % tgt_host
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("select_migration_target: failed. vm('%s')  %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg


@exposed
def migrate_vm(vm_path, tgt_host=None, is_live=True):
    """
    migrate vm

    @return         "success" if success else "failed:<error message>
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("migrate_vm: vm('%s') start..." % vm_path)
        chk_vm_path(vm_path)

        #a vm with status "ERROR" is also allowed to be migrated
        db_chkset_vm_status(vm_path, ("RUNNING", "ERROR"), "PROCESSING")
        try:
            vm = db_get_vm(vm_path)
            srv = db_get_srv(vm['host_srv'])
            sp = ServerProxy(srv['agt_url'])

            if not tgt_host:
                (tgt_host, tgt_sp) = select_tgt_host(vm_path,
                                    any = False,
                                    excludes = (vm['host_srv'], ),
                                    with_attr = {"arch": srv["arch"]})
            else:
                tgt_srv = db_get_srv(tgt_host)
                tgt_sp = ServerProxy(tgt_srv['agt_url'])

            #make sure the tgt host share the same view with current host server,
            rs = sp.utl_find_real_path(vm_path)
            if not rs.startswith("success"):
                raise Exception("host_srv=%s =>%s" % (vm['host_srv'], rs))
            rs = parse_result(rs)
            real_vm_path = rs["real_path"]

            rs = tgt_sp.utl_find_real_path(vm_path)
            if not rs.startswith("success"):
                raise Exception("tgt_host=%s =>%s" % (tgt_host, rs))
            rs = parse_result(rs)
            tgt_real_vm_path = rs["real_path"]

            if real_vm_path != tgt_real_vm_path:
                raise Exception("%s' != '%s'" % (real_vm_path, tgt_real_vm_path))

            #migrate finally
            rs = getattr(sp, "%s_migrate_vm" % vm["vtech"])(real_vm_path, tgt_host, is_live)

            #check result & update db
            logger.info("migrate_vm: vm('%s') from '%s'->'%s' => %s" % \
                    (vm_path, vm["host_srv"], tgt_host ,rs))
            if rs.startswith("success"):
                db_update_vm(vm_path, {"host_srv": tgt_host, "status": "RUNNING"})
            else:
                raise Exception(rs)
        except OVSException, e:
            from OVSExceptionLookup import ERRGRP_SERVER_SELECTION
            if e.iskindof(ERRGRP_SERVER_SELECTION):
                db_chkset_vm_status(vm_path, "", "DOWN")
            else:
                db_chkset_vm_status(vm_path, "", "ERROR")
            raise e
        except Exception, e:
            db_chkset_vm_status(vm_path, "", "ERROR")
            raise e

        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("migrate_vm: failed. vm('%s')  %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg


from OVSImageOperation import createConfigurations, copyConfigIntoImage, cleanConfigurations
CONFIG_CHROOT = "config-chroot"

@exposed
def config_vm(vm_path, property_string):
    """
    Config a guest VM according to the property string. Take the first disk in vm.cfg as the target image.

    @param property_string:   an example, "NIC={eth0={CONFIGURATION_TYPE=DYNAMIC}, eth1={CONFIGURATION_TYPE=STATIC, SUBNET_MASK=255.255.254.0, DEFAULT_GATEWAY=10.182.120.1, IP_ADDRESS=10.182.120.244, HOST_NAME=ovs-w, DNS_SERVER=10.182.122.37/10.182.244.34, DOMAIN_NAME=cn.oracle.com}}, USER_PASSWORD={USER=root, PASSWORD=oracle}"
    @return:    "success" if success else "failed: <error msg>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        db_chkset_vm_status(vm_path, ("DOWN"), "PROCESSING")
        rs = get_vm_vdisks(vm_path)
        if not rs.startswith("success"): raise Exception(rs)
        vdisks = parse_result(rs)["vdisks"].split(",")
        if vdisks:
            vdisk = vdisks[0].split(":")[0]
        else:
            raise Exception("No image found.")
        real_vm_path = db_get_real_vm_path(vm_path)
        image = join(real_vm_path, vdisk)

        config_path = join(real_vm_path, CONFIG_CHROOT)
        val = createConfigurations(property_string, config_path)
        if not val:
            raise Exception("Error creating configurations.")

        val = copyConfigIntoImage(config_path, image)
        if not val:
            raise Exception("Error updating the image.")

        val = cleanConfigurations(config_path)
        if not val:
            raise Exception("Error cleaning configurations.")

        db_update_vm(vm_path, {"status": "DOWN"})
        logger.info("config_vm: success. vm_path('%s') => %s" % (vm_path, val))
        return "success"
    except Exception, e:
        err_msg = fmt_err_msg(e)
        db_update_vm(vm_path, {"status": "DOWN"})
        logger.info("config_vm: failed. vm_path('%s') => %s" % (vm_path, err_msg))
        return "failed: %s" % err_msg


from OVSImageOperation import createPropFile
PROP_CHROOT = "prop_chroot"

@exposed
def copy_props_into_vm(vm_path, props, prop_file):
    """
    Generate a property file and copy it into the VM image. Take the first disk in vm.cfg as the target image.

    @param props:   for example, "k1=v1;k2=v2;k3=v3"
    @param prop_file: absolute path of the property file in the VM, e.g. "/etc/em-agent.conf"
    @return:    "success" if success else "failed: <error msg>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        db_chkset_vm_status(vm_path, ("DOWN"), "PROCESSING")

        rs = get_vm_vdisks(vm_path)
        if not rs.startswith("success"): raise Exception(rs)
        vdisks = parse_result(rs)["vdisks"].split(",")
        if vdisks:
            vdisk = vdisks[0].split(":")[0]
        else:
            raise Exception("No image found.")
        real_vm_path = db_get_real_vm_path(vm_path)
        image = join(real_vm_path, vdisk)

        prop_path = join(real_vm_path, PROP_CHROOT)
        _prop_file = join(prop_path, prop_file.strip("/"))

        val = createPropFile(props, _prop_file)
        if not val:
            raise Exception("Error creating the property file.")

        val = copyConfigIntoImage(prop_path, image)
        if not val:
            raise Exception("Error updating the image.")

        val = cleanConfigurations(prop_path)
        if not val:
            raise Exception("Error cleaning the property file.")

        db_update_vm(vm_path, {"status": "DOWN"})
        logger.info("copy_props_into_vm: success. vm_path('%s'), props('%s'), prop_file('%s') => %s" % (vm_path, props, prop_file, val))
        return "success"
    except Exception, e:
        db_update_vm(vm_path, {"status": "DOWN"})
        err_msg = fmt_err_msg(e)
        logger.info("copy_props_into_vm: failed. vm_path('%s'), props('%s'), prop_file('%s') => %s" % (vm_path, props, prop_file, err_msg))
        return "failed: %s" % err_msg

