#!/usr/bin/env python
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.


import os.path
import string
from xen.xend.sxp import Parser
from xen.xm.XenAPI import Session


#constants

UNKNOWN = "UNKNOWN"
XEN_API_URL = "httpu:///var/run/xend/xen-api.sock"
_g_session = None
_g_session_ref_count = 0


#callable class with partially-applied arguments
#http://www.python.org/dev/peps/pep-0309/

class partial(object):
    def __init__(self, *args, **kw):
        self.fn, self.args, self.kw = (args[0], args[1:], kw)

    def __call__(self, *args, **kw):
        if kw and self.kw:
            d = self.kw.copy()
            d.update(kw)
        else:
            d = kw or self.kw
        return self.fn(*(self.args + args), **d)


class XenAPIObject(object):
    def __init__(self, klass, session, ref):
        self.klass, self.session, self.ref = klass, session, ref

    def __getattr__(self, attr):
        klass = getattr(self.session.xenapi, self.klass)
        return partial(getattr(klass, attr), self.ref)


def get_session():
    """Start a new XenAPI session."""
    global _g_session, _g_session_ref_count
    if not _g_session:
        _g_session=Session(XEN_API_URL)
        _g_session.login_with_password('', '')

    _g_session_ref_count += 1
    return _g_session


def drop_session():
    global _g_session, _g_session_ref_count
    _g_session_ref_count -= 1
    if _g_session_ref_count == 0:
        try:
            _g_session.xenapi.session.logout()
        except:
            pass
        _g_session = None


#alias
session_login = get_session
session_logout = drop_session


def is_vt_enabled():
    """Determine if a machine is VT enabled or not."""

    filename = "/sys/hypervisor/properties/capabilities"
    caps = ""
    if os.path.exists(filename):
        caps = open(filename).read()
    if caps.find("hvm") != -1:
        return True
    return False


def get_network_type():
    """Get the network type of a machine."""

    filename = "/etc/xen/xend-config.sxp"
    buf = ""
    if os.path.exists(filename):
        buf = open(filename).read()
    pin = Parser()
    pin.input(buf)
    d = {}
    while pin.ready():
        val = pin.get_val()
        d[val[0]] = val[1]
    if "network-script" in d and "vif-script" in d:
        if d["network-script"] == "network-bridges" and d["vif-script"] == "vif-bridge":
            return "bridge"
        elif d["network-script"] == "network-nat" and d["vif-script"] == "vif-nat":
            return "nat"
    return UNKNOWN


def get_os_version():
    """Get OS release version."""

    filename = "/etc/ovs-info"
    lines = []
    if os.path.exists(filename):
        lines = open(filename).read().splitlines()
    d = {}
    for line in lines:
        parts = map(string.strip, line.split(":"))
        if len(parts) == 2:
            d[parts[0]] = parts[1]
    if "ovs-release" in d:
        return d["ovs-release"]
    return UNKNOWN


def get_cpu_info():
    """Get CPU information."""
    filename = "/proc/cpuinfo"
    lines = []
    if os.path.exists(filename):
        lines = open(filename).read().splitlines()
    d = {}
    for line in lines:
        parts = map(string.strip, line.split(":"))
        if len(parts) == 2:
            d[parts[0]] = parts[1]
    return d


def is_cpu_compatible(cpu1, cpu2):
    """Determine if cpu1 is compatible with cpu2 for vm live migration."""
    #same vendor and cpu family
    if cpu1["vendor_id"] != cpu2["vendor_id"]: return False
    if cpu1["cpu family"] != cpu2["cpu family"]: return False
    #instruction set compatibility, only support Intel and AMD processors
    vendor = cpu1["vendor_id"]
    if vendor == "GenuineIntel":
        special_flags = ["xd", "sse3", "ssse3", "sse4.1"]
    elif vendor == "AuthenticAMD":
        special_flags = ["nx", "sse3", "ffxsr", "rdtscp", "cmpxch16b", "3dnprefetch"]
    else:
        return False
    flags1 = cpu1["flags"].split()
    flags2 = cpu2["flags"].split()
    diff1 = set(special_flags) - set(flags1)
    diff2 = set(special_flags) - set(flags2)
    if diff1 != diff2:
        return False
    return True


def get_dom0_caps():
    """Get Dom0 capabilities."""
    session = get_session()
    ref = session.xenapi.host.get_all()
    if len(ref) != 1: return UNKNOWN
    caps = session.xenapi.host.get_capabilities(ref[0])
    drop_session()
    return caps


def get_memory_free():
    """Get free memory in bytes."""
    session = get_session()
    ref = session.xenapi.host_metrics.get_all()
    if len(ref) != 1: return UNKNOWN
    memory_free = session.xenapi.host_metrics.get_memory_free(ref[0])
    drop_session()
    return int(memory_free)


def get_memory_actual(name_label):
    """Get actual memory in bytes."""
    session = get_session()
    ref = session.xenapi.VM.get_by_name_label(name_label)
    metrics = session.xenapi.VM.get_metrics(ref[0])
    memory_actual = session.xenapi.VM_metrics.get_memory_actual(metrics)
    drop_session()
    return int(memory_actual)


def get_running_vms():
    session = session_login()
    ref = session.xenapi.VM.get_all()
    l = []
    for i in ref:
        name_label = session.xenapi.VM.get_name_label(i)
        uuid = session.xenapi.VM.get_uuid(i)
        if name_label != "Domain-0":
            l.append((name_label, uuid))
    drop_session()
    return l
