#!/usr/bin/env python
#^#
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import pickle
import os
import logging
import shutil
from os.path import islink, ismount, realpath, exists, join

from OVSConstants import *
from OVSCommons import *
from OVSCDS import *

@interconnect
def clusterm_check_root_ready():
    """Check whether the cluster root is ready.

    Cluster root ready :=
            (/OVS is a link) and
            (repository metadata file exists in /OVS)
    """
    logger = logging.getLogger("ovs.query")
    try:
        if not islink(OVS_BASE):
            raise OVSException(ERR_OVS_NOT_LINK)
        elif not exists(join(OVS_BASE, SR_REPO_META)):
            raise OVSException(ERR_SR_NO_REPO_META, argd={"path": realpath(OVS_BASE)})
        #elif not ismount(realpath(OVS_BASE)):
        #   raise OVSException(ERR_CLUSTER_ROOT_NOT_MOUNT,
        #        argd={"path": realpath(OVS_BASE)})

        rs = SUCC()
        logger.info("clusterm_check_root_ready: %s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("clusterm_check_root_ready: %s" % (errmsg,))
        return FAIL(exception=e)

@interconnect
def clusterm_initialize_sr(sr_uuid=""):
    """Mount all FileSR
    process:
        1) check whether root sr is mounted.
        2) for sr in db: sr.initialize()
    """
    logger = logging.getLogger("ovs.operation")
    try:
        rs = clusterm_check_root_ready()
        DJSON(rs)
        if rs["status"] != OP_STATUS_SUCC:
            raise OVSException(error=rs["error"])

        #load sr
        d = {}
        cds = CDS("sr")
        try:
            if sr_uuid:
                sr = cds.db_sr.get(sr_uuid)
                if not sr:
                    raise OVSException(ERR_DB_SR_NOT_FOUND, argd={"sr_uuid": sr_uuid})
            else:
                d.upate(cds.db_sr)
        finally:
            cds.close()

        #initialize
        if sr_uuid:
            sr.initialize()
        else:
            for sr_uuid, sr in d.items():
                sr.initialize()

        rs = SUCC()
        logger.info("clusterm_initialize_sr:%s%s" % (
                sr_uuid and "sr_uuid=%s " % sr_uuid or " ", rs))

        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("clusterm_initialize_sr: %s" % errmsg)
        return FAIL(exception=e)

@interconnect
def clusterm_init_root_sr(sro=None):
    """Setup cluster root"""
    logger = logging.getLogger("ovs.operation")
    try:
        if sro:
            sr = pickle.loads(sro)
            db_save("LOCAL_DB,node", "cluster_root_sr", sr)
        else:
            try:
                sr = db_load("LOCAL_DB,node", "cluster_root_sr")
            except Exception, e:
                errmsg = fmt_err_msg(e)
                return FAIL(exception=e)

        sr.initialize()
        root_dir = sr.mountpoint

        if not islink(OVS_BASE):
            if ismount(OVS_BASE):
                raise OVSException(ERR_OVS_IS_MOUNTPOINT)
            else:
                if exists(OVS_BASE):
                    backup_path = ovs_backup(OVS_BASE, mode="move")
                    logger.info("clusterm_init_root_sr: backup '%s' to '%s'" % (
                            OVS_BASE, backup_path))
            #now OVS_BASE not exists
            cmd = ['ln', '-sf', root_dir, OVS_BASE]
            run_cmd(cmd)
        else:
            if not realpath(OVS_BASE) == realpath(root_dir):
                os.remove(OVS_BASE)
                cmd = ['ln', '-sf', root_dir, OVS_BASE]
                run_cmd(cmd)

        db_save("LOCAL_DB,node", "local_db", False)
        db_save("cluster", "cluster_root_sr_uuid", format_uuid(sr._uuid))

        rs = {"cluster_root_dir": root_dir}
        rs = SUCC(rs)
        logger.info("clusterm_init_root_sr: %s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("clusterm_init_root_sr: %s" % errmsg)
        return FAIL(exception=e)

@interconnect
def clusterm_init_sr(construct_from_local=False):
    """Initialize SR

    @param sr_uuid:
        if not sr_uuid: initialize all sr that can be found in cluster db.
        else: initialize sr identified by sr_uuid.
    """
    logger = logging.getLogger("ovs.operation")
    try:
        # read from cluster root
        if construct_from_local:
            sr_db = "LOCAL_DB,sr"
        else:
            sr_db = "sr"

        d = db_dump(sr_db)
        for sr_uuid, sr in d.items():
            sr.initialize()

        rs = SUCC()
        logger.info("clusterm_init_sr: success.")
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("clusterm_init_sr: %s" % ( errmsg,))
        return FAIL(exception=e)

@interconnect
def clusterm_set_ocfs2_cluster_conf(cluster_conf):
    """
    @param cluster_conf: <string of %s>
    """ % OCFS2_CLUSTER_CONF
    logger = logging.getLogger("ovs.operation")
    try:
        ocfs2_conf_dir = "/etc/ocfs2"
        if not exists(ocfs2_conf_dir):
            os.mkdir(ocfs2_conf_dir)
        f = open(OCFS2_CLUSTER_CONF, "w")
        try:
            f.write(cluster_conf)
        finally:
            f.close()

        rs = SUCC()
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("clusterm_set_ocfs2_cluster_conf: cluster_conf=\n%s\n=>%s" % (
                cluster_conf, errmsg))
        return FAIL(exception=e)

@interconnect
def clusterm_start_o2cb_service():
    """Start/Reload o2cb service to resync with the config in %s
    """ % OCFS2_CLUSTER_CONF
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("clusterm_start_o2cb_service: reloading config from %s ..." % (
                OCFS2_CLUSTER_CONF,))

        cmd = [join(AGENT_DIR, "utils/config_o2cb.sh")]
        run_cmd(cmd)

        rs = SUCC()
        logger.info("clusterm_start_o2cb_service: => %s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("clusterm_set_ocfs2_cluster_conf: cluster_conf=\n%s\n=>%s" % (
                OCFS2_CLUSTER_CONF, errmsg))
        return FAIL(exception=e)

@interconnect
def clusterm_start_heartbeat(sr_uuid=""):
    """Start heartbeat against this sr.

    NOTE:
        1) If the sr is on ocfs2, mount the sr.
        2) If the sr is on NFS,
            a) create heartbeat file if not already there.
            b) create heartbeat device using dmnfs.
            c) start heartbeat agaist the device.
    """
    logger = logging.getLogger("ovs.operation")
    try:
        if not sr_uuid:
            sr_uuid = db_load("cluster", "cluster_root_sr_uuid")
        sr = db_load("sr", sr_uuid)
        sr.initialize()

        if sr.sp.fs_type != "ocfs2.cluster":
            #heartbeat agaist the heartbeat device.
            hb_file = join(sr.mountpoint, SR_HEARTBEAT_FILE)
            if not exists(hb_file):
                raise OVSException(ERR_SR_NO_HEARTBEAT_FILE, argd={"hb_file": hb_file})
            else:
                try:
                    cmd = [join(AGENT_DIR, "utils/do_heartbeat_nfs.sh"), '--action', 'start', '--heartbeatfile', hb_file]
                    run_cmd(cmd)
                except:
                    logger.debug("clusterm_start_heartbeat: stop heartbeat.")
                    cmd = [join(AGENT_DIR, "utils/do_heartbeat_nfs.sh"), '--action', 'stop', '--heartbeatfile', hb_file]
                    run_cmd(cmd)
                    logger.debug("clusterm_start_heartbeat: start heartbeat.")
                    cmd = [join(AGENT_DIR, "utils/do_heartbeat_nfs.sh"), '--action', 'start', '--heartbeatfile', hb_file]
                    run_cmd(cmd)

        rs = SUCC()
        logger.info("clusterm_start_heartbeat: sr('%s') => %s" % (sr_uuid, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("clusterm_start_heartbeat: sr('%s') => %s" % (sr_uuid, errmsg))
        return FAIL(exception=e)

@interconnect
def clusterm_stop_heartbeat(sr_uuid=""):
    """Stop heartbeat against this sr.

    NOTE:
        1) If the sr is on ocfs2, do NOTHING, will not umount it.
        2) If the sr is on NFS,
            a) start heartbeat agaist the device.
            b) remove the heartbeat device.
            c) will NOT remove the heartbeat file.
    """
    logger = logging.getLogger("ovs.operation")
    try:
        if not sr_uuid:
            sr_uuid = db_load("cluster", "cluster_root_sr_uuid", get=True)
        if sr_uuid:
            sr = db_load("sr", sr_uuid)
            sr.initialize()

            if sr.sp.fs_type != "ocfs2.cluster":
                #heartbeat agaist the heartbeat device.
                hb_file = join(sr.mountpoint, SR_HEARTBEAT_FILE)
                if exists(hb_file):
                    cmd = [join(AGENT_DIR, "utils/do_heartbeat_nfs.sh"), '--action', 'stop', '--heartbeatfile', hb_file]
                    run_cmd(cmd)
        else:
            sr_uuid = ''
            logger.debug("clusterm_stop_heartbeat: cluster root sr not found.")

        rs = SUCC()
        logger.info("clusterm_stop_heartbeat: sr('%s') => %s" % (sr_uuid,rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("clusterm_stop_heartbeat: sr('%s') => %s" % (sr_uuid, errmsg))
        return FAIL(exception=e)


@interconnect
def clusterm_status_heartbeat(sr_uuid="", keep_silence=False):
    """Check heartbeat against this sr.

    NOTE:
        1) If the sr is on ocfs2, do NOTHING, will not umount it.
        2) If the sr is on NFS,
            a) heartbeat agaist the device.
            b) remove the heartbeat device.
            c) will NOT remove the heartbeat file.
    """
    logger = logging.getLogger("ovs.operation")
    try:
        if not sr_uuid:
            sr_uuid = db_load("cluster", "cluster_root_sr_uuid")
        sr = db_load("sr", sr_uuid)
        sr.initialize()

        rs = SUCC(HEARTBEAT_STATUS_INACTIVE)
        if sr.sp.fs_type != "ocfs2.cluster":
            #heartbeat agaist the heartbeat device.
            hb_file = join(sr.mountpoint, SR_HEARTBEAT_FILE)
            if exists(hb_file):
                cmd = [join(AGENT_DIR, "utils/do_heartbeat_nfs.sh"), '--action', 'status', '--heartbeatfile', hb_file]
                try:
                    run_cmd(cmd)
                    rs = SUCC(HEARTBEAT_STATUS_ACTIVE)
                except:
                    pass
        elif sr.sp.fs_type == "ocfs2.cluster":
            rs = SUCC(HEARTBEAT_STATUS_ACTIVE)

        if not keep_silence:
            logger.info("clusterm_status_heartbeat: sr('%s') => %s" % (sr_uuid,rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("clusterm_status_heartbeat: sr('%s') => %s" % (sr_uuid, errmsg))
        return FAIL(exception=e)

def _check(ret):
    ret = DJSON(JSON(ret))
    if ret["status"] == OP_STATUS_FAIL:
        raise OVSException(error=ret["error"])

@interconnect
def clusterm_teardown():
    logger = logging.getLogger("ovs.operation")
    try:
        # stop heartbeat
        ret = clusterm_stop_heartbeat()
        _check(ret)

        # umount
        mount_dir = "/var/ovs/mount"
        cmd = "umount -fl %s/*" % mount_dir
        try:
            old_run_cmd(cmd)
        except Exception, e:
            pass

        # offline o2cb
        cmd = "/etc/init.d/o2cb offline"
        try:
            old_run_cmd(cmd)
        except Exception, e:
            pass

        if exists(mount_dir):
            ovs_backup(mount_dir)
        if exists(OVS_BASE):
            ovs_backup(OVS_BASE)
        elif islink(OVS_BASE):
            os.remove(OVS_BASE)

        cmdlist = [
            "rm -fr %s/db" % CONFIG_DIR,
            "mkdir -p %s" % OVS_BASE,
            "ln -sf %s %s" % (CONFIG_DIR, CONFIG_DIR_CLUSTER),
            "rm -f %s" % OCFS2_CLUSTER_CONF
        ]

        for cmd in cmdlist:
            old_run_cmd(cmd)

        db_save("LOCAL_DB,node", "local_db", True)

        logger.info("clusterm_teardown: success.")
        return SUCC()
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("clusterm_teardown: failed. => %s" % errmsg)
        return FAIL(exception=e)
