#!/usr/bin/env python

import string
import logging

from OVSXHA import ha_join_dlm_domain
from OVSXCluster import clusterm_status_heartbeat
from OVSProtocol import *
from OVSMultiprocess import *
from OVSExceptionLookup import *
from OVSException import *
from OVSConstants import *
from OVSCommons import *

#----------------------------------------------------------------------
#local function

ETXTBUSY = 26
def acquire_master_dlm_lock():
    rs = clusterm_status_heartbeat(keep_silence=True)
    status = GETVAL(rs)
    if status != HEARTBEAT_STATUS_ACTIVE:
        raise OVSException(ERR_NO_HEARTBEAT)

    rs = ha_join_dlm_domain(DLM_DOMAIN, keep_silence=True)
    if not rs.startswith("success"):
        raise Exception(rs)

    #"""
    # close() does not downconvert the lock level nor does it drop the lock. The
    # holder still owns the lock at that level. All close() does is allows any
    # downconvert request to succeed.
    # However, a downconvert request is only generated for queued requests. And
    # O_NONBLOCK is specifically a noqueue dlm request.
    #"""

    #here is the situation,
    # 1) O_CREAT | O_NONBLOCK will create a lock file if it does not exist, whether
    #    we are the lock holder or not.
    # 2) if we hold O_RDWR lock, and we close but not delete it, we still hold it.
    #    afterward, O_RDWR will succeed, but O_RDWR | O_NONBLOCK will not.
    # 3) but if we donnot hold the lock, O_RDWR will hang there waiting,
    #    which is not desirable -- any uninterruptable hang is undesirable.
    # 4) if noboday else hold the lock either, but the lock file exists as side effect
    #    of 1), with O_NONBLOCK, it may result in ETXTBUSY
    #so,
    # a) we need O_NONBLOCK to avoid 3)
    # b) we need to delete it ourself to avoid 2) (for e.g. agtctl restart), but when?
    #   *) if not succeed in 1), remove the lock file (side product), avoid 4)
    #   *) if everything is good, we drop it and we clean out.
    #   *) if stopped with agtctl (killed), agtctl should remove the lock.
    #   *) if crashed, agtctl stop --force should clean out the env.

    try:
        if not exists(DLM_LOCK_MASTER):
            #logger.info("acquire_master_dlm_lock(): touch lock '%s'." % DLM_LOCK_MASTER)
            fd = os.open(DLM_LOCK_MASTER, os.O_CREAT | os.O_NONBLOCK)
            os.close(fd)

        #logger.info("acquire_master_dlm_lock(): create exclusive lock '%s'." % DLM_LOCK_MASTER)
        master_lock = os.open(DLM_LOCK_MASTER, os.O_RDWR | os.O_NONBLOCK)
        #so if we've reached here without error, we've got the lock. close it and hold it.
        os.close(master_lock)

        return True
    except OSError, e:
        if e.errno == ETXTBUSY:
            # if we donnot remove the lock file (the side effect of touch),
            # we may never succeed to hold the exclusive lock, i.e. 4)
            if exists(DLM_LOCK_MASTER):
                os.remove(DLM_LOCK_MASTER)
        else:
            raise e

def release_master_dlm_lock():
    if exists(DLM_LOCK_MASTER):
        os.remove(DLM_LOCK_MASTER)

#----------------------------------------------------------------------
def _get_vip_dev(vip):
    """Get dev the vip is bound to.
    """
    cmd = "ip addr show to %s | awk '/scope global/ {print $NF;}'" % vip
    output = old_run_cmd(cmd)
    dev = output.strip()
    return dev

def ifconfig_vip_up(vip, ip):
    """Bind vip to dev that ip is assigned to.
    """
    dev = _get_vip_dev(vip)
    if dev: #already bound
        return
    cmd = "ip addr show to %s" % ip
    output = old_run_cmd(cmd)
    lines = output.strip().splitlines()
    if len(lines) < 2:
        raise Exception("IP %s is not up." % ip)
    parts = lines[-1].split()
    dev = parts[-1]
    netmask = parts[1].split("/")[-1]

    dev_ip = dev
    if ":" in dev:
        dev_ip = dev.split(":")[0]

    cmd = "ifconfig | awk '/Link encap/ {print $1}'"
    output = old_run_cmd(cmd)
    dev_list = output.splitlines()

    idx = 0
    while True:
        dev_vip = "%s:%s" % (dev_ip, idx)
        if dev_vip not in dev_list:
            break
        else:
            idx += 1

    cmd = "ifconfig %s %s/%s" % (dev_vip, vip, netmask)
    old_run_cmd(cmd)

    #refresh switch arp table
    #http://wiki.wireshark.org/Gratuitous_ARP
    cmd = "arping -q -U -c 2 -I %s %s" % (dev_ip, vip)
    old_run_cmd(cmd)
    cmd = "arping -q -A -c 2 -I %s %s" % (dev_ip, vip)
    old_run_cmd(cmd)

def ifconfig_vip_down(vip):
    """Release vip
    """
    dev = _get_vip_dev(vip)
    if dev:
        cmd = "ifconfig %s down" % dev
        old_run_cmd(cmd)

def i_hold_the_vip(vip):
    """Check whether vip is bounded this server.

    @param vip:
    @return: True if i hold the vip, else False
    """
    dev = _get_vip_dev(vip)

    return dev and True or False

def i_am_the_master():
    """Check whether I am currently the master server.

    @return: True if i am the master or False
    """
    registered_hostname = db_load("LOCAL_DB,node", "registered_hostname", get=True)
    if not registered_hostname:
        return False
    master_hostname = db_load("cluster", "master_hostname", get=True)
    return registered_hostname == master_hostname

#----------------------------------------------------------------------
#@interconnect functions
@interconnect
def clusterm_setup_master_env():
    """Setup master environment

    NOTE:
        the process:
        1) setup master vip.
        2) store master info to master db.

        *) it will NOT try to acquire master dlm lock, which is necessary for
    cooperative reelection of the new master.
    """
    logger = logging.getLogger("os.operation")
    try:
        vip = db_load("cluster", "master_vip")
        ip = db_load("LOCAL_DB,node", "registered_ip")
        ifconfig_vip_up(vip, ip)

        #save master info
        registered_hostname = db_load("LOCAL_DB,node", "registered_hostname")
        srv = db_load("srv", registered_hostname)
        master_agent_url = srv["agt_url"]
        registered_ip = db_load("LOCAL_DB,node", "registered_ip")
        cds = CDS("cluster")
        try:
            cds.db_cluster["master_agt_url"] = master_agent_url
            cds.db_cluster["master_hostname"] = registered_hostname
            cds.db_cluster["master_ip"] = registered_ip
        finally:
            cds.close()

        cds = CDS("srv")
        try:
            for k, v in cds.db_srv.items():
                srv_role = v["srv_role"].split(",")
                if k == registered_hostname:
                    if not "site" in srv_role:
                        srv_role.append("site")
                        v["srv_role"] = ",".join(srv_role)
                        cds.db_srv[k] = v
                else:
                    if "site" in srv_role:
                        srv_role.remove("site")
                        v["srv_role"] = ",".join(srv_role)
                        cds.db_srv[k] = v
        finally:
            cds.close()

        rs = SUCC()
        logger.info("clusterm_setup_master_env: => %s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("clusterm_setup_master_env: => %s" % errmsg)
        return FAIL(exception=e)


@interconnect
def clusterm_cleanup_master_env():
    """Cleanup master environment
    NOTE:
        1) release the master_vip
        2) release master dlm lock

        *) agtctl stop should call this function to step down
           from the master position.
    """
    logger = logging.getLogger("ovs.operation")
    try:
        vip = db_load("cluster", "master_vip", get=True)
        if vip:
            ifconfig_vip_down(vip)

        release_master_dlm_lock()

        rs = SUCC()
        logger.info("clusterm_cleanup_master_env: => %s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("clusterm_cleanup_master_env: => %s" % errmsg)
        return FAIL(exception=e)

#----------------------------------------------------------------------
#@exposed functions

@exposed
def cluster_get_next_master(exclude_hostnames=""):
    """Get next master server candidate according to rules.

    @param exclude_hostnames: "," separated list of hostname.
    @return: hostname of the master candidate.
    """
    logger = logging.getLogger("ovs.operation")
    try:
        exclude_list = []
        if exclude_hostnames:
            exclude_list = map(string.strip, exclude_hostnames.split(","))
        l = get_server_list_with_conn_err()
        exclude_list.extend(l)
        l = get_server_list_with_local_conf()
        exclude_list.extend(l)

        db = db_dump("srv")
        l = []
        for hostname, srv in db.items():
            if hostname in exclude_list:
                continue
            weight = 0
            if srv.get("master_weight"):
                weight = int(srv["master_weight"])
            l.append((weight, hostname))

        if not l:
            raise OVSException(ERR_NO_MASTER_CANDIDATE)

        l.sort()
        next_master = l[-1][1]

        rs = SUCC(next_master)
        logger.info("cluster_get_next_master: => %s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("cluster_get_next_master: => %s" % errmsg)
        return FAIL(exception=e)

