#!/usr/bin/env python

#^#
# OVSXMonitor.py
#
# ovs module, provides monitoring fuction, gether system information,
# performance data, etc
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import socket
import logging
from time import mktime, gmtime, strptime
from os.path import basename, join

import OVSXAPIUtil
from OVSXXenStore import xen_get_caps
from OVSXAPIUtil import XenAPIObject, session_login, session_logout
from OVSTimer import Timer
from OVSConstants import OVS_BASE
from OVSCommons import fmt_err_msg, run_cmd, old_run_cmd, exposed, parse_result, marshall_string, timeout_command

from OVSWrappers import *

from ServerModeHelper import _is_maintenance_mode

@exposed
def sys_vmstat_info():
    """
    system stat data by vmstat
    """
    logger = logging.getLogger("ovs.performance")
    try:
        #stdout = run_cmd("vmstat -n -S M")
        cmdlist = ["vmstat", "-n", "-S", "M"]
        stdout = run_cmd(args=cmdlist)
        lines = stdout.splitlines()
        keys = lines[1].split()
        values = lines[2].split()
        d = dict(zip(keys, values))

        cpu_idle, mem_free = ("%.2f" % (float(d["id"])/100,),
                "%d" % (int(d["free"]) + int(d["buff"]) + int(d["cache"])))

        if _is_maintenance_mode():
            D('maintain mode')
            mem_free = 0

        logger.info("sys_vmstat_info:cpu_idle=%s;mem_free=%s" % (cpu_idle, mem_free))
        return "success:cpu_idle=%s;mem_free=%s" % (cpu_idle, mem_free)
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("sys_vmstat_info: failed! => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def sys_perf_info(host_type="srv", df_path=OVS_BASE):
    """
    cpu_usage, mem_usage, hd_usage,
    hd_io (TODO), network_io (TODO)
    called by performance collector

    @param host_type: if is srv, then try xm info to get free memory
    """
    logger = logging.getLogger("ovs.performance")
    try:
        #stdout = run_cmd("vmstat -n -S M")
        cmdlist = ["vmstat", "-n", "-S", "M"]
        stdout = run_cmd(args=cmdlist)
        lines = stdout.splitlines()
        keys = lines[1].split()
        values = lines[2].split()
        d = dict(zip(keys, values))
        cpu_idle, mem_free = ("%.2f" % (float(d["id"])/100,),
                "%d" % (int(d["free"]) + int(d["buff"]) + int(d["cache"])))

        if host_type == "srv":
            #xm info
            #mem_free from xend view  if available will be preferred
            try:
                cmd = "xm info | grep free_memory"
                stdout = timeout_command(cmd)
                mem_free_xen = int(stdout.split(":")[1].strip())
                #use the bigger mem_free
                if mem_free_xen > int(mem_free):
                    mem_free = "%d" % mem_free_xen
            except Exception, e:
                errmsg = fmt_err_msg(e)
                logger.info("sys_perf_info: xm info=>%s" % errmsg)
        #df
        #stdout = run_cmd("df -B M %s" % df_path)
        cmdlist = ["df", "-B", "M", df_path]
        stdout = run_cmd(args=cmdlist)
        lines = stdout.splitlines()
        idx0 = lines[0].index("Used") + 5
        idx1 = lines[0].index("Use%")
        hd_avail = lines[-1][idx0:idx1].strip(' M')

        rs = "cpu_idle=%s;mem_free=%s;hd_avail=%s" % (cpu_idle, mem_free, hd_avail)
        logger.info("sys_perf_info:%s" % rs)
        return "success:%s" % rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("sys_perf_info: failed! => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def sys_os_info():
    """
    get architecture, cpu number, cpu speed (bogomips), memory size (in MB),

    @return     'success:arch=<val>;mem_total=<val><value from sys_cpu_info()>'
                    if success else 'failed:<error message>'
    """
    logger = logging.getLogger("ovs.query")
    try:
        #import platform
        #arch = platform.architecture()[0]
        rs = xen_get_caps()
        if not rs.startswith("success"):
            raise Exception(rs)
        rs = parse_result(rs)["caps"]
        if "x86_64" in rs:
            arch = "64bit"
        else:
            arch = "32bit"

        stdout = timeout_command("xm info | grep total_memory")
        mem_total = int(stdout.split(":")[1].strip())

        rs = sys_cpu_info()
        if not rs.startswith("success"): raise Exception(rs)
        rs = rs[8:] #remove success:
        rs = "hostname=%s;arch=%s;mem_total=%s;%s" % (socket.gethostname(), arch, mem_total, rs)

        logger.info("sys_os_info:success. %s" % rs)
        return "success:%s" % rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("sys_os_info: failed. %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def sys_cpu_info():
    """
    get chip arch info

    return    "success:vendor_id=<val>;cpu_family=<val>;model=<val>;model_name=<val>;stepping=<val>;bogomips=<val>;cpu_num=<val>"
    """
#        vendor_id       : GenuineIntel
#        cpu family      : 6
#        model           : 15
#        model name      : Intel(R) Core(TM)2 CPU          6400  @ 2.13GHz
#        stepping        : 2

    logger = logging.getLogger("ovs.query")
    try:
        f = open("/proc/cpuinfo")
        try:
            d, count = {}, {}
            for line in f:
                try:
                    k, v = line.split(":")
                    k = "_".join(k.strip().split())
                    d[k] = v.strip()
                    if k in count: count[k] += 1
                    else:count[k] = 1
                except: pass
        finally:
            f.close()

        d_arch = {}
        for k in ("vendor_id", "cpu_family", "model", "model_name", "stepping", "bogomips"):
            if k in d:
                d_arch[k] = d[k]
            else:
                d_arch[k] = ""
        d_arch["cpu_num"] = str(count["bogomips"])
        rs = ";".join(sorted(["%s=%s" % (k, v) for (k, v) in d_arch.items()]))
        logger.info("sys_cpu_info:success. %s" % rs)
        return "success:%s" % rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("sys_cpu_info:failed. %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def xen_hypervisor_info():
    """
    get xm info
    return success:mem_free=<mem_free>;vcaps=<xen_caps>
    """
    logger = logging.getLogger("ovs.query")
    try:
        #cmd = "xm info"
        cmd = ["xm", "info"]
        output = timeout_command(cmd)
        d = {}
        for line in output.splitlines():
            try:
                parts = line.split(":", 1)
                d[parts[0].strip()] = parts[1].strip()
            except: pass

        mem_free = d["free_memory"]
        vcaps = d["xen_caps"]
        if _is_maintenance_mode():
            D('maintain mode')
            mem_free = 0

        rs = "success:mem_free=%s;vcaps=%s" % (mem_free, vcaps)
        logger.info("xen_hypervisor_info:%s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_hypervisor_info: failed. => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def xen_get_xm_info():
    """Get xm info
    return success:<key>=<value>[,<key>=<value>]
    """
    logger = logging.getLogger("ovs.query")
    try:
        #cmd = "xm info"
        cmd = ["xm", "info"]
        output = timeout_command(cmd)
        l = []
        for line in output.splitlines():
            try:
                parts = line.split(":", 1)
                l.append("%s=%s" % (parts[0].strip(), parts[1].strip()))
            except:
                pass

        rs = "success:%s" % ";".join(l)
        logger.info("xen_get_xm_info:%s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_get_xm_info: failed. => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def xen_get_uptime(vm_path):
    """
    get vm uptime
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm_name = basename(vm_path)
        #lines = run_cmd("xm uptime '%s'" % vm_name)
        cmdlist = ["xm", "uptime", vm_name]
        output = run_cmd(args=cmdlist)
        lines = output.split('\n')
        columns = lines[0]
        index = columns.find('Uptime')
        if index < 0:
            raise "Failed to locate 'uptime' column."
        line = lines[1]
        uptime = line[index:].strip()

        logger.info("xen_get_uptime: success. vm_path('%s') => %s" % (vm_path, uptime))
        return "success:uptime=%s" % uptime
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_get_uptime: failed. vm_path('%s') => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg



@exposed
def xen_get_network_config():
    """Get the virtual network configuration of a machine."""

    logger = logging.getLogger("ovs.query")
    try:
        session = OVSXAPIUtil.get_session()
        network_config = []
        for k, v in session.xenapi.network.get_all_records().items():
            network_config.append("id=%s,label=%s" % (v["uuid"], v["name_label"]))
        session.xenapi.session.logout()
        rs = "success:%s" % ";".join(network_config)
        logger.info("xen_get_network_config: %s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_get_network_config: error. %s" % errmsg)
        return "failed:%s" % errmsg


@exposed
def xen_get_server_config():
    """Get the virtual server configuration of a machine."""

    logger = logging.getLogger("ovs.query")
    try:
        network_type = OVSXAPIUtil.get_network_type()
        vt_enabled = OVSXAPIUtil.is_vt_enabled()
        os_version = OVSXAPIUtil.get_os_version()
        rs = "success:network_type=%s,vt_enabled=%s,os_version=%s" % (network_type, vt_enabled, os_version)
        logger.info("xen_get_server_config: %s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_get_server_config: error. %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def xen_get_server_perf_metrics():
    """Get server's performance metrics

    @return:
        1.cpuutil   => <cpu_num>:<cpu_util>[;<cpu_num>:<cpu_util>]
        2.memutil
            memory_total    => (in MB)
            memory_free     => (in MB)
        3.guestvm_count => len(xenapi.VM.get_all()) - 1
        4.diskutil      => zlib.compress($(df -Pm)).encode("base4")
        5.uptime        => zlib.compress($(uptime)).encode("base64")
    """
    logger = logging.getLogger("ovs.query")
    try:
        session = session_login()
        r = []
        try:
            ref = session.xenapi.host.get_all()[0]
            host = XenAPIObject("host", session, ref)
            #cpuutil
            refs = host.get_host_CPUs()
            l = []
            for ref in refs:
                cpu = XenAPIObject("host_cpu", session, ref)
                l.append( (str(cpu.get_number()), cpu.get_utilisation()) )
            r.append( ("cpuutil" , ",".join(["%s:%s" % (k, str(v)) for k, v in l])) )
            #memutil
            ref = host.get_metrics()
            metrics = XenAPIObject("host_metrics", session, ref)
            r.append( ("memory_total", int(metrics.get_memory_total()) / 1024) )
            r.append( ("memory_free", int(metrics.get_memory_free()) / 1024) )
            #guestvm_count
            refs = session.xenapi.VM.get_all()
            r.append( ("guestvm_count", len(refs) - 1) ) #remove dom0
        finally:
            session_logout()

        #diskutil
        cmdlist = ["df", "-Pm"]
        r.append( ("diskutil", "%s" % marshall_string(run_cmd(args=cmdlist).strip())) )
        #uptime
        cmdlist = ["uptime"]
        r.append( ("uptime", "%s" % marshall_string(run_cmd(args=cmdlist).strip())) )

        ms = ";".join(["%s=%s" % (k, str(v)) for k, v in r])
        rs = "success:%s" % ms
        logger.info("xen_get_server_perf_metrics: success. %s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_get_server_perf_metrics: error. => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def xen_get_vm_perf_metrics(vm_path):
    """Get vm's performance metrics

    @return:
        * vm_uuid
        * vm_name
        * vm_uptime     => <now - start_time in secs>
        * vm_state      => <status>[;<status>]
        * vm_cpuutil    => <cpu_num>:<cpu_util>[;<cpu_num>:<cpu_util>] cpu_util in (0, 100)
        * vm_memory_static_min (in MB)
        * vm_memory_static_max (in MB)
        * vm_memory_dynamic_min (in MB)
        * vm_memory_dynamic_max (in MB)
        * vm_memory_actual      (in MB)
        * vm_network_read_rate  => <device>:<rate in kbs>[;<device>:<rate in kbs>]
        * vm_network_write_rate => <device>:<rate in kbs>[;<device>:<rate in kbs>]
        * vm_disk_read_rate     => <device>:<rate in kbs>[;<device>:<rate in kbs>]
        * vm_disk_write_rate    => <device>:<rate in kbs>[;<device>:<rate in kbs>]
        * tot_network_io   => to be implemented
        * tot_disk_io      => to be implemented
    """
    logger = logging.getLogger("ovs.query")
    try:
        vm_cfg = join(vm_path, "vm.cfg")
        cmdlist = ["grep", "uuid", vm_cfg]
        ref = uuid = run_cmd(args=cmdlist).split("=")[-1].strip("\n \"'")
        session = session_login()
        r = [ ("vm_uuid", uuid) ]
        try:
            refs = session.xenapi.VM.get_all()
            if not ref in refs:
                raise Exception("uuid='%s' not found in xenstore." % uuid)
            vm = XenAPIObject("VM", session, ref)
            r.append( ("vm_name", vm.get_name_label()) )
            r.append( ("vm_memory_static_min", str(int(vm.get_memory_static_min())/1024/1024)) )
            r.append( ("vm_memory_static_max", str(int(vm.get_memory_static_max())/1024/1024)) )
            r.append( ("vm_memory_dynamic_min", str(int(vm.get_memory_dynamic_min())/1024/1024)) )
            r.append( ("vm_memory_dynamic_max", str(int(vm.get_memory_dynamic_max())/1024/1024)) )

            VM_metrics = XenAPIObject("VM_metrics",session, vm.get_metrics())
            r.append( ("vm_state", ",".join(VM_metrics.get_state())) )
            r.append( ("vm_memory_actual", str(int(VM_metrics.get_memory_actual())/1024/1024)) )
            r.append( ("vm_cpuutil", ",".join([ "%s:%s" % (k, str(v)) \
                    for k, v in VM_metrics.get_VCPUs_utilisation().items()])) )

            iso_fmt = "%Y%m%dT%H:%M:%S"
            delta = mktime(gmtime()) - mktime(strptime(VM_metrics.get_start_time().value, iso_fmt))
            r.append( ("vm_uptime", repr(delta)) )

            VIFs = vm.get_VIFs()
            lr, lw = [], []
            for ref in VIFs:
                VIF = XenAPIObject("VIF", session, ref)
                VIF_metrics = XenAPIObject("VIF_metrics", session, VIF.get_metrics())
                lr.append("%s:%s" % (VIF.get_device(), VIF_metrics.get_io_read_kbs()))
                lw.append("%s:%s" % (VIF.get_device(), VIF_metrics.get_io_write_kbs()))
            r.append( ("vm_network_read_rate", ",".join(lr)) )
            r.append( ("vm_network_write_rate", ",".join(lw)) )
            #FIXME:
            r.append( ("tot_network_io", "TBD") )

            VBDs = vm.get_VBDs()
            lr, lw = [], []
            for ref in VBDs:
                VBD = XenAPIObject("VBD", session, ref)
                VBD_metrics = XenAPIObject("VBD_metrics", session, VBD.get_metrics())
                lr.append("%s:%s" % (VBD.get_device(), VBD_metrics.get_io_read_kbs()))
                lw.append("%s:%s" % (VBD.get_device(), VBD_metrics.get_io_write_kbs()))
            r.append( ("vm_disk_read_rate", ",".join(lr)) )
            r.append( ("vm_disk_write_rate", ",".join(lw)) )
            #FIXME:
            r.append( ("tot_disk_io", "TBD") )

        finally:
            session_logout()

        ms = ";".join(["%s=%s" % (k, str(v)) for k,v in r])
        rs = "success:%s" % ms
        logger.info("xen_get_vm_perf_metrics: success. %s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_get_vm_perf_metrics: vm('%s')error. => %s" % \
                (vm_path, errmsg))
        return "failed:%s" % errmsg


