#!/usr/bin/env python

#^#
# OVSXUtility.py
#
# ovs agent module for cp/rm/mv/scp etc
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#
import urllib
import tempfile
import sys
import string
import re
import logging
from tempfile import NamedTemporaryFile
from shutil import copytree, copyfile, rmtree, move
from os.path import join, isdir, isfile, islink, ismount, exists, dirname, abspath, isabs, realpath, normpath, basename
from os import makedirs, stat, walk, remove, rename, rmdir, mkdir, listdir
from glob import glob
from OVSXXenVMConfig import xen_unmarshall_cfg, xen_clear_vif_mac, xen_remove_shared_disk, xen_update_uuid, xen_correct_cfg
from OVSConstants import OVS_BASE, OVS_MOUNT, RESERVED_SPACE
from OVSCommons import parse_result, run_cmd, old_run_cmd, fmt_err_msg, chk_vm_path, exposed, get_uuid, strings_common_head
from OVSAsyncTaskAux import save_async_auxdata
from OVSConfig import get_config

import shutil

#----------------------------<> helper function <>----
def get_storage_free_space(dir_path):
    logger = logging.getLogger("ovs.query")

    d = {}
    try:
        if not exists(dir_path):
            return d
        stdout = run_cmd(args = ["df", "--block-size=M", dir_path])
    except Exception, e:
        logger.error("get_storage_free_space: failed. '%s' =>%s" % (dir_path, fmt_err_msg(e)))
        return d

    lines = stdout.splitlines()
    #Mounted on, mp = mount point
    mp_idx0 = lines[0].index("Mounted on")
    #Available
    avail_idx0 = lines[0].index("Used") + 5
    avail_idx1 = lines[0].index("Use%")

    l = []
    success_mapping = False
    for idx, line in enumerate(lines[1:]):
        try:
            if line.startswith(" "): #previous line is for Filesystem
                if success_mapping == True: #and filesystem line done a success mapping
                    del l[-1]
            success_mapping = False

            mp = line[mp_idx0:].strip()
            hd_avail = line[avail_idx0 : avail_idx1].strip(' M')
            l.append((mp, int(hd_avail) - RESERVED_SPACE))

            success_mapping = True
        except: continue

    return dict(l)


from OVSCDS import *

def parse_storage_repos():
    l = []

    d = db_dump("sr")
    for k, v in d.items():
        uuid = v.__dict__["_uuid"]
        l.append(uuid)

    repos_prefix = "/var/ovs/mount"
    l =  [join(repos_prefix, uuid) for uuid in l]

    return l


def find_real_path(vpath, list_all=False):
    """
    Find the real location of path, which could be inside any shared dir,
    e.g., /var/ovs/mount/<uuid>/running_pool/aaa

    If the vpath is a device name, i.e., phy:/bbb, return itself.

    @vpath      e.g., /OVS/running_pool/ccc
    @list_all   if True will return list if multiple paths found,
                else will raise exception.
    """

    logger = logging.getLogger("ovs.query")

    # return directly when the vpath is leading by phy:
    if vpath.startswith('phy:'): return vpath

    l = parse_storage_repos()
    ptn = re.compile("^%s/" % OVS_BASE)
    vpath = ptn.sub("", vpath)
    possible_paths = [join(repo, vpath) for repo in l]

    real_paths = [path for path in possible_paths if exists(path)]
    real_paths = list(set(real_paths))

    if real_paths:
        if not list_all:
            if len(real_paths) > 1:
                raise Exception("Got more than one file/dir with the same name: '%s'" % repr(real_paths))
            return real_paths[0]
        else:
            return real_paths
    else:
        raise Exception("Could not find '%s' in any storage repository." % vpath)


@exposed
def utl_find_real_path(vpath, list_all=False):
    """
    Find the real location of path, which could be inside any shared dir,
    e.g., /var/ovs/mount/<uuid>/running_pool/aaa

    If the vpath is a device name, i.e., phy:/bbb, return itself.

    @vpath      e.g., /OVS/running_pool/ccc
    @list_all   if True will return list if multiple paths found,
                else will raise exception.
    """
    logger = logging.getLogger("ovs.query")
    try:
        list_all = bool(list_all)
        rs = find_real_path(vpath, list_all)
        if list_all: rs = ",".join(rs)
        logger.info("utl_find_real_path: success. path('%s') => %s" % (vpath, rs))
        return "success:real_path=%s" % rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_find_real_path: failed. =>%s" % errmsg)
        return "failed:%s" % errmsg


def get_storage_repos():
    d_repos_space = {}
    l = parse_storage_repos()
    for sr in l:
        d = get_storage_free_space(sr)
        if d:
            d_repos_space[sr] = d[sr]

    if not d_repos_space:
        raise Exception("No storage repository found!")
    # get repos with maximum space
    d = dict([(space, repos) for repos, space in d_repos_space.items()])
    free_space = max(d_repos_space.values())
    repos = d[free_space]
    return repos, free_space


def assemble_real_path(vpath, repos):
    return re.sub(r"^%s" % OVS_BASE, repos, vpath)

#--------------------<> VM manipulation <>----
@exposed
def utl_get_storage_repos():
    logger = logging.getLogger("ovs.query")
    try:
        repos, free_space = get_storage_repos()

        rs = "success:repos=%s;free_space=%d" % (repos, free_space)
        logger.info("util_get_storage_repos: %s" % rs)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_get_storage_repos: failed. =>%s" % errmsg)
        return "failed:%s" % errmsg

def _regen_uln_uuid(cfg_file_path):
    """ re-generate ULN UUID
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("_regen_uln_uuid: %s." % cfg_file_path)
        #cfg = xen_unmarshall_cfg(cfg_file_path)
        #disks = cfg["disks"]

        dest_dir = dirname(cfg_file_path)
        if "'" in dest_dir:
            raise Exception("single quotation mark is not allowed in dest_dir!")
        vdisk_start = "file:"
        #tmp_mount_point = "/tmp/tmp_ovs_clone_mount_point"
        tmp_mount_point = tempfile.mkdtemp('ovs_mount_point')
        if "'" in tmp_mount_point:
            raise Exception("single quotation mark should not exist in tmp_mount_point!")
        uuid_filepath = "etc/sysconfig/rhn/up2date-uuid"

        try:
            mkdir(tmp_mount_point)
        except:
            pass

        disks = []
        for root, dirs, files in walk(dest_dir):
            for name in files:
                if name.endswith('.img'):
                    disks.append(name)

        for disk in disks:
            logger.debug("_regen_uln_uuid: %s." % disk)

            filepath = disk
            try:
                if "'" in filepath:
                    raise Exception("single quotation mark is not allowed in filepath!")
                old_run_cmd("cd '%s' && mount -o loop '%s' '%s'" % (dest_dir, filepath, tmp_mount_point))
            except:
                logger.debug("_regen_uln_uuid: mount -o loop %s %s failed." % (filepath, tmp_mount_point))
                continue

            uuid_file_full_path = join(tmp_mount_point, uuid_filepath)
            if not isfile(uuid_file_full_path):
                logger.debug("/%s doesn't exited in %s" % (uuid_filepath, filepath))
                try:
                    run_cmd(args = ["umount", tmp_mount_point])
                except: pass
                continue

            # re-generate uuid
            logger.info("_regen_uln_uuid: %s re-generate uuid." % filepath)
            newlines = []
            f = open(uuid_file_full_path, 'r')
            try:
                for line in f:
                    if re.match(r"rhnuuid=.{8}-.{4}-.{4}-.{4}-.{12}", line):
                        line = "rhnuuid=%s\n" % get_uuid()
                    newlines.append(line)
            finally:
                f.close()

            f = open(uuid_file_full_path, 'w')
            #f = open("/tmp/up2date-uuid", 'w')
            try:
                f.writelines(newlines)
            finally:
                f.close()

            try:
                run_cmd(args = ["umount", tmp_mount_point])
            except: pass

        return "success:"
    except Exception,e:
        try:
            if ismount(tmp_mount_point):
                try:
                    run_cmd(args = ["umount", tmp_mount_point])
                except: pass
        except:
            pass
        logger.error("_regen_uln_uuid: %s" % e)
        return "failed:%s" % e

def get_sparse_ratio(vpath):
    if "'" in vpath:
        raise Exception("Single quotation mark not allowed!")

    used = 0.0
    allocated = 0.0
    ratio = 0.0

    cmd = "ls -lsgG --block-size=1M '%s' | grep -v '^total'" % vpath

    try:
        output = old_run_cmd("%s | awk '{print $1}'" % cmd)
        l = output.splitlines()
        for i in l: used += int(i)

        output = old_run_cmd("%s | awk '{print $4}'" % cmd)
        l = output.splitlines()
        for i in l: allocated += int(i)

        if allocated > 0:
            ratio = 1 - used / allocated
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger = logging.getLogger("ovs.query")
        logger.error("get_sparse_ratio: failed. => %s" % errmsg)

    return ratio

@exposed
def utl_cp_vm(src_vm_path, dst_vm_path):
    """
    copy vm
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(src_vm_path)
        chk_vm_path(dst_vm_path)

        logger.info("utl_cp_vm: vm('%s') -> vm('%s') start..." % (src_vm_path, dst_vm_path))

        if not exists(src_vm_path):
            raise Exception("VM('%s') does not exist." % src_vm_path)
        if exists(dst_vm_path):
            raise Exception("VM('%s') already exists." % dst_vm_path)

        config = get_config()
        ratio = float(config.get("vm", "disk.sparse.ratio"))
        sparse_ratio = get_sparse_ratio(src_vm_path)
        if sparse_ratio < ratio:
            sparse_flag = False
        else:
            sparse_flag = True

        makedirs(dst_vm_path)

        if isfile(src_vm_path):
            if sparse_flag:
                cmd = ["cp", "--sparse=always", src_vm_path, dst_vm_path]
            else:
                cmd = ["dd", "if=%s" % src_vm_path, "iflag=direct", "of=%s" % dst_vm_path, "bs=1M"]
            run_cmd(args=cmd)

        elif isdir(src_vm_path):
            for dirpath, dirnames, filenames in walk(src_vm_path):
                for _dirname in dirnames:
                    src = join(dirpath, _dirname)
                    dst = src.replace(src_vm_path, dst_vm_path)
                    if not isdir(dst):
                        makedirs(dst)

                for filename in filenames:
                    src = join(dirpath, filename)
                    dst = join(dirpath, filename).replace(src_vm_path, dst_vm_path)
                    if sparse_flag:
                        cmd = ["cp", "--sparse=always", src, dst]
                    else:
                        cmd = ["dd", "if=%s" % src, "iflag=direct", "of=%s" % dst, "bs=1M"]
                    run_cmd(args=cmd)

            dst_vm_cfg_file = join(dst_vm_path, "vm.cfg")
            if exists(dst_vm_cfg_file):
                rs = xen_clear_vif_mac(dst_vm_path)
                if not rs.startswith("success"):
                    raise Exception(rs)
                rs = xen_remove_shared_disk(dst_vm_path)
                if not rs.startswith("success"):
                    raise Exception(rs)
                rs = _regen_uln_uuid(dst_vm_cfg_file)
                if not rs.startswith("success"):
                    raise Exception(rs[len("failed:"):])
                rs = xen_update_uuid(dst_vm_cfg_file)
                if not rs.startswith("success"):
                    raise Exception(rs)
                rs = xen_correct_cfg(dst_vm_cfg_file, dst_vm_path)
                if not rs.startswith("success"):
                    raise Exception(rs)

        rs = utl_get_vm_size(dst_vm_path)
        logger.info("utl_cp_vm: vm (%s) created!" % dst_vm_path)
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_cp_vm: failed! => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def utl_rm_vm(vm_path):
    """
    remove vm
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)

        if not exists(vm_path):
            logger.warning("utl_rm_vm: vm (%s) not exist!" % vm_path)
            return "success:warning=>vm (%s) not exist!" % vm_path

        if isdir(vm_path):
            #allow to remove the vm when creating failed
            #and (exists(join(vm_path, "vm.xml")) or exists(join(vm_path, "vm.cfg"))):
            rmtree(vm_path)
        elif isfile(vm_path): # and vm_path.endswith("tar.gz"):
            remove(vm_path)
        else:
            raise Exception("Not a vm. (config file does not exist.)")
        logger.info("utl_rm_vm: vm (%s) deleted!" % vm_path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_rm_vm: failed! vm('%s') => %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def utl_mv_vm(old_vm_path, new_vm_path):
    """
    Move a vm to its new name
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(old_vm_path)
        chk_vm_path(new_vm_path)

        if not exists(old_vm_path): raise Exception("VM('%s') does not exist." % old_vm_path)
        if exists(new_vm_path): raise Exception("VM('%s') already exists." % new_vm_path)

        parent_dir = dirname(new_vm_path.rstrip("\\/"))
        if not exists(parent_dir):
            makedirs(parent_dir)

        move(old_vm_path, new_vm_path)
        logger.info("utl_mv_vm: success. vm('%s') -> vm('%s')" % (old_vm_path, new_vm_path))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_mv_vm: failed! vm('%s') -> vm('%s') => %s" % \
                (old_vm_path, new_vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def utl_package_vm(vm_path, pkg_path):
    """
    Package vm to a tarball
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(vm_path)
        chk_vm_path(pkg_path)

        parent_dir = dirname(pkg_path.rstrip("\\/"))
        if not exists(parent_dir):
            makedirs(parent_dir)

        logger.info("utl_package_vm: vm('%s') -> '%s' => start..." % (vm_path, pkg_path))
#        if not pkg_path.endswith(".tar.gz"):
#            pkg_path = "%s.tar.gz" % pkg_path.rstrip("\\/")
        vm_path = abspath(vm_path)
        if "'" in vm_path or "'" in pkg_path:
            raise Exception("single quotation mark is not allowed in vm_path or pkg_path!")
        cmd = "cd '%s' && tar -zcfS '%s' ./" % (vm_path, pkg_path)
        old_run_cmd(cmd)

        logger.info("utl_package_vm: success. cmd='%s'" % cmd)
        return utl_get_vm_size(pkg_path)
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_package_vm:failed! vm('%s') -> '%s' => %s" % \
                (vm_path, pkg_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def utl_extract_vm(pkg_path, vm_path):
    """
    extract a packaged vm
    """
    logger = logging.getLogger("ovs.operation")
    try:
        #chk_vm_path(pkg_path)
        #chk_vm_path(vm_path)

        logger.info("utl_extract_vm: '%s' -> '%s' => start..." % (pkg_path, vm_path))
        #if not pkg_path.endswith("tar.gz"): pkg_path = "%s.tar.gz" % pkg_path
        if not isfile(pkg_path): raise Exception("Packaged VM('%s') does not exist." % pkg_path)
        #if isdir(vm_path): raise Exception("vm('%s') already exists!" % vm_path)

        delete_after = False
        if pkg_path == vm_path:
            delete_after = True
            move(pkg_path, "%s.tar.gz" % pkg_path)
            pkg_path = "%s.tar.gz" % pkg_path

        if not isdir(vm_path):
            makedirs(vm_path)
        cmd = ["tar", "tf", pkg_path]
        l = run_cmd(args=cmd).splitlines()
        common_dir = normpath(strings_common_head(l).rsplit("/", 1)[0])
        cmd = ["tar", "-zxfS", pkg_path, "-C", vm_path]
        run_cmd(args=cmd)
        if delete_after:
            remove(pkg_path)

        if common_dir != ".":
            extract_dir = join(vm_path, common_dir)
            if "'" in extract_dir or "'" in vm_path:
                raise Exception("single quotation mark is not allowed in extract_dir or vm_path!")
            cmd = "mv '%s'/* '%s'" % (extract_dir, vm_path)
            old_run_cmd(cmd)
            #remove empty dir
            try:
                while True:
                    cmd = rmdir(extract_dir)
                    extract_dir = extract_dir.rsplit("/", 1)[0]
            except OSError:
                pass
        logger.info("utl_extract_vm: success. cmd='%s'" % cmd)
        return utl_get_vm_size(vm_path)

    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_extract_vm:failed! '%s' -> vm('%s') => %s" % \
                (pkg_path, vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def utl_scp_vm(src_vm_path, dst_vm_path, remote_server, rs_username, rs_passwd):
    """
    cross-site vm copy
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(src_vm_path)

        #TODO: how to check dst_vm_path?
        dst_vm_path = "%s@%s:%s" % (rs_username, remote_server, dst_vm_path)

        #the trailing / in a source name means "copy the contents of this directory".
        #Without the trailing / "copy the directory".
        if isfile(src_vm_path):
            cmd = ["python", join(dirname(__file__),"utils/rcp.py"), src_vm_path, dst_vm_path]
        elif isdir(src_vm_path):
            cmd = ["python", join(dirname(__file__),"utils/rcp.py"), "%s/" % src_vm_path, dst_vm_path]
        else:
            #bug#3813, when src vm path is invalid, possible caused by
            #removing manually or mount point geting stale
            raise Exception("Src VM('%s') is not file or dir." % src_vm_path)
        logger.info("utl_scp_vm: cmd='%s'" % cmd)
        cmd.append(rs_passwd)
        run_cmd(args=cmd)

        logger.info("utl_scp_vm: success. cmd='%s'" % cmd[:-1])
        return utl_get_vm_size(src_vm_path)
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_scp_vm:failed! %s" % errmsg)
        return "failed:%s" % errmsg

#------------------------------<> exposed helper function <>----
@exposed
def utl_get_vm_size(vm_path):
    """
    get the size of the vm
    """
    logger = logging.getLogger("ovs.query")
    try:
        #chk_vm_path(vm_path)

        if exists(vm_path):
            cmd = ["du", "--apparent-size", "--summarize", "--block-size=1M", "--one-file-system", vm_path]
            stdout = run_cmd(args=cmd)
            size = stdout.split("\t")[0]
        else:
            raise Exception("vm('%s') does not exist!" % vm_path)

        logger.info("utl_get_vm_size: vm('%s') => %s" % (vm_path, size))
        return "success:size=%s" % size
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_get_vm_size: failed! vm('%s') => %s" % (vm_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def utl_list_dir(dir_name, level = 1):
    """
    list all files and sub dirs in the dir within the specific directory levels (depth)

    N.B.
       ',' or '=' is not allowed in file/dir name, or it will cause
    misbehavior for the result parser
    """
    names = []
    def walkdir(dir_name, prefix, left_levels):
        if left_levels <= 0:
            return
        left_levels = left_levels - 1
        for root, dirs, files in walk(dir_name):
            for f in files:
                if not f.startswith('.'): names.append(join(prefix, f))
            for d in dirs:
                if not d.startswith('.') and listdir(join(root, d)):
                    names.append('%s/' % join(prefix, d))
                    walkdir(join(root, d), join(prefix, d), left_levels)
            break

    logger = logging.getLogger("ovs.query")
    try:
        if isdir(dir_name):
            walkdir(dir_name, '', level)
            name_str = ','.join(names)
        else:
            raise Exception("utl_list_dir: failed! '%s' is not a directory" % dir_name)

        logger.info("utl_list_dir: directory('%s') level:%s => %s" % (dir_name, level, name_str))
        return "success:names=%s" % name_str
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_list_dir: failed! directory('%s') level:%s => %s" % (dir_name, level, errmsg))
        return "failed:%s" % errmsg

@exposed
def utl_exists(path):
    """
    tell if a file is exists

    @return "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.query")
    try:
        #if you donot wanna impose chk_vm_path, like for phy:/dev/loop123
        no_chk_vm_pathes = ['/dev/']
        if not filter(lambda x: path.startswith(x), no_chk_vm_pathes):
            chk_vm_path(path)
        if exists(path):
            logger.info("utl_exists: success. path='%s'" % path)
            return "success"
        else: raise Exception("Does not exist.")
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_exists:failed. path='%s' => %s" % (path, errmsg))
        return "failed:%s" % errmsg

@exposed
def utl_get_free_space(path):
    """
    get free disk space
    """
    logger = logging.getLogger("ovs.query")
    try:
        if not exists(path):
            raise Exception("dir('%s') does not exist!" % path)
        stdout = run_cmd(args = ["df", "-B", "M", path])
        lines = stdout.splitlines()
        idx0 = lines[0].index("Used") + 5
        idx1 = lines[0].index("Use%")
        hd_avail = lines[-1][idx0:idx1].strip(' M')

        free_space = int(hd_avail) - RESERVED_SPACE
        logger.info("utl_get_free_space: success. path('%s') => %s" % (path, free_space))
        return "success:free_space=%s" % free_space
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_get_free_space: failed. path('%s') => %s" % (path, errmsg))
        return "failed:%s" % errmsg

@exposed
def utl_make_dir(path):
    """
    make dir
    """
    logger = logging.getLogger("ovs.operation")
    try:
        path = abspath(path)
        chk_vm_path(path)
        if not path.startswith(OVS_MOUNT):
            raise Exception("Please make sure the path is under '%s'." % OVS_MOUNT)
        if not isdir(path):
            makedirs(path)

        logger.info("utl_make_dir: success. directory('%s') created" % path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_make_dir: failed. directory('%s') => %s" % (path, errmsg))
        return "failed:%s" % errmsg

@exposed
def utl_remove(path):
    """
    remove a tree/file
    """
    logger = logging.getLogger("ovs.operation")
    try:
        path = abspath(path)
        chk_vm_path(path)
        if not path.startswith(OVS_MOUNT):
            raise Exception("Please make sure the path is under '%s'." % OVS_MOUNT)

        if isdir(path):
            rmtree(path)
        else:
            remove(path)

        logger.info("utl_remove: success. %s removed" % path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_remove: failed. %s => %s" % (path, errmsg))
        return "failed:%s" % errmsg

@exposed
def utl_move(old_path, new_path):
    """
    move a dir/file to its new path
    """
    logger = logging.getLogger("ovs.operation")
    try:
        old_path = abspath(old_path)
        new_path = abspath(new_path)
        chk_vm_path(old_path)
        chk_vm_path(new_path)

        if not old_path.startswith(OVS_MOUNT) or not new_path.startswith(OVS_MOUNT):
            raise Exception("Please make sure the paths are under %s." % OVS_MOUNT)

        move(old_path, new_path)

        logger.info("utl_move: success. %s moved to %s" % (old_path, new_path))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_move: failed. %s to %s => %s" % \
                (old_path, new_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_create_disk(disk_path, size, sparse=True):
    """
    create xen raw disk (a file filled with zeros)
    size is number of MB
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(disk_path)

        if exists(disk_path):
            raise Exception("File %s already exists." % disk_path)

        pdir = dirname(disk_path)
        if not isdir(pdir): makedirs(pdir)

        # always create sprase files
        size = int(size)
        seek = size - 1
        count = 1
        #if sparse: seek = size - 1; count = 1
        #else: seek = 0; count = size
        cmd = ["ionice", "-c3", "dd", "if=/dev/zero", "of=%s" % disk_path, "oflag=direct", "bs=1M", "seek=%d" % seek, "count=%d" % count]
        run_cmd(args=cmd)

        logger.info("xen_create_disk: success. disk('%s')" % disk_path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_create_disk: error. disk('%s')=>%s" % (disk_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def xen_delete_disk(disk_path):
    """
    delete disk physically
    """
    logger = logging.getLogger("ovs.operation")
    try:
        chk_vm_path(disk_path)

        if isdir(disk_path):
            raise Exception("xen disk should be a file, but ('%s') is a dir." % disk_path)
        if not isfile(disk_path):
            logger.warning("xen_delete_disk: disk('%s') not exist." % disk_path)
            return "success"
        else:
            #we assume xen disk consists only one single file
            remove(disk_path)
            logger.info("xen_delete_disk: success. disk('%s')"  % disk_path)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("xen_delete_disk: failed. disk('%s') => %s" % (disk_path, errmsg))
        return "failed:%s" % errmsg

@exposed
def utl_test_url(url, proxy_server="", proxy_user="", proxy_password="", is_dir=False):
    """
    test url --- to check the availability of the specified resource
    """
    logger = logging.getLogger("ovs.operation")
    try:
        f = NamedTemporaryFile()
        try:
            if is_dir and not url.endswith('/'):
                url = "%s/" % url
            if "'" in url:
                raise Exception("single quotation mark is not allowed in URL!")
            if not proxy_server:
                cmd = "wget --tries=2 --timeout=10 -o %s --spider '%s'" % (f.name, url)
            else:
                lower_url = url.lower()
                if lower_url.startswith("http://"):
                    proxy = "http_proxy"
                elif lower_url.startswith("ftp://"):
                    proxy = "ftp_proxy"
                elif lower_url.startswith("https://"):
                    proxy = "https_proxy"
                else:
                    raise Exception("Invalid url('%s')." % url)
                if not proxy_user:
                    cmd = "export %s='%s' && wget --tries=2 --timeout=10 -o %s --spider '%s'" % \
                            (proxy, proxy_server, f.name, url)
                else:
                    cmd = "export %s='%s' && wget --tries=2 --timeout=10 -o %s --proxy-user='%s' --proxy-password='%s' --spider '%s'" % \
                            (proxy, proxy_server, f.name, proxy_user, proxy_password, url)

            lower_url = url.lower()
            if lower_url.startswith('https'):
                cmd = '%s --no-check-certificate ' % cmd

            logger.debug("test_url cmd: %s" % cmd)
            try:
                old_run_cmd(cmd)
            except:
                raise Exception("The specified resource is not available, details: %s" % f.read())

            # a very very special case: ftp + no_proxy + regular file, if  file not eixst, wget doesnot complain, we need to check the log
            # with file exists: ==> SIZE AddRef.rar ... 887503
            # with file exists: ==> SIZE AddRefx.rar ... done.
            if url.lower().startswith("ftp://") and not proxy_server and not is_dir:
                for line in f.readlines():
                    line = line.strip()
                    if line.startswith("==> SIZE") and line.endswith("... done."):
                        raise Exception("The specified resource is not available")

            logger.info("utl_test_url: success. url('%s')" % url)
            return "success"
        finally:
            f.close()
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_test_url: failed. url('%s') => %s" % (url, errmsg))
        return "failed:%s" % errmsg

@exposed
def utl_wget(url, path, is_dir=True, name="", proxy_server="", proxy_user="", proxy_password=""):
    """
    download dir/file with wget
    """
    logger = logging.getLogger("ovs.operation")
    try:
        # download dir/file to a temporary directory and then move it
        from tempfile import mktemp
        logfile = mktemp('.log', 'wget-')
        save_async_auxdata('wgetlog', logfile)
        save_async_auxdata('url', url)

        path = realpath(path)
        tmp_dir = ".wget"
        #to check the existence of "'" which could break old_run_cmd
        if "'" in url:
            raise Exception("single quotation mark is not allowed in parameter: url")
        if "'" in path:
            raise Exception("single quotation mark is not allowed in parameter: path")
        if "'" in name:
            raise Exception("single quotation mark is not allowed in parameter: name")
        if "'" in proxy_server:
            raise Exception("single quotation mark is not allowed in parameter: proxy_server")
        if "'" in proxy_user:
            raise Exception("single quotation mark is not allowed in parameter: proxy_user")
        if "'" in proxy_password:
            raise Exception("single quotation mark is not allowed in parameter: proxy_password")
        # protocol
        lower_url = url.lower()
        if lower_url.startswith("http://"):
            protocol = "http"
        elif lower_url.startswith("ftp://"):
            protocol = "ftp"
        elif lower_url.startswith("https://"):
            protocol = "https"
        else:
            raise Exception("Invalid url('%s')" % url)

        # deal with a directory
        if is_dir:
            # if the url ends with a '/', --mirror will not work for ftp
            url = url.strip('/')
            if protocol != "ftp":
                url = url + '/'

            if name == "":
                name = url.strip('/').split('/')[-1]

            target_dir = '%s/%s/%s' % (path, tmp_dir, name)
            if not exists(target_dir):
                makedirs(target_dir)
            save_async_auxdata('target_dir', target_dir)
            prefix_cmd = "cd '%s'" % target_dir

            reject = "*.htm*"
            parameter = "-o %s  --mirror --continue --tries=3 --dns-timeout=30 --connect-timeout=30 --read-timeout=30 \
--no-parent --no-directories --reject=%s '%s'" % (logfile, reject, url)
            if not proxy_server:
                cmd = "wget %s" % parameter
            else:
                proxy = "%s_proxy" % protocol
                if not proxy_user:
                    cmd = "export '%s'='%s' && ionice -c3 wget %s" % (proxy, proxy_server, parameter)
                else:
                    cmd = "export '%s'='%s' && ionice -c3 wget --proxy-user='%s' --proxy-password='%s' %s" % \
                            (proxy, proxy_server, proxy_user, proxy_password, parameter)

            if protocol == 'https':
                cmd = '%s --no-check-certificate ' % cmd

            cmd = '%s && %s' % (prefix_cmd, cmd)
            logger.debug("wget cmd: %s" % cmd)
            try:
                old_run_cmd(cmd)
            except:
                raise Exception("Error when processing wget.")

            # deal with ftp urls, parse the log file
            if protocol == "ftp":
                f = open(logfile)
                lines = f.readlines()
                f.close()
                dir = urllib.unquote(url.strip('/').split('/')[-1])
                error_line = "No such file '%s'." % dir
                found = False
                for line in lines:
                    if line.strip() == error_line:
                        found = True
                if found:
                    raise Exception("No such directory: url('%s')." % url)

                # remove .listing file
                listing_file = join(target_dir, ".listing")
                rs = utl_remove(listing_file)
                if not rs.startswith("success"):
                    raise Exception(rs)

            # move
            oldpath = target_dir
            newpath = "%s/%s" % (path, name)
            rs = utl_move(oldpath, newpath)
            if not rs.startswith("success"):
                raise Exception(rs)

        # deal with a file
        else:
            target_dir = '%s/%s' % (path, tmp_dir)
            if not exists(target_dir):
                makedirs(target_dir)
            save_async_auxdata('target_dir', target_dir)
            prefix_cmd = "cd '%s'" % target_dir

            if not proxy_server:
                cmd = "ionice -c3 wget -o %s --continue --tries=3 --dns-timeout=30 --connect-timeout=30 --read-timeout=30 '%s'" % (logfile, url)
            else:
                proxy = "%s_proxy" % protocol
                if not proxy_user:
                    cmd = "export '%s'='%s' && ionice -c3 wget -o %s --continue --tries=3 --dns-timeout=30 --connect-timeout=30 --read-timeout=30 '%s'" \
                            % (proxy, proxy_server, logfile, url)
                else:
                    cmd = "export '%s'='%s' && ionice -c3 wget -o %s --continue --tries=3 --dns-timeout=30 --connect-timeout=30 --read-timeout=30 \
--proxy-user='%s' --proxy-password='%s' '%s'" % (proxy, proxy_server, logfile, proxy_user, proxy_password, url)

            if protocol == 'https':
                cmd = '%s --no-check-certificate ' % cmd

            cmd = '%s && %s' % (prefix_cmd, cmd)
            logger.debug("utl_wget: cmd=%s" % cmd)
            try:
                old_run_cmd(cmd)
            except:
                raise Exception("Error when processing wget.")

            #move
            filename = urllib.unquote(url.strip('/').split('/')[-1])
            oldpath = "%s/%s" % (target_dir, filename)
            if name == "": name = filename
            newpath = "%s/%s" % (path, name)
            rs = utl_move(oldpath, newpath)
            if not rs.startswith("success"): raise Exception(rs)
        try:
            remove(logfile)
        except:
            pass

        logger.info("utl_wget: success. url('%s') => path('%s'), name('%s')" % \
                (url, path, name))
        return "success"
    except Exception, e:
        try:
            target_temp_dir = join(path, tmp_dir)
            shutil.rmtree(target_temp_dir)
            remove(logfile)
        except:
            pass
        errmsg = fmt_err_msg(e)
        logger.error("utl_wget: failed. url('%s') to path('%s'), name('%s') => %s" % \
                (url, path, name, errmsg))
        return "failed:%s" % errmsg
#--------------------------------------<> hardware information <>----
@exposed
def utl_get_host_cpu_info():
    """
    get a host CPU information, we assume (which is also a requirement in current design) all the hosts share
    the same hardware configrations(CPU numbers, frequency etc) in the pool

    @return
    "success:<string of cpu informations>" if success else "failed:<error message>"
    examples of string of return value: "success:processor:0,vendor_id:GenuineIntel,...;processor:1,vendor_id:GenuineIntel"
    """
    def get_cpu_info():
        keys = ['processor', 'vendor_id', 'cpu family', 'model', 'model name', 'stepping', 'cpu MHz', 'cache size']
        f = open('/proc/cpuinfo','r')

        try:
            lines = f.readlines()
        finally:
            f.close()

        cpus = []
        for ln in lines:
            if ln.startswith('processor'):
                cpu = []
                cpus.append(cpu)
            if ln.find(':') > 0:
                k, v = ln.split(':')
                k, v = k.strip(), v.strip()
                if k in keys:
                    cpu.append('%s:%s' % (k, v))
        return cpus

    logger = logging.getLogger("ovs.query")
    try:
        cpus = get_cpu_info()
        for i in range(len(cpus)):
            cpus[i] = ','.join(cpus[i])
        rs = ';'.join(cpus)

        logger.info("utl_get_host_cpu_info: success.  %s" % (rs))
        return 'success:%s' % rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_get_host_cpu_info: failed. %s" % (errmsg))
        return "failed:%s" % errmsg

#--------------------------------------<> system passwd <>----
@exposed
def utl_update_sys_passwd(username, passwd):
    """
    update the system password of the user

    @return: "success" if success else "failed"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        cmd = ["python", join(dirname(__file__),"utils/update_password.py"), username, passwd]
        run_cmd(args=cmd)

        logger.info("utl_update_sys_passwd: success.")
        return "success"
    except:
        logger.error("utl_update_sys_passwd: failed.")
        return "failed"

#--------------------------------------<>  <>----
@exposed
def utl_nfs_export_iso(iso_file, mnt_dir, umount=False):
    """Export content of an iso file through mnt_dir (, mainly for para vm installation).

    @iso_file: iso file to be exposed
    @mnt_dir: the dir to mount iso file and put in /etc/exportfs
    @return: "success" if success else "failed:<error message>"
    """
    logger = logging.getLogger("ovs.operation")
    try:
        #start nfs service
        cmd = "netstat -tln | grep -q 2049" #2049 => port for nfs
        try:
            old_run_cmd(cmd)
        except:
            cmd = ["/sbin/service", "portmap", "start"]
            run_cmd(args=cmd)
            cmd = ["/sbin/service", "nfs", "start"]
            run_cmd(args=cmd)

        iso_file = realpath(iso_file)
        mnt_dir = realpath(mnt_dir)
        if "'" in iso_file or "'" in mnt_dir:
            raise Exception("single quotation mark is not allowed in iso_file or mnt_dir!")
        #do mount
        try:
            cmd = "mount | grep -q 'on %s '" % mnt_dir
            old_run_cmd(cmd)

            cmd = "mount | grep -q '%s on %s '" % (iso_file, mnt_dir)
            try:
                old_run_cmd(cmd)
                #mounted to the rite place
            except:
                raise Exception("another device/file already mounted at '%s'." % mnt_dir)
            if umount:
                    cmd = ["umount", mnt_dir]
                    run_cmd(args=cmd)

        except: #not mounted yet
            if not umount:
                if not exists(mnt_dir):
                    makedirs(mnt_dir)
                cmd = ["mount", "-o", "loop", iso_file, mnt_dir]
                run_cmd(args=cmd)

        #do nfs export
        try:
            cmd = ["grep", "-q", '^%s\W' % mnt_dir, "/etc/exports"]
            run_cmd(args=cmd)

            if umount:
                cmd = ["sed", "-i", "-r", "-e", '/^%s\s/d' % mnt_dir.replace("/", "\\/"), "/etc/exports"]
                run_cmd(args=cmd)
                cmd = ["exportfs", "-r"]
                run_cmd(args=cmd)
        except: #not exported yet
            if not umount:
                cmd = "echo '%s *(ro,sync,no_root_squash)' >> /etc/exports" % mnt_dir
                old_run_cmd(cmd)
                cmd = ["exportfs", "-r"]
                run_cmd(args=cmd)

        logger.info("utl_nfs_export_iso: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_nfs_export_iso: failed. %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def utl_nfs_cp(src, dst):
    """
    @src: source file path
    @dst: destination file path

    NOTE: when src/dst file path is an nfs dir, the trailing "/" matters. i.e.
    1) when src=server:/path/to/dir1/ dst=/some/dir2
    it will mount server:/path/to/dir1 and cp server:/path/to/dir1/* to some/dir2/
    2) when src=server:/path/to/dir1 dst=/some/dir2
    it will mount server:/path/to and cp server:/path/to/dir1 to /some/dir2/,
    i.e. /some/dir2/dir1
    """
    logger = logging.getLogger("ovs.operation")
    #------------------------------------------------------
    def _mount(nfs_dir):
        from tempfile import mkdtemp
        td = mkdtemp(suffix="ovs_nfs_")
        cmd = ["mount", "-t", "nfs", nfs_dir, td]
        logger.debug("utl_nfs_cp:_mount() cmd='%s'" % cmd)
        run_cmd(args=cmd)
        return td

    def _umount(mnt_dir):
        cmd = ["umount", mnt_dir]
        logger.debug("utl_nfs_cp:_umount() cmd='%s'" % cmd)
        run_cmd(args=cmd)
        cmd = ["rm", "-rf", mnt_dir]
        logger.debug("utl_nfs_cp:_umount() cmd='%s'" % cmd)
        run_cmd(args=cmd)
    #------------------------------------------------------
    try:
        src_nfs = ":" in src and True or False
        dst_nfs = ":" in dst and True or False
        if not src_nfs and not dst_nfs:
            raise Exception("no nfs location specified!")

        src_dir, src_file_name = src.rsplit("/", 1)
        dst_dir, dst_file_name = dst.rsplit("/", 1)
        if not src_file_name: src_file_name = "*"
        src_local_dir = src_nfs and _mount(src_dir) or src_dir
        dst_local_dir = dst_nfs and _mount(dst_dir) or dst_dir
        try:
            if not dst_nfs and not exists(dst_local_dir):
                makedirs(dst_local_dir)
            cmd = ["cp", "-r", join(src_local_dir, src_file_name), join(dst_local_dir, dst_file_name)]
            logger.debug("utl_nfs_cp: cmd='%s'" % cmd)
            run_cmd(args=cmd)
        finally:
            if src_nfs:
                _umount(src_local_dir)
            if dst_nfs:
                _umount(dst_local_dir)

        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_nfs_cp: failed. %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def utl_get_target_size(target_file):
    """
    get the size of a disk file or a physical device
    @return
    "success:<target_size in bytes>" if success else "failed:<error message>"
    """
    def get_file_size(path):
        f = open(path, 'r')
        try:
            f.seek(0, 2)
            size = f.tell()
            return size
        finally:
            f.close()

    logger = logging.getLogger("ovs.query")
    try:
        if not exists(target_file):
            raise Exception("file %s doesnot exist" % target_file)
        rs = get_file_size(target_file)
        logger.info("utl_get_target_size: success. '%s' => %s" % (target_file, rs))
        return 'success:%s' % rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_get_target_size: failed. '%s' => %s" % (target_file, errmsg))
        return "failed:%s" % errmsg

#--------------------------------------<> share disks <>----
@exposed
def utl_discover_share_disks():
    """
    discover the share disks of this host
    configured in agent.ini: share_disk_pat=/dev/mpath/mpath*:/dev/sd[b-h]*
    select full disk only, not collect the partition, e.g.: list /dev/sdb/ not: /dev/sdb1
    @return
    "success:<devinfo xml>" if  sucess else "failed:<error message>"
    """
    def readlink(path):
        real_path = path
        while islink(real_path):
            try:
                real_path = run_cmd(args = ["readlink", "-f", real_path]).strip()
            except Exception, e:
                logger.warning("Failed to readlink for %s: %s" % (path, str(e)))
                return
        return real_path

    def cmp_by_len(x, y):
        return len(x)- len(y)

    def get_mounted_devs():
        f = open('/etc/mtab', 'r')
        try:
            return [ l.split()[0].strip() for l in f.readlines() ]
        finally:
            f.close()

    def get_mounted_devnums():
        devs = get_mounted_devs()
        ret = []
        for dev in devs:
            if not exists(dev):
                continue
            statinfo = stat(dev)
            if statinfo.st_rdev > 0:
                ret.append(statinfo.st_rdev)
        return ret

    def is_partition_of(d2, d1):
        #to see if block device d2 is a partition of block device d1
        #sda1 is partition of sda
        #mpath1p1 is partition of mpath1
        #mpath10 is NOT partition of mpath1
        if not d2.startswith(d1):
            return False
        n1 = basename(d1)
        n2 = basename(d2)
        if not n1 or not n2 or not n2.startswith(n1):
            return False
        if not n1[-1].isdigit() and n2[len(n1):].isdigit():
            return True
        if n1[-1].isdigit() and n2.startswith(n1 + 'p') and n2[len(n1 + 'p'):].isdigit():
            return True
        return False

    logger = logging.getLogger("ovs.query")
    try:
        config = get_config()
        try:
            share_disk_pat = config.get("security", "share_disk_pat")
        except:
            share_disk_pat = None
        if not share_disk_pat:
            share_disk_pat = "/dev/mpath/*"
        # to get all the share disk candidates
        share_disk_paths = share_disk_pat.split(":")
        share_disk_candidates = []
        for share_disk_path in share_disk_paths:
            share_disk_candidates.extend(glob(share_disk_path))
        if len(share_disk_candidates) == 0:
            raise Exception("No share disk discovered")
        # to remove the partitions, and keep only the full disks
        # e.g.: remove /dev/sdaX but keep /dev/sda
        share_disk_candidates.sort(cmp_by_len)
        i = 0
        while i < len(share_disk_candidates):
            d1 = share_disk_candidates[i]
            j = i + 1
            while j < len(share_disk_candidates):
                d2 = share_disk_candidates[j]
                if is_partition_of(d2, d1):
                    share_disk_candidates.pop(j)
                else:
                    j = j + 1
            i = i + 1
        # the disks must be in /sys/block/*, and not mounted
        block_devs = listdir("/sys/block")
        mounted_devs = get_mounted_devs()
        mounted_devnums = get_mounted_devnums()
        for i in  range(len(share_disk_candidates) - 1, -1, -1):
            real_path = readlink(share_disk_candidates[i])
            if not real_path or not (basename(real_path) in block_devs) or real_path in mounted_devs or stat(real_path).st_rdev in mounted_devnums:
                share_disk_candidates.pop(i)

        if len(share_disk_candidates) == 0:
            raise Exception("No share disk(full disk) discovered")
        #get devinfo
        from devinfo import getDevInfoXML
        rs = ""
        for disk in share_disk_candidates:
            devinfo_xml = getDevInfoXML(disk)
            if devinfo_xml:
                rs = rs + devinfo_xml

        if not rs:
            raise Exception("No share disk with enough information discovered")
        logger.info("utl_discover_share_disks: success. => %s" % (rs))
        return 'success:%s' % rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("utl_discover_share_disks: failed. => %s" % (errmsg))
        return "failed:%s" % errmsg
