#!/usr/bin/python
#============================================================================
# This library is free software; you can redistribute it and/or
# modify it under the terms of version 2.1 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#============================================================================
#

import tempfile
import sys
import os
import commands
from os.path import exists, isfile, realpath, join, islink


SR_REPO_META = ".ovsrepo"

#----------------------------------------------------------------------
def get_repometa(repo_dir):
    d = {}
    repometa_file = join(repo_dir, SR_REPO_META)
    if not exists(repometa_file):
        return d

    f = open(repometa_file)
    try:
        for line in f:
            line = line.strip()
            if not line: continue
            if line.startswith("#"): continue
            k,v = line.split("=", 1)
            d[k] = v
    finally:
        f.close()

    return d

def init_repometa(repo_dir, is_shared=True, description=""):
    repometa_file = join(repo_dir, SR_REPO_META)
    if exists(repometa_file):
        return

    _uuid = commands.getoutput("uuidgen")
    _uuid = _uuid.strip().replace("-", "").upper()
    f = open(repometa_file, "w")
    try:
        f.write("""OVS_REPO_UUID=%s
OVS_REPO_SHARED=%s
OVS_REPO_VERSION=%s
""" % (_uuid, is_shared and "1" or "0", "1"))
    finally:
        f.close()

#----------------------------------------------------------------------
class FileSR(object):
    """ Base class for File based Storage Repos. """

    def __init__(self, storage_provider,
            name_label = 'Untitled',
            name_description = 'File Based Storage Repository',
            content_type = "",
            storage_max = -1, is_shared=True):
        """
        @param storage_max: Maximum disk space to use in bytes. -1 for unlimited.
        """

        self.type = "File Based SR"
        self.name_label = name_label
        self.name_description = name_description
        self.images = {}
        self.sp = storage_provider

        self.physical_size = storage_max
        self.physical_utilisation = 0
        self.virtual_allocation = 0
        self.content_type = ''
        self.is_shared = is_shared

    @property
    def uuid(self):
        if hasattr(self, "_uuid"):
            return self._uuid

        #mount sr
        l = self.sp.mounted()
        if l:
            umount = False
            tmp_repo_dir = l[0][1]
        else:
            umount = True
            tmp_repo_dir = tempfile.mkdtemp(prefix="ovsrepo_")
            self.sp.mount(tmp_repo_dir)

        try:
            #get uuid from repo metadata
            d = get_repometa(tmp_repo_dir)
            if not d:
                init_repometa(tmp_repo_dir, self.is_shared, self.name_label + " || " + self.name_description)
                d = get_repometa(tmp_repo_dir)
            _uuid = d["OVS_REPO_UUID"]
        finally:
            if umount:
                self.sp.umount(tmp_repo_dir)
                if exists(tmp_repo_dir):
                    os.rmdir(tmp_repo_dir)
        #cache it
        self._uuid = _uuid

        return _uuid

    @property
    def mountpoint(self):
        return "/var/ovs/mount/%s" % self.uuid

    def initialize(self):
        mp = self.mountpoint
        #mount sr
        self.sp.mount(mp)
        #permission testing
        f = tempfile.TemporaryFile(dir=mp)
        try:
            f.write("Permission Testing")
        finally:
            f.close()
        #mkdir
        l = ["running_pool", "seed_pool", "iso_pool", "publish_pool", "sharedDisk"]
        for e in l:
            p = join(mp, e)
            if not exists(p):
                if islink(p): os.remove(p)
                os.mkdir(p)

    def get_record(self, transient = True):
        retval = {'uuid': self.uuid,
                  'name_label': self.name_label,
                  'name_description': self.name_description,
                  'virtual_allocation': self.virtual_allocation,
                  'physical_utilisation': self.physical_utilisation,
                  'physical_size': self.physical_size,
                  'type': self.type,
                  'content_type': self.content_type,
                  'VDIs': self.images.keys()}

    def is_valid_vdi(self, vdi_uuid):
        return (vdi_uuid in self.images)

    def get_vdi_by_uuid(self, image_uuid):
        return self.images.get(image_uuid)

    def get_vdi_by_name_label(self, label):
        for image_uuid, image in self.images.items():
            if image.name_label == label:
                return image_uuid
        return None

    def get_vdis(self):
        return self.images.keys()

    def create_vdi(self, vdi_struct):
        raise NotImplementedError()

    def destroy_vdi(self, vdi_struct):
        raise NotImplementedError()
