#!/usr/bin/env python
import random
import os
import commands
from os.path import dirname, join, exists, realpath
from glob import glob
from time import sleep

#----------------------------------------------------------------------
def parse_mount():
    """
    read /proc/mounts and parse to a dict
    """
    l = []
    f = open("/etc/mtab")
    try:
        for line in f:
            try:
                fs_spec, fs_file, fs_vfstype, fs_mntops, fs_freq, fs_passno = line.split()
                if ":" not in fs_spec: #not nfs
                    fs_spec = realpath(fs_spec)
                fs_file = realpath(fs_file)
                l.append((fs_spec, fs_file, fs_vfstype, fs_mntops, fs_freq, fs_passno))
            except:
                continue
    finally:
        f.close()
    return l

def real_path(path):
    if ":" in path: #nfs
        return path
    else:
        return realpath(path)

def get_dev_uuid(dev_spec):
    cmd = "/sbin/blkid -s UUID -o value %s" % dev_spec
    (status, output) = commands.getstatusoutput(cmd)
    if status == 0:
        return output.strip()
    else:
        errmsg = "Error getting device UUID: %s" % dev_spec
        if output.strip():
            errmsg = "%s => %s" % (errmsg, output)
        raise Exception(errmsg)

def get_dev_spec(dev_uuid, dev_spec):
    dev_spec = realpath(dev_spec)
    l = dev_spec.split("/")
    n = len(l)
    if n < 3 or not dev_spec.startswith("/dev"):
        raise Exception("Invalid device: dev_uuid=%s, dev_spec=%s" % (dev_uuid, dev_spec))

    if n > 3:
        # hp servers, /dev/cciss/c0d0p1
        ptn = "%s/*" % "/".join(l[0:-1])
    else:
        # dm-1 -> dm-*
        if dev_spec.startswith("/dev/dm-"):
            ptn = "/dev/dm-*"
        # sde1 -> sd*1
        else:
            v = list(dev_spec)
            idx = -1
            while v[idx] in "0123456789":
                idx -= 1
            v[idx] = "*"
            ptn = "".join(v)

    devs = glob(ptn)
    tgt_dev = []
    for dev in devs:
        try:
            uuid = get_dev_uuid(dev)
        except:
            continue
        if dev_uuid == uuid:
            tgt_dev.append(dev)
    if tgt_dev:
        return tgt_dev[0]
    else:
        raise Exception("No device found: dev_uuid=%s" % dev_uuid)

#----------------------------------------------------------------------
class FileSP(object):
    """Interface class for file based storage providers

    NOTE:
        1) for all file system based storage providers, the plugin class must derive from FileSP.
        2) for each method, if noted as "must implement", plugin class should implement their own version.
    """
    @classmethod
    def get_plugin_info():
        """Metainfo of this plugin

        @return: dict of attributes, e.g.
            { "vendor name" : "Dummy",
              "plugin version": 2.2-3,
              "supported storage types": ("dummy-p1", "dummy-p2", "dummy-p3"),
            }

        NOTE: must implement.
        """
        pass

    @classmethod
    def probe(*argv):
        """Probe the storage provider
        e.g.: scan exported directories from a server

        @return: (status, output)
            status == 0: success. output=<result>
            status != 0: error. status=<error code>, output=<error message>
        """
        return (0, "")

    ### instance method ###
    def __init__(self, fs_spec):
        """Constructor

        @param dev: partition/nfs to be mounted. e.g.,
            nfs-server1:/shared/dir,
            /dev/disk/by-id/scsi-14f70656e66696c000000000002000000da5e27000e000000-part1
        @param mount_options: default mount options.

        NOTE: please only initialize storage attribute values here, and put the actual
              initialization code in initialize().

              must implement.
        """
        self.fs_spec = fs_spec
        if not ":" in fs_spec:
            self.fs_uuid = get_dev_uuid(fs_spec)

    @property
    def info(self):
        """Attributes information about this storage provider
        @return: dict of attributes, e.g.
            { "dev" : "/dev/disk/by-id/scsi-14f70656e66696c000000000002000000da5e27000e000000-part1",
              "mount_options": "-t ocfs2",
            }

        NOTE: must implement.
        """
        return {}

    @property
    def uuid(self):
        """Return UUID
        """
        if hasattr(self, "_uuid"):
            return self._uuid

        seed = self.fs_spec
        random.seed(seed)
        _uuid = "%08x-%04x-%04x-%04x-%012x" % (
                random.randint(0, 0xffffffff),
                random.randint(0, 0xffff),
                random.randint(0, 0xffff),
                random.randint(0, 0xffff),
                random.randint(0, 0xffffffffffff))
        self._uuid = _uuid
        return self._uuid

    @property
    def fs_type(self):
        """The filetype of this storage provider. e.g. "ocfs2", "nfs"

        NOTE: must implement.
        """
        return ""

    def initialize(self, *argv):
        """None repeatable initialization.

        for e.g.: format disk, etc
        """
        pass

    def get_fs_spec(self):
        if ":" in self.fs_spec:
            tgt_dev = self.fs_spec
        else:
            tgt_dev = get_dev_spec(self.fs_uuid, self.fs_spec)
        return tgt_dev

    def mounted(self):
        """List self.fs_spec entries in /etc/mtab

        @return: [<entry in /etc/mtab>]
        """
        fs_spec = self.get_fs_spec()
        l = parse_mount()
        real_fs_spec = real_path(fs_spec)
        l = filter(lambda e: e[0] == real_fs_spec, l)
        return l

    def mount(self, _dir):
        """Mount self.fs_spec to _dir

        @param _dir: mount point
        @param options: mount options
        @return: (status, output)
            status == 0: success.
            status != 0: error. status=<error code>, output=<error message>

        NOTE:_dir cannot have already been mounted to other filesystem
        """
        fs_spec = self.get_fs_spec()

        # check if already mounted
        l = parse_mount()
        if not exists(_dir):
            os.makedirs(_dir)
        real_dir = real_path(_dir)
        l = filter(lambda e: e[1] == real_dir, l)
        if len(l) > 1:
            raise Exception("multiple filesystem mounted on %s" % _dir)
        if len(l) == 1:
            fs_spec, fs_file = l[0][0:2]
            real_fs_spec = real_path(fs_spec)
            if fs_spec == real_fs_spec and fs_file == real_dir:
                return (0, "already mounted.")
            else:
                raise Exception("filesystem %s already mounted to %s" % (fs_spec, _dir))

        # mount
        cmd = "mount %s %s" % (fs_spec, _dir)
        (status, output) = commands.getstatusoutput(cmd)
        if status != 0:
            raise Exception(output)

        # sleep
        sleep(1)

        return (status, output)

    def umount(self, _dir=""):
        """Umount fs_spec from _dir

        if _dir == "": umount fs_spec from all mount point
        """
        # check if already umounted
        fs_spec = self.get_fs_spec()
        l = parse_mount()
        real_fs_spec = real_path(fs_spec)
        if _dir:
            real_dir = real_path(_dir)
            l = filter(lambda e: e[0] == real_fs_spec and e[1] == real_dir, l)
        else:
            l = filter(lambda e: e[0] == real_fs_spec, l)
        if len(l) == 0:
            (status, output) = (0, "not mounted.")

        # umount
        for e in l:
            fs_spec, fs_file = e[0:2]
            cmd = "umount -fl %s" % fs_spec
            (status, output) = commands.getstatusoutput(cmd)
            if status != 0:
                raise Exception(output)

        # sleep
        sleep(1)

        return (status, output)
