#!/usr/bin/env python

class PhySP(object):
    """Interface class for physical device based storage providers

    NOTE:
        1) for all physical device based storage providers, the plugin class must derive from PhySP.
        2) for method raise NotImplementedError(), the plugin class must implement it.
    """
    @classmethod
    def get_plugin_info(klass):
        """Metainfo of this plugin

        @return: dict of attributes, e.g.
            { "vendor name" : "Dummy",
              "plugin version": 2.2-3,
              "supported storage types": ("dummy-p1", "dummy-p2", "dummy-p3"),
            }

        """
        raise NotImplementedError()

    @classmethod
    def probe(klass, *argv):
        """Probe the storage provider
        e.g.: probe the storage without initialize it.

        @return: (status, output)
            status == 0: success. output=<result>
            status != 0: error. status=<error code>, output=<error message>
        """
        return (0, "")

    ### instance method ###
    def __init__(self, *argv):
        """Constructor

        NOTE: please only initialize storage attribute values here, and put the actual
              initialization code in initialize().
        """
        pass

    @property
    def info(self):
        """Attributes information about this storage provider
        @return: dict of attributes, e.g.
            { "server" : "openfiler_server1",
              "port": 3026,
              "username": "openfiler",
              "password": "secret",
            }
        """
        raise NotImplementedError()

    def initialize(self, *argv):
        """None repeatable initialization.
        """
        pass

    def list_devices(self, uniq=False):
        """List devices belongs to this storage provider.

        @param uniq: indicate whether to show only the uniq dev path
            to identify devices among servers.
        """
        raise NotImplementedError()


