# Copyright (C) 2009 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
'''
Created on 2009-6-26

@author: tizhang <tiger.zhang@oracle.com>

External APIs for OVF
'''


from OVSWrappers import *
from OVSOVF.OVSOVF import _OVSOVFManagement
from OVSOVF.OVSOVF import _OVSOvfSet
from OVSOVF.OVSOVF import ErrorHandler
from ovf import validation
from ovf.OvfFile import OvfFile
from OVSOVF.OVSOva import guestEnvironment
from OVSOVF.OVSOva import envTransport, envTransportVFAT
from OVSOVF.OVSOva import createXmConfigFile
from OVSException import *

import os
import libxml2

def list_ova_files(seed_pool_path):
    '''
    List .ova files

    @type seed_pool_path: path
    @param seed_pool_path: target directory

    @return: a list of .ova files
    '''
    try:
        ovs_mgmt = _OVSOVFManagement(seed_pool_path)
        ova_files = ovs_mgmt.get_available_ovas()
        OI(ova_files)
        return ovs_success(ova_files)
    except Exception, e:
        errmsg = fmt_err_msg(e)
        OE(errmsg)
        return ovs_failed(errmsg)

def ova_validate(filename):
    '''
    Validate an ova package

    @type filename: filename
    @param filename: ova filename
    '''
    try:
        OI(filename)
        ovfset = _OVSOvfSet(filename)
        if ovfset.validate():
            return ovs_success()
        else:
            return ovs_failed()
    except Exception, e:
        errmsg = fmt_err_msg(e)
        OE(errmsg)
        return ovs_failed(errmsg)

def ova_extract(filename, files, target_dir):
    '''
    Extract files from ova package

    @type filename: filename
    @param filename: ova filename

    @type files: a list of filename
    @param files: files want to exact

    @type target_dir: path
    @param target_dir: target directory for storing the fetched files
    '''
    try:
        OI('%s %s %s' % (filename, ' '.join(files), target_dir))
        ovfset = _OVSOvfSet(filename)
        ovfset.fetch_files(files, target_dir)
        return ovs_success('')
    except Exception, e:
        errmsg = fmt_err_msg(e)
        OE(errmsg)
        return ovs_failed(errmsg)
    
def ova_getmembers(ova_file):
    '''
    Get a list of members in the ova package.
    
    @param ova_file: ova file path
    @type ova_file: path string
    
    @return: success:[file1,file2,...], failed:errmsg
    @rtype: string
    '''
    try:
        OI(ova_file)
        ovfset = _OVSOvfSet(ova_file)
        return ovs_success(ovfset.get_file_list())
    except Exception, e:
        errmsg = fmt_err_msg(e)
        OE(errmsg)
        return ovs_failed(errmsg)

def ovf_validate(ovf, schema):
    '''
    Validate ovf against the schema.

    @param ovf: ovf file
    @type ovf: ovf filename

    @param schema: schema file
    @type schema: schema filename
    '''
    try:
        if not os.path.exists(ovf):
            raise Exception("OVF file not found")

        if not os.path.exists(schema):
            raise Exception("Schema file not found")

        OI('%s %s' % (ovf, schema))
        ctxt_parser = libxml2.schemaNewParserCtxt(schema)
        ctxt_schema = ctxt_parser.schemaParse()
        del ctxt_parser

        validator  = ctxt_schema.schemaNewValidCtxt()

        handler = ErrorHandler()

        validator.setValidityErrorHandler(handler.error, handler.warning)

        # Test valid document
        ret = validator.schemaValidateFile(ovf, 0)

        del ctxt_schema
        del validator

        D("ret = %d" % ret)

        if ret == 0:
            return ovs_success()
        else:
            return ovs_failed("error no. %d" % ret)
    except Exception, e:
        errmsg = fmt_err_msg(e)
        D(errmsg)
        return ovs_failed(errmsg)

def ovf_get_efile(ovf, id=None, verbose=False):
    """
    """
    try:
        OI(ovf)
        ret = []

        fields = [ "ovf:id", "ovf:size" ]
        if verbose:
            fields = fields + [ "ovf:href", "ovf:chunkSize", "ovf:compress" ]

        ovfFile = OvfFile(ovf)

        for node in ovfFile.document.getElementsByTagName('File'):
            if id == None or id == node.getAttribute('ovf:id'):
                ret.append(map(node.getAttribute, fields))

        return ovs_success(ret)

    except Exception, e:
        errmsg = fmt_err_msg(e)
        OE(errmsg)
        return ovs_failed(errmsg)

def ovf_get_disk(ovf, id=None, verbose=False):
    """
    Get disk elements from ovf file.
    
    @param ovf: ovf file path
    @type ovf: string
    @param id: disk ID
    @type id: string
    @type verbose: including verbose information or not
    @type verbose: Boolean
    @return: success:[disk1,disk2,...] failed:errmsg
    @rtype: string
    """
    try:
        OI(ovf)
        ret = []

        fields = [ "ovf:fileRef", "ovf:diskId" ]
        if verbose:
            fields = fields + [ "ovf:capacity", "ovf:capacityAllocationUnits",
                               " ovf:format", "ovf:populatedSize", "ovf:parentRef"]

        ovfFile = OvfFile(ovf)

        for node in ovfFile.document.getElementsByTagName('Disk'):
            if id == None or id == node.getAttribute('ovf:diskId'):
                ret.append(map(node.getAttribute, fields))

        return ovs_success(ret)

    except Exception, e:
        errmsg = fmt_err_msg(e)
        OE(errmsg)
        return ovs_failed(errmsg)

def ovf_get_network(ovf, id=None, verbose=False):
    """
    Get Network elements from ovf file.
    
    @param ovf: ovf file path
    @type ovf: string
    @param id: network ID
    @type id: string
    @type verbose: including verbose information or not
    @type verbose: Boolean
    @return: success:[network1,network2,...] failed:errmsg
    @rtype: string
    """
    try:
        OI(ovf)
        ret = []

        fields = [ "ovf:id", "ovf:name" ]

        ovfFile = OvfFile(ovf)

        if verbose:
            for node in ovfFile.document.getElementsByTagName('Network'):
                if id == None or id == node.getAttribute('ovf:id'):
                    net = map(node.getAttribute, fields)
                    net.append(node.getElementsByTagName('Description')[0]
                               .firstChild.data)
                    ret.append(net)
        else:
            for node in ovfFile.document.getElementsByTagName('Network'):
                if id == None or id == node.getAttribute('ovf:id'):
                    ret.append(map(node.getAttribute, fields))

        return ovs_success(ret)

    except Exception, e:
        errmsg = fmt_err_msg(e)
        OE(errmsg)
        return ovs_failed(errmsg)

def ovf_get_deploy(ovf, id=None, verbose=False):
    """
    Get Configuration elements from ovf file.
    
    @param ovf: ovf file path
    @type ovf: string
    @param id: Configuration ID
    @type id: string
    @type verbose: including verbose information or not
    @type verbose: Boolean
    @return: success:[conf1,conf2,...] failed:errmsg
    @rtype: string
    """
    try:
        OI(ovf)
        ret = []

        fields = [ "ovf:id", "ovf:default" ]

        ovfFile = OvfFile(ovf)

        if verbose:
            for node in ovfFile.document.getElementsByTagName('Configuration'):
                if id == None or id == node.getAttribute('ovf:id'):
                    deploy = map(node.getAttribute, fields)
                    deploy.append(node.getElementsByTagName('Label')[0]
                                  .firstChild.data)
                    deploy.append(node.getElementsByTagName('Description')[0]
                                  .firstChild.data)
                    ret.append(deploy)
        else:
            for node in ovfFile.document.getElementsByTagName('Configuration'):
                if id == None or id == node.getAttribute('ovf:id'):
                    ret.append(map(node.getAttribute, fields))

        return ovs_success(ret)

    except Exception, e:
        errmsg = fmt_err_msg(e)
        OE(errmsg)
        return ovs_failed(errmsg)

def _get_system(childs):
    ret = []
    presentElements = []
    for node in childs.childNodes:
        if node.nodeName == 'System':
            for sysNode in node.childNodes:
                if sysNode.localName != None:
                    presentElements.append(str(sysNode.localName))

            for i in range(len(presentElements)):
                ret.append(node.getElementsByTagName('vssd:' + presentElements[i])[0]
                           .firstChild.data)
    return ret

def _get_res(ovfFile, aId=None, vhNode=None):
    ret = []
    if vhNode != None:
        foundID = False
        if vhNode.hasAttribute('ovf:id'):
            id = vhNode.getAttribute('ovf:id')
        elif vhNode.parentNode.hasAttribute('ovf:id'):
            id = vhNode.parentNode.getAttribute('ovf:id')
        for node in vhNode.getElementsByTagName('Item'):
            for child in node.childNodes:
                if child.localName != None:
                    if not foundID:
                        foundID = True
                        ret.append(id)
                    ret.append(child.firstChild.data)
    else:
        for vnode in ovfFile.document.getElementsByTagName('VirtualHardwareSection'):
            foundId = False
            for node in vnode.getElementsByTagName('Item'):
                if aId == None:
                    for sysNode in node.childNodes:
                        if sysNode.localName != None:
                            if not foundID:
                                if node.parentNode.hasAttribute('ovf:id'):
                                    ret.append(node.parentNode.getAttribute('ovf:id'))
                                    foundID = True
                                else:
                                    foundID = True
                                    ret.append(node.parentNode.parentNode.
                                               getAttribute('ovf:id'))
                            ret.append(sysNode.firstChild.data)
                else:
                    if node.parentNode.hasAttribute('ovf:id'):
                        id = node.parentNode.getAttribute('ovf:id')
                    elif node.parentNode.parentNode.hasAttribute('ovf:id'):
                        id = node.paranetNode.parentNode.getAttribute('ovf:id')
                    if aId == id:
                        for sysNode in node.childNodes:
                            if sysNode.localName != None:
                                if not foundID:
                                    ret.append(id)
                                    foundID = True
                                ret.append(sysNode.firstChild.data)
    return ret

def ovf_get_hw(ovf, id=None, verbose=False):
    """
    Get VirtualHardwareSection elements from ovf file.
    
    @param ovf: ovf file path
    @type ovf: string
    @param id: VirtualHardwareSection ID
    @type id: string
    @type verbose: including verbose information or not
    @type verbose: Boolean
    @return: success:[hw1,hw2,...] failed:errmsg
    @rtype: string
    """
    try:
        OI(ovf)
        ret = []

        transport = 'ovf:transport'
        idFound = False

        ovfFile = OvfFile(ovf)

        if verbose:
            for node in ovfFile.document.getElementsByTagName('VirtualSystem'):
                if id == None or id == node.getAttribute('ovf:id'):
                    idFound = True
                    for childs in node.childNodes:
                        hw = []
                        if childs.nodeName == 'VirtualHardwareSection':
                            if childs.hasAttribute(transport):
                               hw.append(childs.getAttribute(transport))
                            hw.append(_get_system(childs))
                            hw.append(_get_res(ovfFile, id, childs))
                        ret.append(hw)
        else:
            for node in ovfFile.document.getElementsByTagName('VirtualSystem'):
                if id == None or id == node.getAttribute('ovf:id'):
                    idFound = True
                    for childs in node.childNodes:
                        hw = []
                        if childs.nodeName == 'VirtualHardwareSection':
                            if childs.hasAttribute(transport):
                                hw.append(childs.getAttribute(transport))
                        ret.append(hw)
        if not idFound:
            raise NotImplementedError, "The VirtualSystem ID provided does not\
            match any ID's in the current OVF."

        return ovs_success(ret)
    except Exception, e:
        errmsg = fmt_err_msg(e)
        OE(errmsg)
        return ovs_failed(errmsg)

def ovf_get_product(ovf, aId=None):
    """
    Get ProductSection elements from ovf file.
    
    @param ovf: ovf file path
    @type ovf: string
    @param aId: ProductSection ID
    @type aId: string
    @return: success:[product1,product2,...] failed:errmsg
    @rtype: string
    """    
    try:
        OI(ovf)
        ret = []

        fields = [ "ovf:class", "ovf:instance" ]

        ovfFile = OvfFile(ovf)
        for node in ovfFile.document.getElementsByTagName('ProductSection'):
            if aId != None:
                if (aId == node.getAttribute('ovf:id') or
                    aId == node.parentNode.getAttribute('ovf:id')):
                    for child in node.childNodes:
                        prod = []
                        if child.localeName != None and child.firstChild != None:
                            prod.append(child.firstChild.data)
                            if child.localName == 'Icon':
                                prod.append(child.getAttribute('ovf:fileRef'))
                                prod.append(child.getAttribute('ovf:height'))
                                prod.append(child.getAttribute('ovf:mimeType'))
                                prod.append(child.getAttribute('ovf:width'))
                            if child.locaName == 'Property':
                                prod.append(child.getAttribute('ovf:key'))
                                prod.append(child.getAttribute('ovf:required'))
                                prod.append(child.getAttribute('ovf:userConfigurable'))
                                prod.append(child.getAttribute('ovf:value'))
                                for subChild in child.childNodes:
                                    if subChild.localName:
                                        prod.append(subChild.firstChild.data)
                        ret.append(prod)
            else:
                for child in node.childNodes:
                    prod = []
                    if child.localName != None and child.firstChild != None:
                        prod.append(child.firstChild.data)
                    if child.localName == 'Icon':
                        prod.append(child.getAttribute('ovf:fileRef'))
                        prod.append(child.getAttribute('ovf:height'))
                        prod.append(child.getAttribute('ovf:mimeType'))
                        prod.append(child.getAttribute('ovf:width'))
                    if child.localName == 'Property':
                        prod.append(child.getAttribute('ovf:key'))
                        prod.append(child.getAttribute('ovf:required'))
                        prod.append(child.getAttribute('ovf:userConfigurable'))
                        prod.append(child.getAttribute('ovf:value'))
                        for subChild in child.childNodes:
                            if subChild.localName:
                                prod.append(subChild.firstChild.data)
                    ret.append(prod)

        return ovs_success(ret)
    except Exception, e:
        errmsg = fmt_err_msg(e)
        OE(errmsg)
        return ovs_failed(errmsg)

def ovf_get_os(ovf, aId=None):
    """
    Get OperatingSystemSection elements from ovf file.
    
    @param ovf: ovf file path
    @type ovf: string
    @param aId: OperatingSystemSection ID
    @type aId: string
    @return: success:[os1,os2,...] failed:errmsg
    @rtype: string
    """    
    try:
        OI(ovf)
        ret = []

        ovfFile = OvfFile(ovf)

        for node in ovfFile.document.getElementsByTagName('OperatingSystemSection'):
            if aId != None:
                if (aId == node.getAttribute('ovf:id') or
                    aId == node.parentNode.getAttribute('ovf:id')):
                    for child in node.childNodes:
                        if child.localName == 'Description':
                            ret.append(child.firstChild.data)
            else:
                for child in node.childNodes:
                    if child.localName == 'Description':
                        ret.append(child.firstChild.data)

        return ovs_success(ret)
    except Exception, e:
        errmsg = fmt_err_msg(e)
        OE(errmsg)
        return ovs_failed(errmsg)

def ovf_guest_environment(ovf, ovf_id=None, env_file=None):
    """
    Create guest environment.
    
    @param ovf: ovf file path
    @type ovf: string
    @param ovf_id: ovf ID
    @type ovf_id: string
    @param env_file: output file
    @return: success or failed:errmsg
    @rtype: string
    """
    class EnvCmdOption:
        ovfFile = ovf
        noPrompt = True
        noValue = True
        ovfId = ovf_id
        xmlFile = env_file
        configuration = None
        virtPlatform = None
        pkgDirectory = None

    try:
        OI(ovf)
        guestEnvironment(EnvCmdOption(), '')
        return ovs_success()
    except Exception, e:
        errmsg = fmt_err_msg(e)
        OE(errmsg)
        return ovs_failed(errmsg)

def ovf_env_transport(ovf, aFormat='iso', aOutFile=None):
    """
    Transport a environment file.
    
    @param ovf: ovf file path.
    @type ovf: string
    @param aFormat: target format, can be 'vfat' or 'iso'
    @type aFormat: string
    @param aOutFile: output file name
    @type aOutFile: string
    @return: success or failed:errmsg
    @rtype: string
    """
    class EnvTransportCmdOption:
        ovfFile = ovf
        format = aFormat
        outFile = aOutFile

    try:
        OI(ovf)
        if EnvTransportCmdOption.format == 'vfat':
            envTransportVFAT(EnvTransportCmdOption(), '')
        elif EnvTransportCmdOption.format == 'iso':
            envTransport(EnvTransportCmdOption(), '')
        else:
            raise ValueError('Unsupported format: %s' % aFormat)
        return ovs_success()
    except Exception,e :
        errmsg = fmt_err_msg(e)
        OE(errmsg)
        return ovs_failed(errmsg)
    
def ovf_create_xm_cfg(ovf, aCfgFile=None, aVirtPlatform=None, aEnvDir=None, aConfigId=None):
    """
    Create vm.cfg file.
    
    @param aCfgFile:  Name of the Xen configuration file to create
    @type aCfgFile: string
    @param aVirtPlatform: Virtualization Platform. 'ovm' for Oracle VM
    @type aVirtPlatform: string
    @param aEnvDir:  Path to environment files
    @type aEnvDir: string
    @param aConfigId: Configuration to use
    @type aConfigId: string
    """
    class EnvCreateXmConfigFile:
        ovfFile = ovf
        cfgFile = aCfgFile
        virtPlatform = aVirtPlatform
        envDir = aEnvDir
        configId = aConfigId
    try:
        OI(ovf)
        createXmConfigFile(EnvCreateXmConfigFile(), '')
        return ovs_success()
    except Exception, e:
        errmsg = fmt_err_msg(e)
        OE(errmsg)
        return ovs_failed(errmsg)

if __name__ == '__main__':
    pass
