# Copyright (C) 2009 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.

from xen.xm import XenAPI

from OVSWrappers import *
from OVSException import *
from OVSServerSelection import *

from OVSCDS import CDS

from VmStates import XenVmState, AgentVmState, XenVmPowerState

session = None

def _init_():
    """
    Module construction entry.
    """
    
    global session
    
    # init xen api session
    uri = 'httpu:///var/run/xend/xen-api.sock'
    session = XenAPI.Session(uri)
    session.xenapi.login_with_password('', '')
    
def _dom_name_to_uuid(dom_name):
    """
    Domain name to UUID.
    
    @raise exception: if not found such domain 
    """
    uuids = session.xenapi.VM.get_by_name_label(dom_name)
    if len(uuids) > 0:
        return uuids[0]
    else:
        raise Exception("Not found such domain: %s" % dom_name)
    
def utl_get_vm_record(dom_name):
    """
    Get vm record from xen store.
    
    @param dom_name: domain name
    @type dom_name: string
    """
    uuid = _dom_name_to_uuid(dom_name)
    record = session.xenapi.VM.get_record(uuid)
    return record

def get_vcpus_max(server, dom_name):
    try:
        OI("%s, %s" % (server, dom_name))
        _, sp = get_utility_server(preference=server)
        vcpus_max = sp.utl_get_vm_record(dom_name)['VCPUs_max']
        return ovs_success("vcpus_max=%s" % vcpus_max)
    except Exception, ex:
        errmsg = fmt_err_msg(ex)
        D(errmsg)
        return ovs_failed(errmsg)

def utl_get_domains_stat(domain_names):
    '''
    Get domains status

    @type domain_names: list of strings
    @param domain_names: list of domain names
    '''
    doms_dict = []

    dom_recs = session.xenapi.VM.get_all_records()
    dom_metrics_recs = session.xenapi.VM_metrics.get_all_records()

    for dom_ref, dom_rec in dom_recs.items():
        dom_metrics_rec = dom_metrics_recs[dom_rec['metrics']]

        dom_rec.update({'name': dom_rec['name_label'],
                        'memory_actual': int(dom_metrics_rec['memory_actual'])/1024,
                        'vcpus': dom_metrics_rec['VCPUs_number'],
                        'state': dom_metrics_rec['state'],
                        'cpu_time': dom_metrics_rec['VCPUs_utilisation'],
                        #'start_time': dom_metrics_rec['start_time']
                        })

        doms_dict.append(dom_rec)

    ret = doms_dict
    if domain_names:
        doms = [dom for dom in doms_dict if dom['name'] in domain_names]

        if len(doms) > 0:
            ret = doms
        else:
            ret = []
    
    D("ret: %s" % ret)
    return ret

def get_domains_stat(server, domain_names):
    '''
    Get domains status from a server
    
    @param server: server hostname
    @type server: hostname
    
    @type domain_names: list of strings
    @param domain_names: list of domain names
    '''
    
    try:
        _, sp = get_utility_server(preference=server)
        stats = sp.utl_get_domains_stat(domain_names)
        
        result = []
        for stat in stats:
            result.append("%s=%s" % (stat['name'], ','.join(stat['state'])))
        return ovs_success(";".join(result))
    except Exception,e:
        errmsg = fmt_err_msg(e)
        D(errmsg)
        return ovs_failed(errmsg)

def _get_all_server(*args, **kwargs):
    type = None
    if 'type' in kwargs:
        type = kwargs['type']
    
    srvs = []
    cds = CDS('srv')
    for srv_name,srv in cds.db_srv.iteritems():
        role = srv['srv_role'].split(',')
        if type:
            if type in role:
                srvs.append((srv_name, ServerProxy(srv["agt_url"])))
        else:
            srvs.append((srv_name, ServerProxy(srv["agt_url"])))
                          
    return srvs

def _stat_translate(xen_domain_rec):
    """ convert xen domain state to agent db state """
    
    xen_domain_state = xen_domain_rec['state']
    xen_domain_power_state = xen_domain_rec['power_state']
    
    if XenVmState.paused in xen_domain_state:
        stat = AgentVmState.SUSPEND
    elif XenVmState.crashed in xen_domain_state:
        stat = AgentVmState.ERROR
    elif (XenVmState.shutdown in xen_domain_state) \
        or (XenVmState.dying in xen_domain_state):
        stat = AgentVmState.DOWN
    elif (XenVmState.running in xen_domain_state) \
        or (XenVmState.blocked in xen_domain_state):
        stat = AgentVmState.RUNNING
    else: # len of state is 0
        if xen_domain_power_state == XenVmPowerState.Suspended:
            stat = AgentVmState.DOWN
        elif xen_domain_power_state == XenVmPowerState.Running:
            stat = AgentVmState.RUNNING
        else: # no, it shouldn't happen!
            D("Error: Unexpected xen domain state (%s,%s)" % 
              (",".join(xen_domain_state), xen_domain_power_state))
            stat = AgentVmState.DOWN
    return stat

def _update_agent_db(vm_path, status, name):
    data = {"status": status}
    if name:
        data.update({"name": name})
    db_update_vm(vm_path, data)

def reset_vm_stat(vm_path, domain_name):
    """
    If there the domain is existing in current pool,
    sync its state to db.
    If the domain is not existing, set the state in db
    to powered off.
    
    Finally, return the state.

    @param vm_path: vm path
    @type vm_path: string
    @param domain_name: domain name
    @type domain_name: string
    """
    # traverse the servers in current pool.
    try:
        OI("%s, %s" % (vm_path, domain_name))
        srvs = _get_all_server()
        
        name = None
        status = AgentVmState.DOWN        
        for srv_name,srv_sp in srvs:
            D("traverse %s" % srv_name)
            stat = srv_sp.utl_get_domains_stat(domain_name)
            if len(stat) > 0:
                name = stat[0]['name']
                if stat[0]['power_state'] == XenVmPowerState.Suspended:
                    D("Warning: %s, %s" % (name, stat[0]['power_state']))
                    
                status =  _stat_translate(stat[0])
                
                # found a running domain, stop here
                break
        
        _update_agent_db(vm_path, status, name)
        
        return ovs_success(status)
    except Exception, ex:
        errmsg = fmt_err_msg(ex)
        D(errmsg)
        return ovs_failed(errmsg)

if __name__ == '__main__':
    pass
