#!/bin/bash

#----------------------------------------------------------------------
DMNFS_HB_FILE_SIZE_MB=300
DMNFS_HB_FILE_NODE_SLOTS=32
DMNFS_HB_FILE_BLOCK_SIZE="2k"
DMNFS_HB_FILE_CLUSTER_SIZE="4k"
DMNFS_HB_FILE_LABEL="server_pool_hb"

hb_get_devices_dmnfs()
{
    heartbeatfile=$1
    # returns the device name for all dmnfs devices for the heartbeat file
    # does NOT output the leading component (/dev/mapper/)
    /sbin/dmsetup table --target=nfs | grep dmnfs | \
            sed -n -e "s#\(dmnfs[0-9]*\): [0-9]* [0-9]* nfs $heartbeatfile [0-9]*#\1#p"
}

hb_device_enable_dmnfs()
{
    heartbeatfile=$1

    devs=$(hb_get_devices_dmnfs $heartbeatfile)
    if [ -n "$devs" ]; then
        numdevs=$(echo $devs|wc -w)
        if [[ -z "$numdevs" || $numdevs > 1 ]]; then
            echo "Incorrect number of devices ($numdevs) found for $heartbeatfile in /dev/mapper: "
            echo "$devs"
            echo "There should be at most one dm-nfs device for $heartbeatfile."
            echo "Please correct this with dmsetup before continuing."
            return 1
        fi
        echo "/dev/mapper/$devs"
        return 0
    fi

    i=0
    while true; do
        dmnfsdev="dmnfs$i"
        devpath="/dev/mapper/$dmnfsdev"
        if [ ! -e $devpath ]; then
            # here we have an unused dm-nfs node
            break
        fi
        i=$[i+1]
    done

    # use only last full block, no partials
    blocksize=512
    blocks=$(/usr/bin/stat -c '%s' $heartbeatfile)
    if [ "x$blocks" == "x" ]; then
        echo "dmnfs setup for $heartbeatfile failed.  blocks is invalid."
        return 1
    fi
    blocks=$[ $blocks / $blocksize ]

    echo -n "0 $blocks nfs $heartbeatfile 0" | /sbin/dmsetup create $dmnfsdev

    echo "$devpath"
    return 0
}

# returns 1 on failure
# returns 0 and outputs number of refs on success
hb_started()
{
    hbdev="$1"
    [ -z "$hbdev" ] && return 1

    #64684A7DD84B42D6AA56096797B213D5: 1 refs
    refs=$(/sbin/ocfs2_hb_ctl -I -d "$hbdev" 2> /dev/null |grep ": [0-9]* refs" | sed 's/^.*: \([0-9]*\) refs/\1/')

    [[ -z "$refs" || "$refs" == 0 ]] && return 1
    echo "$refs"
    return 0
}

hb_start()
{
    # no harm in rerunning, will return success
    hbdev="$1"
    [ -z "$hbdev" ] && return 1
    hb_started "$hbdev" && return 0
    /sbin/ocfs2_hb_ctl -S -d "$hbdev" dmnfs
}

hb_stop()
{
    hbdev="$1"
    [ -z "$hbdev" ] && return 1
    hb_started  "$hbdev" || return 1
    /sbin/ocfs2_hb_ctl -K -d "$hbdev" dmnfs
}

#----------------------------------------------------------------------
create_heartbeat_file()
{
    heartbeatfile=$1
    echo "create heartbeat file: $heartbeatfile"

    [ -f $heartbeatfile ] && return 0
    /bin/dd if=/dev/zero of="$heartbeatfile" bs=1M count=$DMNFS_HB_FILE_SIZE_MB || return 1
    /sbin/mkfs.ocfs2 -L $DMNFS_HB_FILE_LABEL -N $DMNFS_HB_FILE_NODE_SLOTS \
        -b $DMNFS_HB_FILE_BLOCK_SIZE -C $DMNFS_HB_FILE_CLUSTER_SIZE --force \
        "$heartbeatfile"
}

remove_heartbeat_file()
{
    heartbeatfile=$1
    echo "remove heartbeat file: $heartbeatfile"

    [ ! -f $heartbeatfile ] && return 0
    stop_heartbeat $heartbeatfile && rm -f $heartbeatfile
}

start_heartbeat()
{
    heartbeatfile=$1
    echo "start heartbeat against: $heartbeatfile"

    hb_dev_name=$(hb_device_enable_dmnfs $heartbeatfile)
    [[ $? == 0 ]] && hb_start $hb_dev_name
}

stop_heartbeat()
{
    heartbeatfile=$1
    echo "stop heartbeat against: $heartbeatfile"

    hb_dev=/dev/mapper/$(hb_get_devices_dmnfs $heartbeatfile)
    #no heartbeat device found
    [[ "$hb_dev" == "/dev/mapper/" ]] && return 0

    while hb_started "$hb_dev"; do
        hb_stop "$hb_dev"
    done
    dmsetup remove $hb_dev
}

status_heartbeat()
{
    heartbeatfile=$1
    echo "status heartbeat against: $heartbeatfile"

    hb_dev=/dev/mapper/$(hb_get_devices_dmnfs $heartbeatfile)
    hb_started $hb_dev
}

#----------------------------------------------------------------------
show_usage()
{
    echo "$0 --action <[create|remove|start|stop]> --heartbeatfile <heartbeat file path>"
}
#======================================================================
#eval to prevent string expantion
eval set -- $(getopt --long action:,heartbeatfile: --options a:b: -- "$@")

action=''
heartbeatfile=''
while [[ ${#@} > 0 ]]; do
    case "$1" in
    --action | -a)
        action="$2"
        shift 2
    ;;
    --heartbeatfile | -b)
        heartbeatfile="$2"
        heartbeatfile=`readlink -f $heartbeatfile`
        shift 2
    ;;
    --)
        break
    ;;
    *)
        show_usage #()
        exit 1
    ;;
    esac
done

case $action in
create)
    create_heartbeat_file $heartbeatfile
;;
remove)
    remove_heartbeat_file $heartbeatfile
;;
start)
    start_heartbeat $heartbeatfile
;;
stop)
    stop_heartbeat  $heartbeatfile
;;
status)
    status_heartbeat $heartbeatfile
;;
*)
    show_usage #()
    exit 1
esac

