#!/usr/bin/env python
#
#^#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

from os.path import dirname, join, exists
import sys
sys.path.append(join(dirname(__file__), ".."))

import pickle
from xmlrpclib import ServerProxy
from OVSCDS import *
from OVSSiteCluster import *
from repos import *
from do_rpc import *
from OVSCommons import get_host_by_name

import logging
import logging.handlers

#======================================================================
#setup the logging system, no dependency to the log server.

formatter = logging.Formatter(fmt="%(asctime)s %(levelname)s=> %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
handler = logging.handlers.RotatingFileHandler(join(LOG_DIR, "ovs_upgrade.log"), "a", 1024*1024, 3)
handler.setFormatter(formatter)
logging.getLogger("").setLevel(logging.DEBUG)
logging.getLogger("").addHandler(handler)

#======================================================================

def need_upgrade():
    upgrade_flag = db_load("LOCAL_DB,node", "upgrade_flag", get=True)
    return not upgrade_flag

def parse_mount():
    l = []
    l_file = []
    f1 = open("/etc/mtab")
    f2 = open("/etc/fstab")
    try:
        # parse /etc/mtab
        for line in f1:
            fs_spec, fs_file, fs_vfstype = line.split()[0:3]
            if (":" in fs_spec or fs_spec.startswith("/dev")) and fs_file.startswith("/OVS"):
                l.append((fs_spec, fs_file, fs_vfstype))
                l_file.append(fs_file)
        # parse /etc/fstab
        for line in f2:
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            fs_spec, fs_file, fs_vfstype = line.split()[0:3]
            if (":" in fs_spec or fs_spec.startswith("/dev")) and fs_file.startswith("/OVS"):
                if not fs_file in l_file:
                    l.append((fs_spec, fs_file, fs_vfstype))
    finally:
        f1.close()
        f2.close()
    return l


def upgrade():
    l = parse_mount()
    for i in l:
        if i[1] == "/OVS":
            l.remove(i)
            l.append(i)
            break

    logging.debug("SR list: %s" % repr(l))

    root_sr = None
    for fs_spec, fs_file, fs_vfstype in l:
        # umount
        try:
            run_cmd(["sync"])
            run_cmd(["umount", "-fl", fs_file])
        except:
            pass
        # get type
        if fs_vfstype == "nfs":
            sp_type = "OVSSPNFS"
        elif fs_vfstype == "ocfs2":
            sp_type = "OVSSPOCFS2"
        else:
            sp_type = "OVSSPPartition"
        # new repo
        sr_uuid = new_repo(sp_type, fs_spec)

        logging.debug("SR %s created: %s" % (fs_spec, sr_uuid))

        if fs_file == "/OVS":
            root_sr = sr_uuid

    if root_sr:
        master_hostname = db_load("LOCAL_DB,master", "master_hostname", get=True)
        slave_hostname = db_load("LOCAL_DB,master", "slave_hostname", get=True)
        if master_hostname and slave_hostname and master_hostname != slave_hostname:
            logging.debug("Slave node: not sync agent config.")
        else:
            # set root
            set_root(root_sr)

            agt_url = get_local_agt_url()
            agt_sp = ServerProxy(agt_url)

            # sync config
            agt_sp.cluster_sync_agent_config("", root_sr)
            logging.debug("Master node: sync agent config.")

            # init sr
            sr = db_load("sr", root_sr)
            root_sro = pickle.dumps(sr)
            agt_sp.clusterm_init_root_sr(root_sro)
            agt_sp.clusterm_init_sr(True)
        # remove shutdown flag
        d = db_dump("srv")
        for srv in d.keys():
            agt_sp.set_shutdown_mode(srv, False)


def pre_upgrade():
    fstab = "/etc/fstab"
    mtab = "/etc/mtab"
    repos = "/etc/ovs/repositories"

    for e in [fstab, mtab, repos]:
        if exists(e):
            cmd = "cp %s %s.bak" % (e, e)
            old_run_cmd(cmd)
            logging.debug("%s backed up to %s.bak" % (e, e))


def update_master_weight(weight=1):
    hostname = db_load("LOCAL_DB,node", "registered_hostname", get=True)
    if hostname:
        d = db_load("srv", hostname, get=True)
        if d:
            d["master_weight"] = weight
            db_save("srv", hostname, d)


def post_upgrade():
    ha_enable = db_load("LOCAL_DB,master", "pool_ha_enable", get=True)
    hostname = db_load("LOCAL_DB,master", "slave_hostname", get=True)
    master_hostname = db_load("LOCAL_DB,master", "master_hostname", get=True)
    master_agt_url = db_load("LOCAL_DB,master", "master_agt_url", get=True)
    if ha_enable:
        db_save("cluster", "pool_ha_enable", ha_enable)
    if hostname:
        db_save("LOCAL_DB,node", "registered_hostname", hostname)
        ip = get_host_by_name(hostname)
        db_save("LOCAL_DB,node", "registered_ip", ip)
    if master_hostname:
        db_save("cluster", "master_hostname", master_hostname)
    if master_agt_url:
        db_save("cluster", "master_agt_url", master_agt_url)

    # update master_weight of this node
    # master -> 1 and slave -> 0
    if master_hostname and master_hostname == hostname:
        update_master_weight(1)
    else:
        update_master_weight(0)

    # set flag
    db_save("LOCAL_DB,node", "upgrade_flag", True)

    # update /etc/fstab
    old_run_cmd("sed -i '/\\s*\/OVS\\s/d' /etc/fstab")

    # remove /etc/ovs/repositories
    run_cmd(["rm", "-f", "/etc/ovs/repositories"])


if __name__ == "__main__":
    # for 2.1.5 to 2.2.0 upgrade
    if need_upgrade():
        pre_upgrade()
        upgrade()
        post_upgrade()
